#!/bin/sh
# the next line calls 
#     tclsh "$0" "$@"
# where - "$0" is the filename of the script being executed, i.e.  pfm.tcl
#       - "$@" stands for all arguments with which the script was called
# The backslash at the end of this comment makes the next line a continuation
# of the comment for tclsh, but not for sh. So, the next line is only
# executed in 'sh', not in 'tclsh' \
exec tclsh "$0" "$@"
#############################################################################
#
# This is Postgres Forms (pfm), a client application for PostgreSQL.
#
# Copyright (C) 2004-2007 Willem Herremans
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send bug reports and feature requests by E-mail to the address
# that you can find in procedure cmdAbout.
#
# The home page of pfm can be found at
#
# http://pgfoundry.org/projects/pfm/
#############################################################################

############################################################################
# pfm.tcl script arguments                                                 #
############################################################################

# The pfm.tcl script can be called with 0 or 1 argument(s).
#
# The optional argument should be the fully qualified name of the directory
# in which pfm is installed.
#
# If pfm.tcl is called without arguments, pfm.tcl assumes that it is
# installed in the current working directory. If this assumption is
# wrong, pfm will not be able to find the on-line help files and the
# SQL files that are needed to install the pfm_* tables.
#

package require Tk

########################################################################
# Begin namespace options                                              #
########################################################################

# pfmOptions is an array with all the options for Postgres Forms.
# Up to now, the options are:
#
#     - browser : a web browser used to display the help file.
#
#     - dblist : a list of data base names from which the user
#       can choose one at Open data base.
#
#     - host, port, user and dbname: the default values for the
#       connection parameters proposed by Open data base.
#
#     - fontmonospace: a monospace font to use for the GUI
#
#     - fontproportional : a proportional font to use for the GUI
#
#     - printcmd : UNIX command which accepts data on stdin and sends
#       them to the printer.
#
#     - printencoding: character encoding used by pfm to send text to print command.
#
#     - psql: the command used to call psql.
#
#     - tmpdir : the temporary directory used by pfm
#
#     - usePGPASSWORD: whether or not to use the environment variable PGPASSWORD
#                      to store the password.
#
# pfmOptions are stored in file ~/.pfmrc. This file is read by proc initOptions.
#
# On Windows platforms the options are stored in %APPDATA%\pfm\pfm.conf
#
# pfmOptions can be modified by menu Tools -> Options.
#


namespace eval options {
    variable pfmOptions
    variable newOptions
    variable optionList
	
    proc getDefault {option} {
	global env
	global tcl_platform
	variable ::pfm::pfmConfig

	switch -- $option {
	    "browser" {
		switch -- $tcl_platform(platform) {
		    "unix" {
			set value $pfmConfig(defaultBrowser)
		    }
		    "windows" {
			if {[info exists env(ProgramFiles)]} then {
			    set iexplorer \
				[file normalize [file join $env(ProgramFiles) \
						     {Internet Explorer} \
						     {iexplore.exe}]]
			    set value [list $iexplorer %s]
			} else {
			    set value {iexplore.exe %s}
			}
		    }
		    default {
			set value {}
		    }
		}
	    }
	    "dblist" {
		set value [list $tcl_platform(user)]
	    }
	    "dbname" {
		set value $tcl_platform(user)
	    }
	    "fontmonospace" {
		set value [list courier -12]
	    }
	    "fontproportional" {
		set value [list helvetica -12]
	    }
	    "fontstyle" {
		set value normal
	    }
	    "host" {
		set value {}
	    }
	    "port" {
		set value {5432}
	    }
	    "printcmd" {
		switch -- $tcl_platform(platform) {
		    "unix" {
			set value $pfmConfig(defaultPrintcmd)
		    }
		    "windows" {
			if {[info exists env(ProgramFiles)]} then {
			    set wordpad \
				[file normalize [file join $env(ProgramFiles) \
						     {Windows NT} \
						     {Bureau-accessoires} \
						     {wordpad.exe}]]
			    set value [list $wordpad %s]
			} else {
			    set value {wordpad.exe %s}
			}
		    }
		}
	    }
	    "printencoding" {
		set value [encoding system]
	    }
	    "psql" {
		switch -- $tcl_platform(platform) {
		    "unix" {
			set value {psql}
		    }
		    "windows" {
			set value {psql.exe}
		    }
		    default {
			set value {}
		    }
		}
	    }
	    "tmpdir" {
		switch -- $tcl_platform(platform) {
		    "unix" {
			set value {/tmp}
		    }
		    "windows" {
			if {[info exists env(TEMP)]} then {
			    set value [file normalize $env(TEMP)]
			} else {
			    set value [file normalize "~/tmp"]
			}
		    }
		    default {
			set value {}
		    }
		}
	    }
	    "usePGPASSWORD" {
		set value {yes}
	    }
	    "user" {
		set value $tcl_platform(user)
	    }
	    default {
		set value {}
	    }
	}
	return $value
    }

    proc setDefaultOptions {OptionsName erase} {
	# If $erase, all options are set to their default.
	# Else, only the options that are not defined yet are set to
	# their defaults.
	# This makes it possible to add new options in new versions of pfm
	# without having to convert the .pfmrc file.
	# From version 1.0.4. on, this procedure does not call
	# pg_conndefaults anymore.

	upvar $OptionsName Options
	variable optionList
	variable fontFam
	variable fontSize

	foreach option $optionList {
	    if {![info exists Options($option)] || $erase} then {
		set Options($option) [getDefault $option]
		if {[lsearch {fontmonospace fontproportional} $option] \
			>= 0} then {
		    set fontFam($option) [lindex $Options($option) 0]
		    set fontSize($option) \
			[expr - [lindex $Options($option) 1]]
		}
	    }
	}
	return
    }

    proc saveOptions {} {

	variable pfmOptions
	variable optionList
	global tcl_platform
	global env

	if {$tcl_platform(platform) eq {windows}} then {
	    set filename [file join $env(APPDATA) pfm pfm.conf]
	    set dirname [file join $env(APPDATA) pfm]
	    if {![file exists $dirname]} then {
		file mkdir $dirname
	    }
	} else {
	    set filename "~/.pfmrc"
	}
	set rcFile [open $filename w]
	foreach option $optionList {
	    puts $rcFile [list $option $pfmOptions($option)]
	}
	close $rcFile
	return
    }

    proc initOptions { } {

	variable pfmOptions
	variable optionList
	global tcl_platform
	global env

	set optionList {}
	lappend optionList browser
	lappend optionList dblist
	lappend optionList dbname
	lappend optionList fontmonospace
	lappend optionList fontproportional
	lappend optionList fontstyle
	lappend optionList host
	lappend optionList port
	lappend optionList printcmd
	lappend optionList printencoding
	lappend optionList psql
	lappend optionList tmpdir
	lappend optionList usePGPASSWORD
	lappend optionList user
	if {$tcl_platform(platform) eq {windows}} then {
	    set filename [file join $env(APPDATA) pfm pfm.conf]
	} else {
	    set filename "~/.pfmrc"
	}
	if { [file exists $filename]} then {
	    set rcFile [open $filename r]
	    while {![eof $rcFile]} {
		set line [gets $rcFile]
		if {[string length $line] != 0} then {
		    if {[catch {
			set pfmOptions([lindex $line 0]) [lindex $line 1]
		    } errMsg]} then {
			append errMsg "\nError parsing '${line}' in options file '$filename'."
			append errMsg "\nCheck options using Tools -> Options."
			tk_messageBox -type ok -icon error -message $errMsg
		    }
		}
	    }
	    close $rcFile
	    # Next lines make it possible to add new options in new versions of
	    # pfm without having to convert the .pfmrc file: all options not
	    # present in .pfmrc, are set to their default values.
	    setDefaultOptions pfmOptions 0
	    setFonts
	    saveOptions
	} else {
	    setDefaultOptions pfmOptions 1
	    setFonts
	    saveOptions
	}
	return
    }

    proc setFonts {} {
	# This procedure defines the named fonts prop, propbold, mono and
	# monobold.
	# It also defines the rules for the fonts in the options database.
	# From version 1.4.3 on, all fonts are set here, instead of at the
	# time of the widget creation.

	variable pfmOptions

	if {[lsearch -exact [font names] prop] >= 0} then {
	    font delete prop
	}
	if {[lsearch -exact [font names] propbold] >= 0} then {
	    font delete propbold
	}
	if {[lsearch -exact [font names] mono] >= 0} then {
	    font delete mono
	}
	if {[lsearch -exact [font names] monobold] >= 0} then {
	    font delete monobold
	}
	if {$pfmOptions(fontstyle) eq {bold}} then {
	    set weight bold
	} else {
	    set weight normal
	}
	set fontsdefined 0
	while {!$fontsdefined} {
	    if {[catch {
		font create prop \
		    -family [lindex $pfmOptions(fontproportional) 0] \
		    -size [lindex $pfmOptions(fontproportional) 1] \
		    -weight normal
		font create propbold \
		    -family [lindex $pfmOptions(fontproportional) 0] \
		    -size [lindex $pfmOptions(fontproportional) 1] \
		    -weight $weight
	    } errMsg]} then {
		set pfmOptions(fontproportional) [getDefault fontproportional]
		append errMsg "\nOption 'fontproportional' has been reset to $pfmOptions(fontproportional)"
		tk_messageBox -type ok -icon warning -message $errMsg
	    } else {
		set fontsdefined 1
	    }
	}
	set fontsdefined 0
	while {!$fontsdefined} {
	    if {[catch {
		font create mono \
		    -family [lindex $pfmOptions(fontmonospace) 0] \
		    -size [lindex $pfmOptions(fontmonospace) 1] \
		    -weight normal
		font create monobold \
		    -family [lindex $pfmOptions(fontmonospace) 0] \
		    -size [lindex $pfmOptions(fontmonospace) 1] \
		    -weight $weight
	    } errMsg]} then {
		set pfmOptions(fontmonospace) [getDefault fontmonospace]
		append errMsg "\nOption 'fontmonospace' has been reset to $pfmOptions(fontmonospace)"
		tk_messageBox -type ok -icon warning -message $errMsg
	    } else {
		set fontsdefined 1
	    }
	}

	# At http://wiki.tcl.tk/1062, I have found how
	# to change the font of the tk_messageBox:
	#     option add *Dialog.msg.font prop.
	# At http://wiki.tcl.tk/10424, I have found how to put all font
	# settings at 1 point. I have used that here:
	option clear
	option add *font propbold
	option add *report.fmReport.lsb.font monobold
	option add *Entry.font prop
	option add *Text.font mono
	return
    }

    proc getOptionHelp {option} {
	variable ::pfm::installDir
	variable ::pfm::pfmConfig
	
	set fileName [file join $installDir $pfmConfig(helpDir) \
			  help_${option}.txt]
	if {[catch {open $fileName r} helpChannel]} then {
	    set help {No help available}
	} else {
	    set help [read $helpChannel]
	    close $helpChannel
	}
	if {$option eq {printencoding}} then {
	    append help \
		"\nThe following character encodings are available on you system:\n"
	    foreach charEncoding [lsort [encoding names]] {
		append help "\n$charEncoding"
	    }
	}
	return $help
    }

    proc cmdOptions {} {

	variable pfmOptions
	variable newOptions
	variable optionList
	variable fontFam
	variable fontSize

	destroy .options
	toplevel .options -class Toplevel
	wm transient .options .
	set x [expr [winfo rootx .] + 20]
	set y [expr [winfo rooty .] + 20]
	wm geometry .options +$x+$y
	wm title .options "pfm - Options"
	set rowidx 0
	foreach option $optionList {
	    set newOptions($option) $pfmOptions($option)
	    label .options.lb$option -text $option
	    switch -- $option {
		"fontmonospace" -
		"fontproportional" {
		    frame .options.fm$option
		    set fontFam($option) [lindex $newOptions($option) 0]
		    set fontSize($option) [lindex $newOptions($option) 1]
		    set fontSize($option) [expr - $fontSize($option)]
		    button .options.bn$option -text "Default" -underline 1 \
			-pady 0 -takefocus 0 \
			-command [list ::options::getDefaultFont $option]
		    grid .options.lb$option -in .options -column 0 -row $rowidx \
			-columnspan 1 -rowspan 1
		    grid .options.fm$option -in .options -column 1 -row $rowidx \
			-columnspan 3 -rowspan 1 -sticky we
		    grid .options.bn$option -in .options -column 4 -row $rowidx \
			-columnspan 1 -rowspan 1
		    button .options.fm$option.button -pady 0 \
			-textvariable ::options::fontFam($option) \
			-command [list ::options::getFontFam $option]
		    bind .options.fm$option.button <Alt-KeyPress-e> \
			[list ::options::getDefaultFont $option]
		    spinbox .options.fm$option.spinbox \
			-textvariable ::options::fontSize($option) \
			-from -30.0 -to 30.0 -increment 1.0 -format %1.0f \
			-width 4 -background white -takefocus 0
		    pack .options.fm$option.spinbox \
			-in .options.fm$option \
			-side right
		    pack .options.fm$option.button \
			-in .options.fm$option \
			-side right -expand 1 -fill x
		}
		default {
		    entry .options.ent$option \
			-textvar ::options::newOptions($option) \
			-width 40 -background white
		    button .options.bn$option -text "Expand" -underline 1 \
			-pady 0 -takefocus 0 \
			-command "::options::cmdExpand $option .options.ent$option"
		    bind .options.ent$option <Alt-KeyPress-x> \
			"::options::cmdExpand $option .options.ent$option"
		    grid .options.lb$option -in .options -column 0 -row $rowidx \
			-columnspan 1 -rowspan 1
		    grid .options.ent$option -in .options -column 1 -row $rowidx \
			-columnspan 3 -rowspan 1
		    grid .options.bn$option -in .options -column 4 -row $rowidx \
			-columnspan 1 -rowspan 1
		}
	    }
	    incr rowidx
	}
	bind .options <KeyPress-Down> \
	    {focus [tk_focusNext [focus -displayof .options]]}
	bind .options <KeyPress-Up> \
	    {focus [tk_focusPrev [focus -displayof .options]]}

	button .options.btnDefault -text {Defaults} \
	    -underline 0 \
	    -command [namespace code \
			  {setDefaultOptions newOptions 1}]
	button .options.btnOK -text {OK} \
	    -command [namespace code cmdOptionOK] -underline 0
	button .options.btnCancel -text {Cancel} \
	    -command [namespace code cmdOptionCancel] -underline 0
	grid .options.btnDefault -column 1 -row $rowidx -sticky we
	grid .options.btnOK -column 2 -row $rowidx -sticky we
	grid .options.btnCancel -column 3 -row $rowidx -sticky we
	focus .options.ent[lindex $optionList 0]
	bind .options <Alt-KeyPress-d> "pfm::PressButton .options.btnDefault"
	bind .options <Alt-KeyPress-o> "pfm::PressButton .options.btnOK"
	bind .options <Alt-KeyPress-c> "pfm::PressButton .options.btnCancel"
	bind .options <KeyPress-Escape> "pfm::PressButton .options.btnCancel"
	return
    }

    proc getFontFam {option} {
	variable fontfamilies
	variable fontFam
	variable selFam

	set selFam fontFam($option)
	destroy .options.getfont
	toplevel .options.getfont -class Toplevel
	wm transient .options.getfont .options
	set x [winfo rootx .options.fm$option]
	set y [winfo rooty .options.fm$option]
	wm geometry .options.getfont 250x200+$x+$y
	wm title .options.getfont "$option"
	listbox .options.getfont.lsb -background white \
	    -yscrollcommand {.options.getfont.vsb set}
	focus .options.getfont.lsb
	scrollbar .options.getfont.vsb -orient vertical \
	    -command {.options.getfont.lsb yview} -takefocus 0
	place .options.getfont.lsb \
	    -anchor nw -x 0 -y 0 \
	    -width -20 -relwidth 1 \
	    -relheight 1
	place .options.getfont.vsb \
	    -anchor ne -x 0 -y 0 -relx 1 \
	    -width 20 \
	    -relheight 1
	set fontfamilies {}
	foreach family [lsort [font families]] {
	    switch -- $option {
		"fontproportional" {
		    if {![font metrics [list $family] -fixed]} then {
			.options.getfont.lsb insert end $family
			lappend fontfamilies $family
		    }
		}
		"fontmonospace" {
		    if {[font metrics [list $family] -fixed]} then {
			.options.getfont.lsb insert end $family
			lappend fontfamilies $family
		    }
		}
	    }
	}
	.options.getfont.lsb selection clear 0 end
	.options.getfont.lsb selection set 0 0
	.options.getfont.lsb activate 0
	.options.getfont.lsb see 0
	bind .options.getfont.lsb <ButtonPress-1> {
	    after idle {
		set ::options::selFam \
		    [lindex $::options::fontfamilies \
			 [.options.getfont.lsb curselection]]
		destroy .options.getfont
	    }
	}
	bind .options.getfont.lsb <KeyPress-Return> {
	    set ::options::selFam \
		[lindex $::options::fontfamilies \
		     [.options.getfont.lsb curselection]]
	    destroy .options.getfont
	}
	bind .options.getfont.lsb <KeyPress-Escape> {
	    after idle {
		destroy .options.getfont
	    }
	}
	tkwait window .options.getfont.lsb
	set fontFam($option) $selFam
	return
    }

    proc getDefaultFont {option} {
	variable fontFam
	variable fontSize

	set defFont [getDefault $option]
	set fontFam($option) [lindex $defFont 0]
	set fontSize($option) [expr - [lindex $defFont 1]]
	return
    }

    proc cmdExpand {option optionWidget} {
	variable wrapOn

	set wrapOn 1
	destroy .options.expand
	toplevel .options.expand -class Toplevel
	wm transient .options.expand .options
	set x [expr [winfo rootx .options] + 20]
	set y [expr [winfo rooty .options] + 20]
	wm geometry .options.expand 600x400+$x+$y
	wm title .options.expand "pfm - Option: $option"

	# Help text and vertical scrollbar
	text .options.expand.help -background LightYellow2 -wrap word \
	    -takefocus 0 \
	    -yscrollcommand {.options.expand.helpscroll set}
	.options.expand.help insert end "Help for option '$option':\n\n"
	set endIdx [.options.expand.help index "end -1 chars"]
	.options.expand.help tag add helpTitle 1.0 $endIdx
	.options.expand.help tag configure helpTitle -foreground {medium blue}
	.options.expand.help insert end [getOptionHelp $option]
	.options.expand.help configure -state disabled
	scrollbar .options.expand.helpscroll -orient vertical \
	    -command {.options.expand.help yview} -takefocus 0
	place .options.expand.help \
	    -anchor nw -x 0 -y 0 \
	    -width -20 -relwidth 1 \
	    -relheight 0.4
	place .options.expand.helpscroll \
	    -anchor ne -x 0 -y 0 -relx 1 \
	    -width 20 \
	    -relheight 0.4 

	# Expanded options and scrollbars
	text .options.expand.text -background white -wrap word \
	    -yscrollcommand {.options.expand.vsb set} \
	    -xscrollcommand {.options.expand.hsb set}
	scrollbar .options.expand.vsb -orient vertical -command \
	    {.options.expand.text yview} -takefocus 0
	scrollbar .options.expand.hsb -orient horizontal -command \
	    {.options.expand.text xview} -takefocus 0
	place .options.expand.text \
	    -anchor nw -x 0 -rely 0.4 \
	    -relwidth 1 -width -20 \
	    -relheight 0.6 -height -50	    
	place .options.expand.vsb \
	    -anchor ne -relx 1 -rely 0.4  \
	    -width 20 \
	    -relheight 0.6 -height -50
	place .options.expand.hsb \
	    -anchor sw -x 0 -y -30 -rely 1 \
	    -relwidth 1 -width -20 \
	    -height 20
	focus .options.expand.text

	# Buttons frame
	frame .options.expand.buttons
	place .options.expand.buttons \
	    -anchor sw -x 0 -y 0 -rely 1 \
	    -relwidth 1 \
	    -height 30

	# Buttons
	button .options.expand.buttons.bnBrowse -text {Paste filename} \
	    -underline 0 -takefocus 0 \
	    -command {
		.options.expand.text insert insert \
		    [file normalize \
			 [tk_getOpenFile -title {Paste filename} \
			      -parent .options.expand]]
	    }
	button .options.expand.buttons.bnDefault -text {Default} \
	    -takefocus 0 -underline 0 \
	    -command "::options::cmdOneDefault $option"
	button .options.expand.buttons.bnOK -text {OK} \
	    -takefocus 0 -underline 0 \
	    -command "::options::cmdExpandOK $optionWidget"
	radiobutton .options.expand.buttons.rbWrap -text {Wrap} -value 1 \
	    -variable ::options::wrapOn \
	    -command {.options.expand.text configure -wrap word} \
	    -takefocus 0 -underline 0
	radiobutton .options.expand.buttons.rbTruncate -text {Truncate} \
	    -value 0 -takefocus 0 -underline 0 \
	    -variable ::options::wrapOn \
	    -command {.options.expand.text configure -wrap none}
	button .options.expand.buttons.bnCancel -text {Cancel} \
	    -takefocus 0 -underline 0 \
	    -command {destroy .options.expand}
	bind .options.expand <Alt-KeyPress-t> \
	    "::pfm::PressButton .options.expand.buttons.rbTruncate"
	bind .options.expand <Alt-KeyPress-w> \
	    "::pfm::PressButton .options.expand.buttons.rbWrap"
	bind .options.expand <Alt-KeyPress-d> \
	    "::pfm::PressButton .options.expand.buttons.bnDefault"
	bind .options.expand <Alt-KeyPress-p> \
	    "::pfm::PressButton .options.expand.buttons.bnBrowse"
	bind .options.expand <Alt-KeyPress-o> \
	    "::pfm::PressButton .options.expand.buttons.bnOK"
	bind .options.expand <Alt-KeyPress-c> \
	    "::pfm::PressButton .options.expand.buttons.bnCancel"
	bind .options.expand <KeyPress-Escape> \
	    "::pfm::PressButton .options.expand.buttons.bnCancel"
	pack .options.expand.buttons.bnCancel -side right
	pack .options.expand.buttons.bnOK -side right
	pack .options.expand.buttons.bnBrowse -side right
	pack .options.expand.buttons.bnDefault -side right
	pack .options.expand.buttons.rbWrap -side left
	pack .options.expand.buttons.rbTruncate -side left
	.options.expand.text insert end [$optionWidget get]
	return
    }

    proc cmdOneDefault {option} {

	.options.expand.text delete 1.0 end
	.options.expand.text insert end [getDefault $option]
	return
    }

    proc cmdExpandOK { optionWidget} {

	$optionWidget delete 0 end
	$optionWidget insert 0 [.options.expand.text get 1.0 "end -1 chars"]
	destroy .options.expand
	return
    }

    proc cmdOptionOK {} {

	variable pfmOptions
	variable newOptions
	variable fontFam
	variable fontSize
	variable ::pfm::currentDB
	variable optionList
	destroy .options
	foreach option $optionList {
	    switch -- $option {
		"fontmonospace" -
		"fontproportional" {
		    if {[catch {
			set pfmOptions($option) \
			    [list $fontFam($option) [expr -$fontSize($option)]]
			# puts $pfmOptions($option)
		    } errMsg]} then {
			append errMsg "\nOption '$option' has not been modified."
			tk_messageBox -type ok -icon error -message $errMsg
		    }
		}
		default {
		    set pfmOptions($option) $newOptions($option)
		}
	    }
	}
	setFonts
	saveOptions
	if {[info exists currentDB]} {
	    ::pfm::refreshFormsList
	}
	return
    }

    proc cmdOptionCancel {} {

	destroy .options
	return
    }

    proc addToDBlist {dbName} {
	variable pfmOptions

	if { [lsearch $pfmOptions(dblist) $dbName] == -1 } then {
	    lappend pfmOptions(dblist) $dbName
	    set pfmOptions(dblist) [lsort $pfmOptions(dblist)]
	}
	set pfmOptions(dbname) $dbName
	saveOptions
	return
    }


    ###############################################################
    # Main of namespace options                                   #
    ###############################################################

    # initOptions has to be postponed until configuration has
    # been read. It is moved to the main of namespace pfm


}

###############################################################
# End of namespace options                                    #
###############################################################


################################################################
#                                                              #
# Begin of namespace pfm                                       #
#                                                              #
################################################################

# widget is an associative array containing aliases for widget path names
#
# dbName is the name of the currently open data base. It is filled out
# by proc cmdOpenOK.
#
# passMatrix an associative array which is filled by readPgPass with the 
# contents of the ~/.pgpass file:
#      passMatrix($i,hostname) = hostname of $i-th entry in pgpass
#      passMatrix($i,port) = port of $i-th entry in pgpass
#      passMatrix($i,database) = database name of $i-th entry in pgpass
#      passMatrix($i,username) = username of $i-th entry in pgpass
#      passMatrix($i,password) = password of $i-th entry in pgpass
#
# psqlChannel the channel linked to the psql command pipeline. It is set
# in cmdOpenOK and unset in cmdCloseDataBase.
#
# currentDB contains the postgres data base handle for the open data base
# It is filled in by proc cmdOpenDataBase, and it is used throughout this
# application.
#
# formsArray contains all the tuples of pfm_form, where
# formsArray($name,$attribute) contains the value of attribute '$attribute'
# in the tuple of pfm_form for which name=$name. It is also filled by 
# refreshFormsList.
#
# formsIndex is an array for which formsIndex($n) contains the
# name of the form in the n-th form in the forms listbox. It is
# filled by refreshFormsList.
#
# connInfoList: the list of parameters for the conninfo in pg_connect.
#
# pfmMode can be either:
#
#    -  "normal": only the forms with showform = true are displayed;  or
#
#    -  "design": only the forms with showform = false are displayed.
#
# pfmConfig is an array containing configuration parameters. On UNIX platforms,
# pfmConfig is read from the config-file pfm.conf in pfm's installation directory 
# or in /etc. On the Windows platform pfmConfig is hardcoded.
# Currently the array elements are:
#
#    - docDir : the directory in which the on-line documentation is installed.
#    - helpDir : the directory in which the options on-line help texts and the
#                license are installed.
#    - exampleDir : the directory in which the example databases are installed.
#    - defaultBrowser : the default value for the "browser" option
#    - defaultPrintcmd : the default value for the "printcmd" option.

namespace eval pfm {

    variable pfmVersion {1.4.3}
    variable API
    variable pfmConfig
    variable installDir
    variable currentDB
    variable psqlChannel
    variable errChannel
    variable formsArray
    variable formsIndex
    variable widget
    variable dbName
    variable pfmMode {normal}
    variable connInfoList
    variable passMatrix

    proc initRootWindow { } {

	variable widget

	wm title . "pfm - No data base opened"
	wm geometry . 350x300

	# Menubar of root window
	menu .menubar -tearoff 0
	.menubar add cascade -menu .menubar.database -label {Database} \
	    -underline 0
	.menubar add cascade -menu .menubar.reports \
	    -label {Reports/Queries} -underline 0 -state disabled
	.menubar add cascade -menu .menubar.tools -label {Tools} \
	    -underline 0
	.menubar add cascade -menu .menubar.help -label {Help} \
	    -underline 0

	# Make it the root window's bar menu
	. configure -menu .menubar

	# Database pull down menu
	menu .menubar.database -tearoff 0
	.menubar.database add command \
	    -command [namespace code {cmdOpenDataBase}] \
	    -label "Open ..." -underline 0 \
	    -accelerator {Cntrl-o}
	.menubar.database add command \
	    -command [namespace code cmdCloseDataBase] -label Close \
	    -state disabled -underline 0 \
	    -accelerator {Cntrl-w}
	.menubar.database add command \
	    -command {destroy .} -label Quit \
	    -underline 0 \
	    -accelerator {Cntrl-q}

	# Reports/Queries pull down menu
	menu .menubar.reports -tearoff 0
	.menubar.reports add command \
	    -command {::report::cmdReportSQL sql} -label "Run SQL" \
	    -underline 4 \
	    -accelerator {Cntrl-s}
	.menubar.reports add command \
	    -command {::report::cmdReportSQL report} -label "Run report" \
	    -underline 4 \
	    -accelerator {Cntrl-r}

	# Tools pull down menu
	menu .menubar.tools -tearoff 0
	.menubar.tools add command \
	    -command ::pfm::installPfm \
	    -label "Install pfm_* tables" -state disabled \
	    -underline 8
	.menubar.tools add command \
	    -command ::pfm::cmdInstallExample \
	    -label "Install example database ..." -state disabled \
	    -underline 8
	.menubar.tools add command \
	    -command ::options::cmdOptions -label "Options" \
	    -underline 0

	# Help menu
	menu .menubar.help -tearoff 0
	.menubar.help add command \
	    -command ::help::cmdDisplayManual -label {Help file} \
	    -underline 0 -accelerator {F1}
	.menubar.help add command \
	    -command ::help::cmdLicense -label {License} \
	    -underline 0
	.menubar.help add command \
	    -command [namespace code cmdAbout] -label About \
	    -underline 0
	bind . <KeyPress-F1> ::help::cmdDisplayManual

	# The radio buttons
	radiobutton .rbNormal -text {Normal mode} -value {normal} \
	    -variable ::pfm::pfmMode -command ::pfm::refreshFormsList \
	    -state disabled -takefocus 0 \
	    -underline 0
	radiobutton .rbDesign -text {Design mode} -value {design} \
	    -variable ::pfm::pfmMode -command ::pfm::refreshFormsList \
	    -state disabled -takefocus 0 \
	    -underline 1
	bind . <Alt-KeyPress-n> "::pfm::PressButton .rbNormal"
	bind . <Alt-KeyPress-e> "::pfm::PressButton .rbDesign"
 	place .rbNormal \
	    -anchor n -x 0 -y 0 -relx 0.25
 	place .rbDesign \
	     -anchor n -x 0 -y 0 -relx 0.75

	# The open button

	button .btnOpen \
	    -command {::pfm::cmdOpenQuery} \
	    -padx 0 -text {Open form} \
	    -underline 0 \
	    -state disabled
	bind . <Alt-KeyPress-o> {::pfm::PressButton .btnOpen}
	place .btnOpen \
	    -anchor sw -x 0 -rely 1 \
	    -relwidth 1

	# Listbox + Scrollbar  + Title
	label .lbList -text "List of forms"
	listbox .lsbForms \
	    -yscrollcommand {.vsb set} -background white
	scrollbar .vsb -orient vertical \
	    -command {.lsbForms yview} -takefocus 0
	set widget(lsbForms) .lsbForms
	bind .lsbForms <KeyPress-Return> ::pfm::cmdOpenQuery
	set alreadyUsedHeight 0
	incr alreadyUsedHeight [winfo reqheight .rbNormal]
	incr alreadyUsedHeight [winfo reqheight .lbList]
	incr alreadyUsedHeight [winfo reqheight .btnOpen]
 	place .lbList \
	    -anchor n -x 0 -y 20 -relx 0.5
 	place .lsbForms \
	    -anchor nw -x 0 -y 40 \
	    -width -20 -relwidth 1 \
	    -height -$alreadyUsedHeight -relheight 1 
 	place .vsb \
	    -anchor ne  -x 0 -y 40 -relx 1 \
	    -width 20 \
	    -height -$alreadyUsedHeight -relheight 1

	focus .lsbForms


	# Bindings

	# Accelerators

	bind . <Control-KeyPress-o> {::pfm::cmdOpenDataBase}
	bind . <Control-KeyPress-w> {::pfm::cmdCloseDataBase}
	bind . <Control-KeyPress-q> {destroy .}
	bind . <Control-KeyPress-s> {::report::cmdReportSQL sql}
	bind . <Control-KeyPress-r> {::report::cmdReportSQL report}

	# Cleanup no matter how user exits pfm.
	# Bug 1052: The initial solution:
	#     "bind . <Destroy> ::pfm::cmdExit"
	# caused cmdExit to be called for every <Destroy> event to any of
	# the root window's child widgets. So, when pfm's main window is
	# destroyed, cmdExit, which in turn calls "exit", wass called multiple
	# times. Sometimes, this caused the Tcl/Tk interpreter to get locked.
	# Here is the more sofisticated solution.

	# Add a bindtag to the toplevel .
	set tags [bindtags .]
	lappend tags MainWindow
	bindtags . $tags
	bind MainWindow <Destroy> ::pfm::cmdExit
	# This binding causes cmdExit to be called only when
	# pfm's main window is destroyed.

	# Use Return rather than Space to invoke buttons.
	bind Button <KeyPress-Return> {event generate %W <KeyPress-space>}

	return
    }

    proc cmdAbout {} {
	variable ::pfm::installDir
	variable API
	variable pfmVersion

	set aboutMsg \
	    "Postgres Forms (pfm) Version $pfmVersion. Copyright (C)"
	append aboutMsg \
	    " Willem Herremans 2004-2007\n\nPostgres Forms comes with ABSOLUTELY NO WARRANTY;"
	append aboutMsg \
	    " see 'Help -> License' for details."
	append aboutMsg \
	    "\n\nThis is free software, and you are welcome to redistribute it under"
	append aboutMsg " certain conditions; see 'Help -> License' for details."
	append aboutMsg \
	    "\n\nSend bug reports or other comments to comments.pfm@scarlet.be."
        append aboutMsg "\n\n$API"
	append aboutMsg "\n\npfm is installed in $installDir"
	::pfm::pfmDialog .about "pfm - About" $aboutMsg 400 0 "OK"
	return
    }


    proc cmdOpenDataBase { } {
	# Let the user specify a data base to open

	variable ::options::pfmOptions
	variable connInfoList
	variable currentDB
	variable dbName
	if { [string equal -nocase -length 1 $pfmOptions(usePGPASSWORD) y] } then {
	    set connInfoList {host port user password}
	    # The password will be prompted for and stored in PGPASSWORD.
	} else {
	    set connInfoList {host port user}
	    # The password will not be prompted for. A properly configured
	    # ~/.pgpass file is required.
	}
	if { ![info exists currentDB]} then {
	    set dbName $pfmOptions(dbname)
	    destroy .opendb
	    toplevel .opendb -class Toplevel
	    wm transient .opendb .
	    set x [expr [winfo rootx .] + 20]
	    set y [expr [winfo rooty .] + 20]
	    wm geometry .opendb +$x+$y
	    wm title .opendb "pfm - Open data base"
	    set rowidx 0
	    foreach connItem $connInfoList {
		if { [info exists pfmOptions($connItem)] } then {
		    set connInfo $pfmOptions($connItem)
		} else {
		    set connInfo {}
		}
		label .opendb.lbl$connItem -text $connItem
		entry .opendb.val$connItem -width 30 -background white
		if { [string equal $connItem {password}] } then {
		    .opendb.val$connItem configure -show *
		}
		grid .opendb.lbl$connItem -in .opendb -row $rowidx -column 0 \
			-rowspan 1 -columnspan 1
		grid .opendb.val$connItem -in .opendb -row $rowidx -column 1 \
			-rowspan 1 -columnspan 3 -sticky {we}
		.opendb.val$connItem insert end $connInfo
		incr rowidx
	    }
	    label .opendb.lbldbname -text {dbname}
	    grid .opendb.lbldbname -in .opendb -row $rowidx -column 0 \
		-rowspan 1 -columnspan 1
	    entry .opendb.valdbname -width 30 -background white \
		-textvariable ::pfm::dbName
	    if {[string equal -nocase -length 1 $pfmOptions(usePGPASSWORD) y]} then {
		focus .opendb.valpassword
	    } else {
		focus .opendb.valdbname
	    }
	    button .opendb.btndbname -image ::img::down \
		-command ::pfm::cmdSelectDB
	    grid .opendb.valdbname -in .opendb -row $rowidx -column 1 \
		-rowspan 1 -columnspan 2
	    grid .opendb.btndbname -in .opendb -row $rowidx -column 3 \
		-rowspan 1 -columnspan 1
	    button .opendb.btnOK -text OK \
		-command {::pfm::cmdOpenOK} -underline 0
	    button .opendb.btnCancel -text Cancel \
		-command {::pfm::cmdOpenCancel} -underline 0
	    bind .opendb.valdbname <KeyPress-Return> \
		"::pfm::PressButton .opendb.btnOK"
	    bind .opendb <Alt-KeyPress-o> \
		"::pfm::PressButton .opendb.btnOK"
	    bind .opendb <Alt-KeyPress-c> \
		"::pfm::PressButton .opendb.btnCancel"
	    bind .opendb <KeyPress-Escape> \
		"::pfm::PressButton .opendb.btnCancel"
	    incr rowidx
	    grid .opendb.btnOK -column 1 -row $rowidx -sticky we
	    grid .opendb.btnCancel -column 2 -row $rowidx -sticky we
	} else {
	    tk_messageBox -message "First close data base $dbName" -type ok \
		-icon info -parent .
	}
	return
    }

    proc cmdSelectDB {} {
	variable ::options::pfmOptions
	variable dbName

	destroy .opendb.selectDB
	toplevel .opendb.selectDB -class Toplevel
	wm transient .opendb.selectDB .opendb
	# set x [expr [winfo pointerx .opendb] - 200]
	# set y [winfo pointery .opendb]
	set x [winfo rootx .opendb.valdbname]
	set y [winfo rooty .opendb.valdbname]
	wm geometry .opendb.selectDB 250x200+$x+$y
	wm title .opendb "pfm - Select database"
	listbox .opendb.selectDB.lsb -background white \
	    -yscrollcommand {.opendb.selectDB.vsb set}
	focus .opendb.selectDB.lsb
	scrollbar .opendb.selectDB.vsb -orient vertical \
	    -command {.opendb.selectDB.lsb yview} -takefocus 0
	place .opendb.selectDB.lsb \
	    -anchor nw -x 0 -y 0 \
	    -width -20 -relwidth 1 \
	    -relheight 1
	place .opendb.selectDB.vsb \
	    -anchor ne -x 0 -y 0 -relx 1 \
	    -width 20 \
	    -relheight 1 
	foreach item $pfmOptions(dblist) {
	    .opendb.selectDB.lsb insert end $item
	}
	set currentItem [lsearch $pfmOptions(dblist) $dbName]
	if {$currentItem < 0} then {
	    set currentItem 0
	}
	.opendb.selectDB.lsb selection clear 0 end
	.opendb.selectDB.lsb selection set $currentItem $currentItem
	.opendb.selectDB.lsb activate $currentItem
	.opendb.selectDB.lsb see $currentItem	
	bind .opendb.selectDB.lsb <ButtonPress-1> {
	    after idle {
		set ::pfm::dbName [lindex $::options::pfmOptions(dblist) \
				       [.opendb.selectDB.lsb curselection]]
		destroy .opendb.selectDB
		focus .opendb.btnOK
	    }
	}
	bind .opendb.selectDB.lsb <KeyPress-Return> {
	    set ::pfm::dbName [lindex $::options::pfmOptions(dblist) \
				   [.opendb.selectDB.lsb curselection]]
	    destroy .opendb.selectDB
	    focus .opendb.btnOK
	}
	bind .opendb.selectDB.lsb <KeyPress-Escape> {
	    after idle {
		destroy .opendb.selectDB
		focus .opendb.valdbname
	    }
	}
	return
    }

    proc cmdOpenOK {} {
	global env
	variable connInfoList
	variable currentDB
	variable dbName
	variable ::options::pfmOptions

	set env(PGCLIENTENCODING) "UNICODE"
	set connInfo {}
	lappend connInfo "dbname = '$dbName'"
	set dbHost {}
	set dbPort {}
	set dbUser {}
	set dbPassword {}
	foreach connItem $connInfoList {
	    set connItemValue [.opendb.val$connItem get]
	    if { ![string equal $connItemValue {}] } then {
		lappend connInfo "$connItem = $connItemValue"
	    }
	    switch $connItem {
		host {
		    set dbHost $connItemValue
		}
		port {
		    set dbPort $connItemValue
		}
		user {
		    set dbUser $connItemValue
		}
		password {
		    set dbPassword $connItemValue
		}
	    }
	}
	if { [string equal -nocase -length 1 $pfmOptions(usePGPASSWORD) n] } then {
	    # Get password from ~/.pgpass
	    # Pgtcl is able to read the .pgpass file, but pgin.tcl is not.
	    # That is why pfm reads it.
	    set dbPassword [findPassword $dbHost $dbPort $dbName $dbUser]
	    if { $dbPassword ne {} } then {
		lappend connInfo "password = $dbPassword"
	    }
	}
	set connInfo [join $connInfo]
	if { [catch {set currentDB [pg_connect -conninfo $connInfo]} errorMsg]} then {
	    # Database could not be opened for some reason.
	    tk_messageBox -message $errorMsg -type ok -icon error -parent .opendb
	} else {
	    openPsql $dbName $dbHost $dbPort $dbUser $dbPassword
	    destroy .opendb
	    wm title . "pfm - Database : $dbName"
	    check_pfm_tables
	    refreshFormsList
	    .rbNormal configure -state normal
	    .rbDesign configure -state normal
	    .btnOpen configure -state normal
	    # disable Database -> Open
	    .menubar.database entryconfigure 0 -state disabled
	    # enable Database -> Close
	    .menubar.database entryconfigure 1 -state normal
	    # enable Reports and Querries
	    .menubar entryconfigure 1 -state normal
	    ::options::addToDBlist $dbName
	}
	return
    }

    proc openPsql {dbName dbHost dbPort dbUser dbPassword} {
	variable psqlChannel
	variable errChannel
	global env
	variable ::options::pfmOptions

	if { [info exists psqlChannel] } then {
	    close $psqlChannel
	    unset psqlChannel
	}
	set openCommand "|"
	lappend openCommand $pfmOptions(psql)
	lappend openCommand {--echo-queries}
	if { ![string equal $dbName {}] } then {
	    lappend openCommand "--dbname"
	    lappend openCommand $dbName
	}
	if { ![string equal $dbHost {}] } then {
	    lappend openCommand "--host"
	    lappend openCommand $dbHost
	}
	if { ![string equal $dbPort {}] } then {
	    lappend openCommand "--port"
	    lappend openCommand $dbPort
	}
	if { ![string equal $dbUser {}] } then {
	    lappend openCommand "--username"
	    lappend openCommand $dbUser
	}
	if { [string equal -nocase -length 1 $pfmOptions(usePGPASSWORD) y] } then {
	    set env(PGPASSWORD) $dbPassword
	}
	lappend openCommand "2>@$errChannel"
	# Connecting to psql.
	if { [catch { open $openCommand RDWR } psqlChannel] } then {
	    # For security reasons, the PGPASSWORD environment variable is cleared.
	    if { [string equal -nocase -length 1 $pfmOptions(usePGPASSWORD) y] } then {
		unset env(PGPASSWORD)
	    }
	    tk_messageBox -message $psqlChannel -type ok -icon error -parent .opendb
	    unset psqlChannel
	} else {
	    # For security reasons, the PGPASSWORD environment variable is cleared.
	    if { [string equal -nocase -length 1 $pfmOptions(usePGPASSWORD) y] } then {
		unset env(PGPASSWORD)
	    }
	    # When openeing a database, pfm puts the environment variable 
	    # PGCLIENTENCODING UNICODE, which means in fact utf-8 (see
	    # chapter '20.2 Character set support' of PostgreSQL documentation.
	    # The next statement puts the command pipeline to utf-8 as well.
	    fconfigure $psqlChannel -encoding utf-8
	    fileevent $psqlChannel readable ::report::showResult
	}
	return
    }

    proc readPgPass { } {
	variable passMatrix
	global tcl_platform
	global env

	# This procedure reads the ~/.pgpass file if it exists and if it
	# has the right permissions (00600, i.e. rw for owner only).
	# It parses this file and stores the result in passMatrix.
	# This procedure supports the backslash escape for : and backslash.
	# backslash backslash is read as backslash
	# backslash ':' is read as ':' and not interpreted as entry separator
	# backslash 'anything else' is read as 'anything else'
	#                                      (i.e. backslash is dropped)
	# ':' is interpreted as entry separator

	# On Windows platforms, the pgpass file is
	# %APPDATA%\postgresql\pgpass.conf

	set seqnr 0
	if {$tcl_platform(platform) eq {windows}} then {
	    set filename [file join $env(APPDATA) postgresql pgpass.conf]
	} else {
	    set filename "~/.pgpass"
	}
	if { [file exists $filename] } then {
	    if {$tcl_platform(platform) eq {unix}} then {
	        set filePermission [file attributes $filename -permissions]
	        set first [expr [string length $filePermission] - 3]
	        set filePermission [string range $filePermission $first end]
	    } else {
	        set filePermission "600"
	    }
	    if { $filePermission ne "600" } then {
		set map {0 --- 1 --x 2 -w- 3 -wx 4 r-- 5 r-x 6 rw- 7 rwx}
		set filePermission [string map $map $filePermission]
		set errMsg "The permissions on ~/.pgpass are '$filePermission'."
		set errMsg "$errMsg\nThey should be 'rw-------'"
		tk_messageBox -type ok -icon error -message $errMsg -parent .opendb
	    } else {
		if { [catch {open $filename r} pgPass ] } then {
		    tk_messageBox -type ok -icon error -message $pgPass \
			-parent .opendb
		} else {
		    set argList {hostname port database username password}
		    while { ![eof $pgPass] } {
			if {[gets $pgPass current_line] > 0} then {
			    incr seqnr
			    foreach name $argList {
				set passMatrix($seqnr,$name) {}
			    }
			    set arg {}
			    set argNr 0
			    set last [expr [string length $current_line] - 1]
			    for {set i 0} {$i <= $last} {incr i} {
				set curChar [string index $current_line $i]
				switch -- $curChar {
				    "\\" {
					# This is the way to write 1 backslash:
					# NOT with curly braces.
					# Skip the backslash and copy the next character
					incr i
					append arg [string index $current_line $i]
				    }
				    ":" {
					# end of an arg
					set name [lindex $argList $argNr]
					if {$name ne {}} then {
					    set passMatrix($seqnr,$name) $arg
					}
					# puts "$seqnr, $name : $arg"
					set arg {}
					incr argNr
				    }
				    default {
					# just copy the character
					append arg $curChar
				    }
				}
			    }
			    # We are at end of line. Just copy the last arg.
			    set name [lindex $argList $argNr]
			    if {$name ne {}} then {
				set passMatrix($seqnr,$name) $arg
			    }
			    # puts "$seqnr, $name : $arg"
			    set arg {}
			    incr argNr
			}
		    }
		    close $pgPass
		}
	    }
	}
	return $seqnr
    }

    proc findPassword {hostname port database username} {
	variable passMatrix

	# This procedure tries to get the password from ~/.pgpass
	# It returns the found password. If it does not find
	# a password, it returns the empty string.

	set nr_of_lines [readPgPass]
	set found 0
	set password {}
	for {set seqnr 1} {($seqnr <= $nr_of_lines) && (!$found)} {incr seqnr} {
	    if { ([string equal $hostname $passMatrix($seqnr,hostname)] || \
		      [string equal {*} $passMatrix($seqnr,hostname)]) && \
		     ([string equal $port $passMatrix($seqnr,port)] || \
			  [string equal {*} $passMatrix($seqnr,port)]) && \
		     ([string equal $database $passMatrix($seqnr,database)] || \
			  [string equal {*} $passMatrix($seqnr,database)]) && \
		     ([string equal $username $passMatrix($seqnr,username)] || \
			  [string equal {*} $passMatrix($seqnr,username)]) } then {
		set found 1
		set password $passMatrix($seqnr,password)
	    }
	}
	array unset passMatrix
	return $password
    }



    proc cmdOpenCancel {} {

	destroy .opendb
	return
    }

    proc cmdCloseDataBase {} {
	#Close data base that is currently open

	variable currentDB
	variable dbName
	variable formsArray
	variable widget
	variable psqlChannel

	if { [info exists currentDB] } then {
	    pg_disconnect $currentDB
	    unset currentDB
	}
	if { [info exists psqlChannel] } then {
	    if { [catch {close $psqlChannel} errMsg] } then {
		tk_messageBox -message $errMsg -type ok -icon error -parent .
	    }
	    unset psqlChannel
	}
	array unset formsArray
	$widget(lsbForms) delete 0 end
	$widget(lsbForms) see 0
	wm title . "pfm - No data base opened"
	.btnOpen configure -state disabled
	.rbNormal configure -state disabled
	.rbDesign configure -state disabled
	# enable Database -> Open ...
	.menubar.database entryconfigure 0 -state normal
	# disable Database -> Close
	.menubar.database entryconfigure 1 -state disabled
	# disable reports/querries
	.menubar entryconfigure 1 -state disabled
	# disable install pfm_* tables and install example database
	.menubar.tools entryconfigure 0 -state disabled
	.menubar.tools entryconfigure 1 -state disabled
	destroy .form
	destroy .query
	destroy .report
	return
    }

    proc cmdExit {} {

	variable currentDB
	variable ::options::pfmOptions

	if {[info exists currentDB]} {
	    cmdCloseDataBase
	}
	# remove temporary files of this session
	set tmpFile [file join $pfmOptions(tmpdir) "pfm_$pfmOptions(user)_[pid].sql"]
	if {[file exists $tmpFile]} then {
	    catch {file delete $tmpFile}
	}
	set tmpFile [file join $pfmOptions(tmpdir) "pfm_$pfmOptions(user)_[pid].txt"]
	if {[file exists $tmpFile]} then {
	    catch {file delete $tmpFile}
	}
	exit
    }

    proc refreshFormsList {} {

	variable currentDB
	variable ::options::pfmOptions
	variable formsArray
	variable formsIndex
	variable widget
	variable pfmMode

	array unset formsArray
	array unset formsIndex
	switch $pfmMode {
	    design {
		set sqlwhere "WHERE showform='f'"
	    }
	    normal -
	    default {
		set sqlwhere "WHERE showform='t'"
	    }
	}
	set formQuery "SELECT name,tablename,pkey,showform,view,sqlselect,sqlfrom,groupby"
	append formQuery " FROM pfm_form $sqlwhere ORDER BY name"
	set resQuery [pg_exec $currentDB $formQuery]
	pg_result $resQuery -assignbyidx formsArray
	set lastTuple [expr [pg_result $resQuery -numTuples] -1]
	$widget(lsbForms) delete 0 end
	set listIndex 0
	for {set tupleNr 0} {$tupleNr <= $lastTuple} {incr tupleNr} {
	    set tuple [pg_result $resQuery -getTuple $tupleNr]
	    $widget(lsbForms) insert end [lindex $tuple 0]
	    set formsIndex($listIndex) [lindex $tuple 0]
	    incr listIndex
	}
	foreach index [array names formsIndex] {
	    set form $formsIndex($index)
	    if {$formsArray($form,pkey) eq {}} then {
		set formsArray($form,view) {t}
	    }
	}
	pg_result $resQuery -clear
	$widget(lsbForms) selection clear 0 end
	$widget(lsbForms) selection set 0 0
	$widget(lsbForms) see 0
	$widget(lsbForms) activate 0
	return
    }


    proc installPfm {} {

	variable installDir
	variable currentDB

	if { ![info exists currentDB] } then {
	    tk_messageBox -message "There is no database open!" -type ok \
		-icon error
	} else {
	    ::report::cmdReportSQL {sql}
	    set fileName [file join $installDir install_pfm.sql]
	    if { ![file exists $fileName] } then {
		set errMsg "$fileName does not exist!"
		tk_messageBox -message $errMsg -icon error -type ok -parent .report
	    } else {
		.report.fmSQL.text delete 1.0 end
		set convertedFile [::report::ConvertToUTF-8 $fileName {iso8859-1}]
		if {$convertedFile ne {}} then {
		    .report.fmSQL.text insert end "\\i '$convertedFile'"
		    set msgInfo "The pfm_* tables will be installed."
		    append msgInfo "\nPlease, check the output that will be generated."
		    append msgInfo "\nThen close window \"Run SQL\"."
		    tk_messageBox -message $msgInfo -type ok -icon info -parent .report
		    ::report::cmdRunSQL
		    tkwait window .report
		    refreshFormsList
		    # disable install pfm_* tables and install example database
		    .menubar.tools entryconfigure 0 -state disabled
		    .menubar.tools entryconfigure 1 -state disabled
		} 
	    }
	}
	return
    }

    proc cmdInstallExample {} {

	variable installDir
	variable currentDB
	variable dbName
	variable pfmConfig

	if { ![info exists currentDB] } then {
	    tk_messageBox -message "There is no database open!" -type ok \
		-icon error
	} else {
	    set message \
		"You are about to install an example database in the "
	    append message \
		"currently opened database, which is '${dbName}'. "
	    append message \
		"This database should be completely empty. "
	    append message \
		"It should not even contain the pfm_*  tables. "
	    append message \
		"Is this OK?"
	    set proceed [tk_messageBox -message $message -type yesno \
			     -icon warning -parent .]
	    if { [string equal $proceed "yes"] } then {
		set initialDir $pfmConfig(exampleDir)
		set title "Select an example database"
		set fromEncoding {iso8859-1}
		set fileTypes {
		    {{SQL statements} {.sql} }
		    {{All files} *}
		}
		set defaultExt ".sql"
		::report::cmdReportSQL {sql}
		set fileName [tk_getOpenFile -title $title -filetypes $fileTypes \
				  -defaultextension $defaultExt -parent .report \
				  -initialdir $initialDir]
		if {![string equal $fileName {}]} then {
		    set example [file tail $fileName]
		    .report.fmSQL.text delete 1.0 end
		    set convertedFile [::report::ConvertToUTF-8 $fileName {iso8859-1}]
		    if {$convertedFile ne {}} then {
			.report.fmSQL.text insert end "\\i '$convertedFile'"
			set msgInfo "The example database '${example}' will be installed."
			append msgInfo "\nPlease, check the output that will be generated."
			append msgInfo "\nThen close window \"Run SQL\"."
			tk_messageBox -message $msgInfo -type ok -icon info -parent .report
			::report::cmdRunSQL
			tkwait window .report
			refreshFormsList
			# disable install pfm_* tables and install example database
			.menubar.tools entryconfigure 0 -state disabled
			.menubar.tools entryconfigure 1 -state disabled
		    }
		}
	    }
	}
	return
    }

    proc check_pfm_tables {} {

	variable currentDB
	variable pfmVersion

	set queryDef {
	    SELECT COUNT(*) AS pfm_exists FROM pg_tables WHERE
	    tablename IN ('pfm_form','pfm_attribute','pfm_value','pfm_value_list',
			  'pfm_link','pfm_report','pfm_section')
	}
	set queryRes [pg_exec $currentDB $queryDef]
	set nrInstalled [lindex [pg_result $queryRes -getTuple 0] 0]
	pg_result $queryRes -clear
	if { $nrInstalled == 0} then {
	    # enable Install pfm_* tables and Install example database
	    .menubar.tools entryconfigure 0 -state normal
	    .menubar.tools entryconfigure 1 -state normal
	    set msgWarn \
		"The pfm_* tables have not been installed in this database. "
	    append msgWarn \
		"You are currently not able to use or define forms or reports."
	    append msgWarn \
		"\nHint: Use 'Tools -> Install pfm_* tables' or 'Tools -> "
	    append msgWarn \
		"Install example database ...'."
	    tk_messageBox -type ok -icon warning -message $msgWarn -parent .
	} else {
	    # disable Install pfm_* tables and Install example database
	    .menubar.tools entryconfigure 0 -state disabled
	    .menubar.tools entryconfigure 1 -state disabled	    
	    set queryDef {SELECT version FROM pfm_version ORDER BY seqnr DESC}
	    set queryRes [pg_exec $currentDB $queryDef]
	    set status [pg_result $queryRes -status]
	    if { [string equal $status "PGRES_TUPLES_OK"] } then {
		set numTuples [pg_result $queryRes -numTuples]
		if { $numTuples > 0 } then {
		    pg_result $queryRes -assign versionArray
		    set DBversion $versionArray(0,version)
		} else {
		    set msgWarn "pfm_version exists, but contains no data. "
		    append msgWarn "pfm allows you to open this database, "
		    append msgWarn "but it is an indication that the pfm_* tables "
		    append msgWarn "are not properly installed."
		    tk_messageBox -message $msgWarn -type ok -icon warning -parent .
		    # We assume that no conversion is needed.
		    set DBversion $pfmVersion
		}
	    } else {
		# Older versions of pfm did not record version in the database.
		set DBversion "1.0.4"
	    }
	    pg_result $queryRes -clear
	    switch -- [versionCompare $pfmVersion $DBversion] {
		1 {
		    # pfm version is newer than database version
		    if {![convertDB $DBversion]} then {
			set msgWarn "This version of the pfm_* tables (${DBversion}) may be "
			append msgWarn "incompatible with this version of pfm (${pfmVersion}). "
			append msgWarn "Continue at your own risk."
			tk_messageBox -type ok -icon warning -message $msgWarn -parent .
		    }
		}
		0 {
		    # pfm version = database version, no action required
		}
		-1 {
		    # pfm version is older than database version
		    set msgWarn "This version of the pfm_* tables (${DBversion}) may be "
		    append msgWarn "incompatible with this version of pfm (${pfmVersion}). "
		    append msgWarn "Continue at your own risk."
		    tk_messageBox -type ok -icon warning -message $msgWarn -parent .
		}
	    }
	}
	return
    }

    proc versionCompare {v1 v2} {

	set v1List [split $v1 "."]
	set v2List [split $v2 "."]
	set result 0
	for {set i 0} {($i <= 2) && ($result == 0)} {incr i} {
	    if {[lindex $v1List $i] < [lindex $v2List $i]} then {
		set result -1
	    } else {
		if {[lindex $v1List $i] > [lindex $v2List $i]} then {
		    set result 1
		}
	    }
	}
	return $result
    }

    proc convertDB {DBversion} {
       	variable installDir
	variable currentDB

	# From pfm version 1.3.0 on, we do not convert the database to the latest
	# version of pfm. We only convert it to 1.2.0. Future versions of pfm
	# will also convert the database to version 1.2.0 until there is a need
	# to change the structure of the pfm_tables.
	#
	# Similarly, the SQL-scripts install_pfm.sql and install_<exampledb>.sql,
	# of this and future versions of pfm will also create pfm_tables of version 1.2.0,
	# until there is a need to change the structure of the pfm_tables.

	set toVersion {1.2.0}
	switch $DBversion {
	    {1.0.4} {
		set proceed [tk_messageBox -message \
		       "Convert pfm-* tables from version $DBversion to version ${toVersion}" \
				 -type yesno -icon question -parent .]
		if { [string equal $proceed "yes"] } then {
		    set converted [specialConversion {1.0.4} {1.1.0}]
		    set converted [specialConversion {1.1.0} $toVersion]
		} else {
		    set converted 0
		}
	    }
	    {1.1.0} -
	    {1.1.1} {
		set proceed [tk_messageBox -message \
		       "Convert pfm-* tables from version $DBversion to version ${toVersion}?" \
				 -type yesno -icon question -parent .]
		if { [string equal $proceed "yes"] } then {
		    set converted [specialConversion {1.1.0} $toVersion]
		} else {
		    set converted 0
		}
	    }
	    default {
		# The database version is between 1.2.0 and the current pfm version.
		# Since the structure of the pfm_tables has not changed between
		# version 1.2.0 and the current pfm version, there is no need to touch the
		# pfm tables.
		set converted 1
	    }
	}
	return $converted
    }

    proc specialConversion {DBversion toVersion} {
       	variable installDir
	variable currentDB

	# When this procedure is called, the file 
	# '$installDir/convert_from_$DBversion.sql' should exist.

	::report::cmdReportSQL {sql}
	set fileName [file join $installDir convert_from_$DBversion.sql]
	if {![file exists $fileName]} then {
	    set errMsg "$fileName does not exist!"
	    tk_messageBox -message $errMsg -icon error -type ok -parent .report
	    set converted 0
	} else {
	    .report.fmSQL.text delete 1.0 end
	    set convertedFile [::report::ConvertToUTF-8 $fileName {iso8859-1}]
	    if {$convertedFile ne {}} then {
		.report.fmSQL.text insert end "\\i '$convertedFile'"
		set msgInfo "The database will be converted to $toVersion."
		append msgInfo "\nPlease, check the output that will be generated."
		append msgInfo "\nThen close window \"Run SQL\"."
		tk_messageBox -message $msgInfo -type ok -icon info -parent .report
		::report::cmdRunSQL
		tkwait window .report
		refreshFormsList
		set converted 1
	    } else {
		set converted 0
	    }
	}
	return $converted
    }
    

    proc cmdOpenQuery {} {
	variable formsIndex
	variable widget

	set formNr [$widget(lsbForms) curselection]
	if { $formNr == {} } then {
	    set formNr 0
	}
	if {[info exists formsIndex($formNr)]} then {
	    set formName $formsIndex($formNr)
	    ::form::OpenQuery $formName 1
	}
	return
    }

    proc defineEditMenu {} {
	
	menu .mnEdit -tearoff 0
	.mnEdit add command \
	    -command {::pfm::cmdEditCopy} -label Copy
	.mnEdit add command \
	    -command {::pfm::cmdEditCut} -label Cut
	.mnEdit add command \
	    -command {::pfm::cmdEditPaste} -label Paste
	# Entry and Text are class names. They refer to all entry and text widgets.
	bind Entry <ButtonPress-3> ::pfm::popUpEditMenu
	bind Text <ButtonPress-3> ::pfm::popUpEditMenu
    }

    proc popUpEditMenu {} {

	# Set focus to the widget on which button 3 (right button) is pressed.
	set x [winfo pointerx .]
	set y [winfo pointery .]
	set widget [winfo containing -displayof . $x $y]
	focus $widget
	if {[selection own -displayof . -selection PRIMARY] ne $widget ||
	    [catch {selection get -displayof . -selection PRIMARY \
			-type STRING}]} then {
	    # The widget does not own the selection or there is no selection
	    # in the widget
	    .mnEdit entryconfigure 0 -state disabled
	    .mnEdit entryconfigure 1 -state disabled
	} else {
	    .mnEdit entryconfigure 0 -state normal
	    if {[$widget cget -state] eq {normal}} then {
		.mnEdit entryconfigure 1 -state normal
	    } else {
		.mnEdit entryconfigure 1 -state disabled
	    }
	}
	if {[catch {clipboard get -displayof . -type STRING}]} then {
	    .mnEdit entryconfigure 2 -state disabled
	} else {
	    if {[$widget cget -state] eq {normal}} then {
		.mnEdit entryconfigure 2 -state normal
	    } else {
		.mnEdit entryconfigure 2 -state disabled
	    }
	}
	tk_popup .mnEdit $x $y
	return
    }

    proc cmdEditCopy {} {

	clipboard clear -displayof .
	clipboard append -displayof . -format STRING -type STRING -- \
	    [selection get -displayof . -selection PRIMARY -type STRING]
	selection clear -displayof . -selection PRIMARY
	return
    }
 
    proc cmdEditCut {} {

	# selection own -selection PRIMARY $widget
	clipboard clear -displayof .
	clipboard append -displayof . -format STRING -type STRING -- \
	    [selection get -displayof . -selection PRIMARY -type STRING]
	set widget [focus -displayof .]
	$widget delete sel.first sel.last
	selection clear -displayof . -selection PRIMARY
	return
    }
    
    proc cmdEditPaste {} {
	set widget [focus -displayof .]
	$widget insert insert [clipboard get -displayof . -type STRING]
	return
    }

    proc pfmDialog {widget title message msgWidth defaultButton buttonList} {
	variable ::pfm::dialogChoice

	set dialogChoice 0
	destroy $widget
	toplevel $widget -class Toplevel
	# set x [winfo pointerx [winfo parent $widget]]
	# set y [winfo pointery [winfo parent $widget]]
	wm transient $widget [winfo parent $widget]
	# wm geometry $widget +$x+$y
	wm title $widget $title
	message $widget.message -width $msgWidth -justify left \
	    -text $message
	set lastButton [expr [llength $buttonList] - 1]
	grid ${widget}.message -row 0 -column 0 -columnspan [expr $lastButton + 1]
	for {set button 0} {$button <= $lastButton} {incr button} {
	    set cmdButton "set ::pfm::dialogChoice ${button}\n"
	    append cmdButton "destroy $widget"
	    button ${widget}.button${button} \
		-text [lindex $buttonList $button] \
		-command $cmdButton
	    grid ${widget}.button${button} -row 1 -column $button
	}
	focus ${widget}.button${defaultButton}
	tkwait window $widget
	return $dialogChoice
    }

    proc PressButton {button} {

	if {([info command $button] ne {}) && \
		([$button cget -state] ne {disabled}) } then {
 	    if {[$button cget -relief] eq {raised} } then {
 		$button configure -relief sunken
 		set command "$button configure -relief raised"
 		append command "\n$button invoke"
 		after 100 $command
 	    } else {
 		after idle "$button invoke"
 	    }
	}
	return
    }



    #########################################################
    # Main of namespace pfm                                 #
    #########################################################

    if { $argc == 0 } then {
	set installDir [file normalize [pwd]]
    } else {
	set installDir [lindex $argv 0]
    }
    if {[string equal $tcl_platform(platform) "unix"]} then {
	set configFile [file join $installDir {pfm.conf}]
	if {![file exists $configFile]} then {
	    set configFile {/etc/pfm.conf}
	}
	if {[catch {open $configFile r} confChan]} then {
	    tk_messageBox -type ok -icon error -message $confChan
	    exit
	} else {
	    while {![eof $confChan]} {
		set line [gets $confChan]
		if {([string length $line] != 0) && \
			([string compare -length 1 $line "#"] != 0)} then {
		    set pfmConfig([lindex $line 0]) [lindex $line 1]
		}
	    }
	    close $confChan
	    if {[string compare -length 1 $pfmConfig(docDir) "/"] != 0} then {
		# The directory indicated is relative to the installDir
		set pfmConfig(docDir) [file join $installDir $pfmConfig(docDir)]
	    }
	    if {[string compare -length 1 $pfmConfig(helpDir) "/"] != 0} then {
		# The directory indicated is relative to the installDir
		set pfmConfig(helpDir) [file join $installDir $pfmConfig(helpDir)]
	    }
	    if {[string compare -length 1 $pfmConfig(exampleDir) "/"] != 0} then {
		# The directory indicated is relative to the installDir
		set pfmConfig(exampleDir) [file join $installDir $pfmConfig(exampleDir)]
	    }
	}
    } else {
	# On non-UNIX platforms there is no configuration file.
	set pfmConfig(docDir) [file join $installDir {doc}]
	set pfmConfig(helpDir) [file join $installDir {help}]
	set pfmConfig(exampleDir) [file join $installDir {examples}]
	# The default values for "browser" and "printcmd" are not read
	# from pfmConfig. They are calculated in the procedure getDefault.
    }
    ::options::initOptions
    initRootWindow
    defineEditMenu


    # Init errChannel. This the channel to which stderr of psql is redirected.

    switch -- $tcl_platform(platform) {
	"unix" {
	    # We use UNIX's cat instead of cat.tcl
	    set openCommand "| cat"
	}
	"windows" {
	    # Cat does not require Tk. Hence it is better to use tclkitsh
	    # for faster start up.
	    set tclkit [file join $installDir tclkit tclkitsh.exe]
	    set cat [file join $installDir "cat.kit"]
	    set openCommand "|"
	    lappend openCommand $tclkit
	    lappend openCommand $cat
	}
    }
    if {[catch {open $openCommand RDWR} errChannel]} then {
	tk_messageBox -type ok -icon error -message $errChannel
    } else {
	fconfigure $errChannel -encoding utf-8
	fileevent $errChannel readable ::report::showError
    }

}

#############################################################################
# end of namespace pfm                                                      #
#############################################################################

#############################################################################
# begin of namespace help                                                   #
#############################################################################
#

namespace eval help {

    proc cmdDisplayManual {} {

	variable ::pfm::pfmConfig
	variable ::options::pfmOptions
	
	set url "file://${pfmConfig(docDir)}/index.html"
	set command {exec}
	set map {%s}
	lappend map $url
	foreach arg $pfmOptions(browser) {
		lappend command [string map $map $arg]
	}
	    lappend command {&}
	# puts $command
	if { [catch $command errMsg]} then {
	    tk_messageBox -type ok -icon error -parent . \
		-message "$command failed\n$errMsg"
	}
	return
    }


    proc cmdLicense {} {

	variable ::pfm::pfmConfig

	set fileName [file join $pfmConfig(helpDir) {gpl.txt}]
	if { [catch {open $fileName r} license_ch] } then {
	    fileNotFound {license} $license_ch
	} else {
	    toplevel .license -class Toplevel
	    wm geometry .license 600x600
	    wm title .license "pfm - License"
	    label .license.lbl \
		-text {GNU GENERAL PUBLIC LICENSE - Version 2, June 1991}
	    text .license.txt -background white \
		-yscrollcommand {.license.vsb set} -takefocus 0
	    scrollbar .license.vsb -orient vertical \
		-command {.license.txt yview} -takefocus 0
	    button .license.bnOK -command {destroy .license} -text {OK} \
	        -underline 0
	    focus .license.bnOK
	    bind .license <Alt-KeyPress-o> {::pfm::PressButton .license.bnOK}
	    bind .license <KeyPress-Escape> {::pfm::PressButton .license.bnOK}
	    bind .license <KeyPress-Prior> {
		.license.txt yview scroll -1 pages
	    }
	    bind .license <KeyPress-Next> {
		.license.txt yview scroll 1 pages
	    }
	    bind .license <KeyPress-Up> {
		.license.txt yview scroll -1 units
	    }
	    bind .license <KeyPress-Down> {
		.license.txt yview scroll 1 units
	    }
	    place .license.lbl \
		-anchor n -x 0 -y 0 -relx 0.5 
	    place .license.txt \
		-anchor nw -x 0 -y 20 \
		-width -20 -relwidth 1 \
		-relheight 1 -height -50
	    place .license.vsb \
		-anchor ne -x 0 -y 20 -relx 1 \
		-width 20 \
		-height -50 -relheight 1
	    place .license.bnOK \
		-anchor s -x 0 -y 0 -relx 0.5 -rely 1
	    set license_text [read $license_ch]
	    close $license_ch
	    .license.txt insert end $license_text
	}
	return
    }

    proc fileNotFound {what reason} {

	switch $what {
	    help {
		set msgText "Normally you should see the help file now, but $reason"
	    }
	    license {
		set msgText "Normally you should see the GNU General Public License now, but $reason"
	    }
	    default {
		set msgText $reason
	    }
	}
	tk_messageBox -message $msgText -type ok -icon error
	return
    }

}
######################################################################
# End of namespace help                                              #
######################################################################

######################################################################
# Begin of namespace form                                            #
######################################################################

# activeForm contains the name of the currently opened form.
# It is filled by several procedures:
#     1. OpenQuery
#     2. cmdFollowLink
#     3. cmdBack
#
# formAttribList contains the list of attributes of the active form,
# as defined in data base table pfm_attributes.
#
# attributeArray contains the complete attributes defintion of the
# active form as defined by pfm_attributes.
#
# tableAttribList contains only the attributes that are not "tgReadOnly".
# The purpose is to include only the attributes of the table referred
# to by pfm_form.tablename.
#
# formAttribList, tableAttribList and attributeArray are filled by 
# proc getAttributes.
#
# getAttributes is called by:
#     1. OpenQuery
#     2. cmdFollowLink
#     3. cmdBack
#
# WhereSelected is a boolean which indicates whether the user is
# pasting into the "where" or into the "order by" text entry
# of the query. Its value is controlled by the radio buttons
# on the query window.
#
# recordArray contains all the records selected by the query defined
# by the user in the query window. recordArray($tupleNr,$attribute)
# indicates the value of $attribute of $tupleNr. It is filled
# by proc OpenForm. It is the so called "internal buffer".
# On top of the attribute values, recordArray also contains
# a status for each record: recordArray($tupleNr,23status47)
# contains the status the record indicated by $tupleNr. The
# status can be : "Not modified", "Updated", "Deleted", "Added",
# "After last", "Not added", "Updating", "Adding". 
# The attribute name "23status47" has been chosen
# to avoid name conflicts with real table attributes.
#
# lastRecord is the tupleNr of the last tuple in recordArray. This
# is in fact a dummy, empty record, functioning as a sentinel.
#
# curRecord is the tupleNr of the record that is currently displayed
# on the screen.
#
# txtRecord is the textvar linked to the attribute labels,
# entries or buttons on the form. When the current record
# is displayed, the values of recordArray($curRecord,$attribute)
# are copied to txtRecord($attribute).
#
# formStack contains the subsequent queries issued by the user
# as a result of following links. lastFormOnStack is a stack
# pointer on this stack. The first query is pushed on the stack
# by proc cmdExecuteQuery. If the user clicks on a link button
# another query is pushed on the stack by proc cmdFollowLink.
# Any time the user presses back, a query defintion is popped
# of the stack. The elements that are kept on the stack are:
#     formId
#     queryDef
#     intro (the information displayed on top of the form)
#     displayKey (the list of pkey attribute values of the record 
#                 that was displayed at
#                 the time a link button was pressed and which
#                 is displayed again when the user presses "Back")
#
# linksArray is loaded with all the links originating from the
# active form. It is filled by proc displayLinks, which is called
# from OpenForm. Its structure is linksArray($link,$attribute)
# where $link is an index for the link (starting from 0) and
# where $attribute is any attribute of pfm_link.
#
# widget is an array containing aliases for some widget
# path names
#
# windowSize is and associative array in which the window size
# and window position of the form windows are stored, such that
# when the user resizes and replaces a window, pfm remembers size
# and position until pfm is closed.
#
# focusRow is de attribute row that has the focus on the form window.


namespace eval form {
    variable activeForm
    variable formState
    variable formAttribList
    variable tableAttribList
    variable attributeArray
    variable WhereSelected
    variable queryWhere
    variable queryOrderBy
    variable lastRecord
    variable recordArray
    variable curRecord
    variable txtRecord
    variable formStack
    variable lastFormOnStack
    variable linksArray
    variable widget
    variable windowSize
    variable focusRow

    #############################################################
    # Procedures that are called from the corresponding cmdXXX  #
    # procedures in namespace pfm                               #    
    #############################################################


    proc OpenQuery {formName clear} {

	variable ::pfm::formsArray
	variable activeForm
	variable WhereSelected
	variable queryWhere
	variable queryOrderBy
	variable formAttribList
	variable attributeArray
	variable widget

	set WhereSelected true
	if {$clear} then {
	    set queryWhere {}
	    set queryOrderBy {}
	}
	set activeForm $formName
	showQueryWindow
	wm title .query "pfm - Open form : $formName"
	$widget(lbSelect) insert end \
	    "SELECT $formsArray($activeForm,sqlselect)\nFROM $formsArray($activeForm,sqlfrom)"
	if { ![string equal $formsArray($activeForm,groupby) {}] } then {
	    $widget(lbSelect) insert end \
		"\nGROUP BY $formsArray($activeForm,groupby)"
	}
	getAttributes $activeForm
	fillPasteMenus
	return
    }

    proc fillPasteMenus {} {
	variable ::pfm::formsArray
	variable formAttribList
	variable attributeArray
	variable widget

	set names 0
	set values 0
	foreach attribute $formAttribList {
	    .query.menubar.mnNames add command \
		-label $attribute \
		-command "::form::cmdPasteAttribute $attribute"
	    set names 1
	    switch $attributeArray($attribute,typeofget) {
		"tgDirect" -
		"tgReadOnly" -
		"tgExpression" { }
		"tgList" -
		"tgLink" {
		    .query.menubar.mnValues add command \
			-label $attribute \
			-command \
			"::form::cmdSelectFromList $attribute .query.entWhere paste"
		    set values 1
		}
	    }
	}
	if {$names} then {
	    .query.menubar entryconfigure 1 -state normal
	}
	if {$values} then {
	    .query.menubar entryconfigure 2 -state normal
	}
	return
    }


    ################################################################
    #                                                              #
    # Procedures for query window (now renamed to "Open form")     #
    #                                                              #
    ################################################################

    proc showQueryWindow { } {

	variable widget
	variable windowSize

	destroy .query
	destroy .form
	toplevel .query -class Toplevel
	bind .query <Destroy> {
	    set ::form::windowSize(.query) \
		[string map {{+0+0} {}} [wm geometry .query]]
	}
	wm geometry .query $windowSize(.query)


	# Text for sql statement
	text .query.lbSelect -wrap word -background white \
	    -yscrollcommand {.query.vsbSelect set} -takefocus 0
	scrollbar .query.vsbSelect -orient vertical \
	    -command {.query.lbSelect yview} -takefocus 0
	set widget(lbSelect) .query.lbSelect
	place .query.lbSelect \
	    -anchor nw -x 0 -y 0 \
	    -relwidth 1 -width -20 \
	    -height -74 -relheight 1
	place .query.vsbSelect \
	    -anchor ne -x 0 -y 0 -relx 1 \
	    -width 20 \
	    -height -74 -relheight 1

	# radio buttons and entries
	radiobutton .query.rabWhere \
	    -text WHERE -value true  -variable ::form::WhereSelected \
	    -underline 0 -takefocus 0 -command {focus .query.entWhere}
	entry .query.entWhere -background white \
	    -textvariable ::form::queryWhere
	radiobutton .query.rabOrderBy \
	    -text {ORDER BY} -value false  -variable ::form::WhereSelected \
	    -underline 0 -takefocus 0 -command {focus .query.entOrderBy}
	entry .query.entOrderBy -background white \
	    -textvariable ::form::queryOrderBy
	place .query.rabWhere \
	    -anchor sw -x 0 -y -52 -rely 1 \
	    -height 22
	place .query.rabOrderBy \
	    -anchor sw -x 0 -y -30 -rely 1 \
	    -height 22
	set buttonWidth [winfo reqwidth .query.rabOrderBy]
	place .query.entWhere \
	    -anchor sw -x $buttonWidth -y -52 -rely 1 \
	    -width -$buttonWidth -relwidth 1 \
	    -height 22 
	place .query.entOrderBy \
	    -anchor sw -x $buttonWidth -y -30 -rely 1 \
	    -width -$buttonWidth -relwidth 1 \
	    -height 22



	# Frame for query buttons
	frame .query.buttons -borderwidth 2 -relief groove
	place .query.buttons \
	    -anchor sw -x 0 -y 0 -rely 1 \
	    -relwidth 1 \
	    -height 30

	# Menubar

	menu .query.menubar -tearoff 0
	.query configure -menu .query.menubar
	.query.menubar add cascade \
	    -menu .query.menubar.mnForm \
	    -label {Form} -underline 0 \
	    -state normal
	.query.menubar add cascade \
	    -menu .query.menubar.mnNames \
	    -label {paste attribute Name} -underline 16 \
	    -state disabled
	.query.menubar add cascade \
	    -menu .query.menubar.mnValues \
	    -label {paste attribute Value} -underline 16 \
	    -state disabled

	# The From menu
	menu .query.menubar.mnForm -tearoff 0
	.query.menubar.mnForm add command \
	    -label {Close} -underline 0 -accelerator {Esc} \
	    -command ::form::cmdQuitQuery

	# The attribute Names menu
	menu .query.menubar.mnNames -tearoff 0
	# This menu is filled by fillPasteMenus

	# The attribute Values menu
	menu .query.menubar.mnValues -tearoff 0
	# This menu is filled by fillPasteMenus

	# The query buttons
	button .query.buttons.execute -command \
	    {::form::cmdExecuteQuery $::form::queryWhere $::form::queryOrderBy \
		 "User defined query:"} -text Run -underline 0
	button .query.buttons.quit -command ::form::cmdQuitQuery -text Quit \
	    -underline 0
	pack .query.buttons.quit -side right
	pack .query.buttons.execute -side right

	focus .query.entWhere
	bind .query <Alt-KeyPress-w> {::pfm::PressButton .query.rabWhere}
	bind .query <Alt-KeyPress-o> {::pfm::PressButton .query.rabOrderBy}
	bind .query <Alt-KeyPress-r> \
	    {::pfm::PressButton .query.buttons.execute}
# 	bind .query <Alt-KeyPress-Return> \
# 	    {::pfm::PressButton .query.buttons.execute}
	bind .query <KeyPress-Escape> \
	    {::pfm::PressButton .query.buttons.quit}
	bind .query <Alt-KeyPress-q> \
	    {::pfm::PressButton .query.buttons.quit}
	bind .query.entWhere <FocusIn> {set ::form::WhereSelected true}
	bind .query.entOrderBy <FocusIn> {set ::form::WhereSelected false}

	return
    }


    proc cmdExecuteQuery {sqlWhere sqlOrderBy intro} {
	# This procedure prepares the SQL SELECT statement of the
	# query to be made, it initialises the formStack and
	# it calls OpenForm. It is called when the user presses
	# the Execute button on the query window.

	variable ::pfm::formsArray
	variable activeForm
	variable formStack
	variable lastFormOnStack
	variable windowSize

	set sqlAttrib $formsArray($activeForm,sqlselect)
	if { !$formsArray($activeForm,view) && \
		 ([lsearch $formsArray($activeForm,pkey) {oid}] >= 0) && \
		 ([regexp {\moid\M} $sqlAttrib] == 0)} then {
	    set sqlAttrib \
		"\"$formsArray($activeForm,tablename)\".oid, $sqlAttrib"
	}
	set sqlFrom $formsArray($activeForm,sqlfrom)
	set groupby $formsArray($activeForm,groupby)
	set queryDef "SELECT $sqlAttrib\nFROM $sqlFrom"
	if { ![string equal $groupby {}] } then {
	    append queryDef "\nGROUP BY $groupby"
	    if { ![string equal $sqlWhere {}] } then {
		# If there is a GROUP BY clause, the sqlWhere must become
		# a HAVING clause.
		append queryDef "\nHAVING $sqlWhere"
	    }
	} else {
	    if { ![string equal $sqlWhere {}] } then {
		append queryDef "\nWHERE $sqlWhere"
	    }
	}
	if { ![string equal $sqlOrderBy {}] } then {
	    append queryDef "\nORDER BY $sqlOrderBy"
	}
	array unset formStack
	set lastFormOnStack 0
	set formStack($lastFormOnStack,formId) $activeForm
	set formStack($lastFormOnStack,queryDef) $queryDef
	set formStack($lastFormOnStack,intro) $intro
	set formStack($lastFormOnStack,displayKey) {}
	destroy .query
	OpenForm $queryDef $intro {}
	return
    }

    proc cmdPasteAttribute {attribute} {
	# The user pastes an attribute name into either the "where" or
	# "order by" text entries, depending on WhereSelect, the value
	# of which is determined by the radio buttons .query.rabWhere
	# and .query.rabOrderBy

	variable WhereSelected
	if {$WhereSelected} then {
	    .query.entWhere insert insert \"$attribute\"
	} else {
	    .query.entOrderBy insert insert \"$attribute\"
	}
	return
    }


    proc cmdQuitQuery {} {
	variable windowSize
	
	destroy .query
	::pfm::refreshFormsList
	return
    }



    ################################################################
    #                                                              #
    # Procedures for form window                                   #
    #                                                              #
    ################################################################

    proc showFormWindow { } {

	variable widget
	variable windowSize

	destroy .query
	if { [winfo exists .form] } then {
	    displayRecordFrame
 	    destroy .form.frmLinkBtn
 	    frame .form.frmLinkBtn -borderwidth 2 -relief sunken	
 	    place .form.frmLinkBtn -in .form.pLink \
 		-anchor nw -x 0 -y 30 \
 		-relwidth 1 \
 		-relheight 1 -height -30
	    .form.txtQuery delete 1.0 end
	    .form.txtResult delete 1.0 end
	} else {
	    toplevel .form -class Toplevel
	    bind .form <Destroy> {
		set ::form::windowSize(.form) \
		    [string map {{+0+0} {}} [wm geometry .form]]
	    }
	    wm geometry .form $windowSize(.form)

	    # Main menu
	    menu .form.menu -tearoff 0
	    .form.menu add cascade -menu .form.menu.mnForm -label {Form} \
		-underline 0
	    .form.menu add cascade -menu .form.menu.mnRecord -label {Record} \
		-underline 0
	    .form.menu add cascade -menu .form.menu.mnGoto \
		-label {Go to} -underline 0
	    .form.menu add cascade -menu .form.menu.mnSearch \
		-label {Search} -underline 0
	    .form.menu add command -label {Help} -underline 0 \
		-command ::form::cmdHelp
	    bind .form <KeyPress-F1> ::form::cmdHelp

	    # Make it the main menu
	    .form configure -menu .form.menu

	    # Menu Form
	    menu .form.menu.mnForm -tearoff 0
	    .form.menu.mnForm add command \
		-command {::form::cmdQuitForm} \
		-label {Close} -underline 0 -accelerator {Esc}

	    # Menu Record
	    menu .form.menu.mnRecord -tearoff 0
	    .form.menu.mnRecord add command \
		-command {::form::cmdAdd} \
		-label {Add} -underline 0 -accelerator {Alt-a}
	    .form.menu.mnRecord add command \
		-command {::form::cmdDelete} \
		-label {Delete} -underline 0 -accelerator {Alt-d}
	    .form.menu.mnRecord add command \
		-command {::form::cmdUpdate} \
		-label {Update} -underline 0 -accelerator {Alt-u}

	    # Menu Go to
	    menu .form.menu.mnGoto -tearoff 0
	    .form.menu.mnGoto add command \
		-command {::form::cmdFirst} \
		-label {First} -underline 0 -accelerator {Home, Cntrl-Home}
	    .form.menu.mnGoto add command \
		-command {::form::cmdPrev} \
		-label {Previous} -underline 0 -accelerator {Page Up, Left}
	    .form.menu.mnGoto add command \
		-command {::form::cmdNext} \
		-label {Next} -underline 0 -accelerator {Page Down, Right}
	    .form.menu.mnGoto add command \
		-command {::form::cmdLast} \
		-label {Last} -underline 0 -accelerator {End, Cntrl-End}

	    # Menu Search is defined in displayAttribNames

	    # Bind up and down arrows to move input focus to
	    # higher or lower attribute on the form.

	    bind .form <KeyPress-Down> \
		{focus [tk_focusNext [focus -displayof .form]]}
	    bind .form <KeyPress-Up> \
		{focus [tk_focusPrev [focus -displayof .form]]}

	    # Define and place panes

	    frame .form.pQuery
	    frame .form.pForm
	    frame .form.pLink
	    frame .form.pResult

 	    place .form.pQuery -in .form \
 		-anchor nw -x 0 -y 0 \
 		-relwidth 1 \
 		-relheight 0.2
 	    place .form.pForm -in .form \
 		-anchor nw -x 0 -y 0 -rely 0.2 \
 		-relwidth 0.8 \
 		-relheight 0.6
 	    place .form.pLink -in .form \
 		-anchor ne -x 0 -y 0 -relx 1 -rely 0.2 \
 		-relwidth 0.2 \
 		-relheight 0.6
 	    place .form.pResult -in .form \
 		-anchor sw -x 0 -rely 1 \
 		-relwidth 1 \
 		-relheight 0.2

	    # the contents of pQuery pane

	    text .form.txtQuery -background white \
		-wrap word -yscrollcommand {.form.scrollQuery set} \
		-takefocus 0
	    .form.txtQuery tag configure blueTag -foreground {medium blue}
	    .form.txtQuery tag configure greenTag -foreground {green4}
	    .form.txtQuery tag configure redTag -foreground {red3}
	    scrollbar .form.scrollQuery -orient vertical \
		-command {.form.txtQuery yview} -takefocus 0
	    place .form.txtQuery -in .form.pQuery \
		-anchor nw -x 0 -y 0 \
		-relwidth 1 -width -20 \
		-relheight 1
	    place .form.scrollQuery -in .form.pQuery \
		-anchor ne -relx 1 -y 0 \
		-width 20 \
		-relheight 1

	    # text area for Result

	    text .form.txtResult -background white \
		-wrap word -yscrollcommand {.form.scrollResult set} \
		-takefocus 0
	    .form.txtResult tag configure blueTag -foreground {medium blue}
	    .form.txtResult tag configure greenTag -foreground {green4}
	    .form.txtResult tag configure redTag -foreground {red3}
	    scrollbar .form.scrollResult -orient vertical \
		-command {.form.txtResult yview} -takefocus 0
	    place .form.txtResult -in .form.pResult \
		-anchor nw -x 0 -y 0 \
		-relwidth 1 -width -20 \
		-relheight 1
	    place .form.scrollResult -in .form.pResult \
		-anchor ne -relx 1 -y 0 \
		-width 20 \
		-relheight 1 


	    # The form body

	    # The status bar

	    frame .form.frmStatus -borderwidth 2 -relief groove
	    place .form.frmStatus -in .form.pForm \
		-anchor nw -x 0 -y 0 \
		-relwidth 1 \
		-height 30
 	    label .form.frmStatus.lblRecord \
 		    -borderwidth 1 -textvar ::form::txtRecord(23nr47) 
	    label .form.frmStatus.lblFormName \
		    -borderwidth 1 -textvar ::form::activeForm
	    label .form.frmStatus.lblStatus \
		    -borderwidth 1 -textvar ::form::txtRecord(23status47)
	    place .form.frmStatus.lblRecord -in .form.frmStatus \
		    -anchor w -x 10 -rely 0.5
	    place .form.frmStatus.lblFormName -in .form.frmStatus \
		    -anchor center -rely 0.5 -relx 0.5
	    place .form.frmStatus.lblStatus -in .form.frmStatus \
		    -anchor e -relx 1 -x -10 -rely 0.5 

	    # The record frame

	    displayRecordFrame

	    # The buttons bar

	    frame .form.frmButtons -borderwidth 2 -relief groove
	    place .form.frmButtons -in .form.pForm \
		-anchor sw -x 0 -rely 1 \
		-relwidth 1 \
		-height 30

	    # The buttons are defined but not displayed yet. That is
	    # up to the procedures displayBrowseButtons and
	    # displayEditButtons

	    label .form.frmButtons.left
	    button .form.frmButtons.btnUpdate \
		-text Update -command ::form::cmdUpdate -takefocus 0 \
		-underline 0
	    button .form.frmButtons.btnAdd \
		-text Add -command ::form::cmdAdd -takefocus 0 \
		-underline 0
	    button .form.frmButtons.btnDelete \
		-text Delete -command ::form::cmdDelete -takefocus 0 \
		-underline 0
	    button .form.frmButtons.btnOK \
		-text OK -command ::form::cmdOK -underline 0
	    button .form.frmButtons.btnCancel \
		-text Cancel -command ::form::cmdCancel -underline 0
	    label .form.frmButtons.right

	    # The link frame

	    # The title bar

	    frame .form.frmLink1 -borderwidth 2 -relief groove
	    place .form.frmLink1 -in .form.pLink \
		-anchor nw -x 0 -y 0 \
		-relwidth 1 \
		-height 30
	    label .form.frmLink1.lblLinks \
		-borderwidth 1 -text Links
	    place .form.frmLink1.lblLinks -in .form.frmLink1 \
		-anchor center -x 0 -relx 0.5 -y 0 -rely 0.5 

	    # The body of the links frame

	    frame .form.frmLinkBtn -borderwidth 2 -relief sunken
	    place .form.frmLinkBtn -in .form.pLink \
		-anchor nw -x 0 -y 30 \
		-relwidth 1 \
		-relheight 1 -height -30

	    # The links are displayed by displayLinks procedure
	}
	return
    }

    proc displayRecordFrame {} {
	variable widget

	destroy .form.canvas
	destroy .form.vscroll
	canvas .form.canvas -yscrollcommand {.form.vscroll set} \
	    -borderwidth 2 -relief sunken
	scrollbar .form.vscroll -orient vertical \
	    -command {.form.canvas yview} -takefocus 0
	frame .form.canvas.frmRecord -borderwidth 2 -relief raised
	.form.canvas create window 20 20 \
	    -window .form.canvas.frmRecord -anchor nw
	set widget(record) .form.canvas.frmRecord
	place .form.canvas -in .form.pForm \
	    -anchor nw -x 0 -y 30 \
	    -relwidth 1 -width -20 \
	    -relheight 1 -height -60
	place .form.vscroll -in .form.pForm \
	    -anchor nw -relx 1 -x -20 -y 30 \
	    -width 20 \
	    -relheight 1 -height -60
	return
    }

    ################################################################
    # Commands for browse, update, add and quit buttons            #
    ################################################################

    proc identCurRecord {withTable} {
	variable ::pfm::formsArray
	variable activeForm
	variable recordArray
	variable curRecord
	variable attributeArray

	set whereClause {}
	foreach pkey $formsArray($activeForm,pkey) {
	    if {[info exists recordArray($curRecord,$pkey)]} then {
		if {$whereClause ne {}} then {
		    set whereClause "$whereClause AND"
		}
		if {$withTable} then {
		    set attribute "\"$formsArray($activeForm,tablename)\".\"$pkey\""
		} else {
		    set attribute "\"$pkey\""
		}
		if {[info exists attributeArray($pkey,typeofattrib)] } then {
		    if {$attributeArray($pkey,typeofattrib) eq {taQuoted}} then {
			set value [string map {' ''} $recordArray($curRecord,$pkey)]
			set whereClause \
			    "$whereClause ($attribute = '$value')"
		    } else {
			set whereClause \
			    "$whereClause ($attribute = $recordArray($curRecord,$pkey))"
		    }
		} else {
		    set whereClause \
			"$whereClause ($attribute = $recordArray($curRecord,$pkey))"
		}
	    } else {
		set whereClause FALSE
		set errMsg "The pkey attribute '$pkey' of form '$activeForm' is not returned by the form's SQL statement. Check the form's definition!"
		tk_messageBox -type ok -icon error -message $errMsg -parent .form
		break
	    }
	}
	return $whereClause
    }

    proc cmdHelp {} {

	variable wrapOn
	variable ::pfm::currentDB
	variable activeForm
	variable windowSize

	set queryDef "SELECT help FROM pfm_form WHERE name='$activeForm'"
	set queryRes [pg_exec $currentDB $queryDef]
	set helpText [lindex [pg_result $queryRes -getTuple 0] 0]
	if { [string length $helpText] == 0 } then {
	    set helpText "No help available for $activeForm."
	}
	pg_result $queryRes -clear
	set wrapOn 1
	destroy .form.help
	toplevel .form.help -class Toplevel
	bind .form.help <Destroy> {
	    set ::form::windowSize(.form.help) \
		[string map {{+0+0} {}} [wm geometry .form.help]]
	}
	wm transient .form.help .form
	wm geometry .form.help $windowSize(.form.help)
	wm title .form.help "Help for $activeForm"

	# text widget
	text .form.help.text -wrap word -background white \
	    -yscrollcommand {.form.help.vscroll set} \
	    -xscrollcommand {.form.help.hscroll set} \
	    -takefocus 0
	scrollbar .form.help.vscroll -orient vertical \
	    -command {.form.help.text yview} -takefocus 0
	scrollbar .form.help.hscroll -orient horizontal \
	    -command {.form.help.text xview} -takefocus 0
	place .form.help.text \
	    -anchor nw -x 0 -y 0 \
	    -relwidth 1 -width -20 \
	    -relheight 1 -height -50
	place .form.help.vscroll \
	    -anchor ne -x 0 -y 0 -relx 1 \
	    -width 20 \
	    -height -50 -relheight 1
	place .form.help.hscroll \
	    -anchor sw -x 0 -y -30 \
	    -width -20 -relwidth 1 \
	    -rely 1 -height 20

	# Frame for help buttons
	frame .form.help.buttons
	place .form.help.buttons \
	    -anchor sw -x 0 -y 0 -rely 1 \
	    -relwidth 1 \
	    -height 30

	# The buttons
	radiobutton .form.help.buttons.rbWrap -text {Wrap} -value 1 \
	    -variable ::form::wrapOn \
	    -command {.form.help.text configure -wrap word} -takefocus 0 \
	    -underline 0
	radiobutton .form.help.buttons.rbTruncate -text {Truncate} \
	    -value 0 -variable ::form::wrapOn \
	    -command {.form.help.text configure -wrap none} -takefocus 0 \
	    -underline 0
	button .form.help.buttons.bnCancel -text {Cancel} \
	    -command {destroy .form.help}
	pack .form.help.buttons.bnCancel -side right
	pack .form.help.buttons.rbWrap -side left
	pack .form.help.buttons.rbTruncate -side left

	focus .form.help.buttons.bnCancel
	bind .form.help <KeyPress-Escape> \
	    "::pfm::PressButton .form.help.buttons.bnCancel"
	bind .form.help <Alt-KeyPress-c> \
	    "::pfm::PressButton .form.help.buttons.bnCancel"
	bind .form.help <Alt-KeyPress-w> \
	    "::pfm::PressButton .form.help.buttons.rbWrap"
	bind .form.help <Alt-KeyPress-t> \
	    "::pfm::PressButton .form.help.buttons.rbTruncate"
	bind .form.help <KeyPress-Prior> {
	    .form.help.text yview scroll -1 pages
	}
	bind .form.help <KeyPress-Next> {
	    .form.help.text yview scroll 1 pages
	}
	bind .form.help <KeyPress-Up> {
	    .form.help.text yview scroll -1 units
	}
	bind .form.help <KeyPress-Down> {
	    .form.help.text yview scroll 1 units
	}
	.form.help.text insert end $helpText
	.form.help.text configure -state disabled
	return
    }

    proc cmdAdd {} {
	variable attributeArray
	variable tableAttribList
	variable txtRecord
	variable ::pfm::currentDB
	variable activeForm
	variable windowSize

	if {[winfo exists .form.search]} then {
	    destroy .form.search
	}
	.form.txtResult delete 1.0 end
	foreach attribute $tableAttribList {
	    set defVal $attributeArray($attribute,default)
	    if {$defVal ne {}} then {
		if { [string index $defVal 0] eq {=}} then {
		    set queryDef [string range $defVal 1 end]
		    set queryRes [pg_exec $currentDB $queryDef]
		    if { [pg_result $queryRes -status] eq {PGRES_TUPLES_OK}} then {
			if {[pg_result $queryRes -numTuples] == 1} then {
			    set txtRecord($attribute) \
				[lindex [pg_result $queryRes -getTuple 0] 0]
			} else {
			    set errMsg "$queryDef\nhas returned [pg_result $queryRes -numTuples] records."
			    set errMsg "$errMsg\nCheck the definition of the default value for $attribute of $activeForm in pfm_attribute.\n"
			    bell
			    displayOnForm .form.txtResult $errMsg {red}
			}
		    } else {
			set errMsg "$queryDef\n[pg_result $queryRes -error]"
			set errMsg "$errMsg\nCheck the definition of the default value for $attribute of $activeForm in pfm_attribute.\n"
			bell
			displayOnForm .form.txtResult $errMsg {red}
		    }
		    pg_result $queryRes -clear
		} else {
		    set txtRecord($attribute) $defVal
		}
	    }
	}
	newFormState add
	return
    }

    proc cmdUpdate {} {

	# variable ::pfm::currentDB
	# variable ::pfm::formsArray
	variable txtRecord
	variable windowSize
	# variable recordArray
	# variable curRecord
	# variable tableAttribList
	# variable activeForm
	# Bug 679 : The start of the transaction is postponed until
	# the user presses [OK]. Instead a 'reload record' is executed
	# to minimize the time window during which another user can
	# modify or delete the current record.

	if { ![string equal $txtRecord(23status47) {After last}] && \
		![string equal $txtRecord(23status47) {Deleted}] && \
		![string equal $txtRecord(23status47) {Not added}] } then {
	    .form.txtResult delete 1.0 end
	    if {[winfo exists .form.search]} then {
		destroy .form.search
	    }
	    if { [reloadRecord] } then {
		newFormState update   
	    }	
	}
	return
    }


    proc cmdDelete {} {

	variable ::pfm::currentDB
	variable ::pfm::formsArray
	variable curRecord
	variable activeForm
	variable txtRecord
	variable recordArray
	variable formAttribList
	if { ($txtRecord(23status47) ne {After last}) && \
		 ($txtRecord(23status47) ne {Deleted}) && \
		 ($txtRecord(23status47) ne {Not added}) } then {
	    set question "Do you really want to delete the current record?"
	    set answer [tk_messageBox -type yesno -default no -icon question \
			    -message $question -parent .form]
	    if {$answer eq {yes}} then {
		set queryDef "DELETE FROM \"$formsArray($activeForm,tablename)\""
		set queryDef "$queryDef WHERE [identCurRecord 0]"
		.form.txtResult delete 1.0 end
		displayOnForm .form.txtResult "$queryDef\n" {black}
		set queryRes [pg_exec $currentDB $queryDef]
		set status [pg_result $queryRes -status]
		if { [string equal $status {PGRES_COMMAND_OK}] } then {
		    set recordArray($curRecord,23status47) "Deleted"
		    set txtRecord(23status47) "Deleted"
		    foreach attribute $formAttribList {
			set recordArray($curRecord,$attribute) ""
			set txtRecord($attribute) ""
		    }
		    displayOnForm .form.txtResult $status {green}
		} else {
		    set status $status\n[pg_result $queryRes -error]
		    bell
		    displayOnForm .form.txtResult $status {red}
		}
		pg_result $queryRes -clear
	    }
	}
	return
    }

    proc cmdFirst {} {
	variable curRecord

	if {$curRecord != 0} {
	    set curRecord 0
	    filltxtRecord $curRecord
	    .form.txtResult delete 1.0 end

	}
	return
    }

    proc cmdNext {} {

	variable curRecord
	variable lastRecord
	if {$curRecord < $lastRecord} then {
	    incr curRecord
	    filltxtRecord $curRecord
	    .form.txtResult delete 1.0 end
	}
	return
    }

    proc cmdPrev {} {

	variable curRecord
	if {$curRecord > 0} then {
	    incr curRecord -1
	    filltxtRecord $curRecord
	    .form.txtResult delete 1.0 end
	}
	return
    }

    proc cmdLast {} {

	variable curRecord
	variable lastRecord

	if {$curRecord < $lastRecord} then {
	    set curRecord [expr $lastRecord - 1]
	    filltxtRecord $curRecord
	    .form.txtResult delete 1.0 end
 	}
 	return
    }


    proc cmdQuitForm {} {
	variable windowSize
	variable formStack
	
	destroy .form
	::pfm::refreshFormsList
	OpenQuery $formStack(0,formId) 0
	return
    }



    ################################################################
    # Commands for OK and Cancel buttons                           #
    ################################################################

    proc cmdOK {} {

	variable ::pfm::currentDB
	variable formState

	# reworked because of bug 679
	.form.txtResult delete 1.0 end
	switch $formState {
	    "update" {
		set success [updateRecord]
		reloadRecord
	    }
	    "add" {
		if {[addRecord]} then {
		    reloadRecord
		}
	    }
	}
	newFormState browse
	return
    }

    proc cmdCancel {} {

	variable ::pfm::currentDB
	variable curRecord
	variable formState

	# reworked because of bug 679
	filltxtRecord $curRecord
	if { [string equal $formState {update}] } then {
	    set rollbackStatus "Update cancelled."
	} else {
	    set rollbackStatus "No record inserted."
	}
	.form.txtResult delete 1.0 end
	displayOnForm .form.txtResult $rollbackStatus {blue}
	newFormState browse
	return
    }

    ################################################################
    # Commands for selecting a value from a list                   #
    ################################################################

    
    proc cmdSelectFromList {attribute button action} {

	variable ::pfm::currentDB
	variable activeForm
	variable attributeArray
	variable displayList
	variable searchString
	variable txtRecord

	set searchString {}
	set base [winfo parent $button]
	set x [winfo rootx $button]
	set y [winfo rooty $button]
	set window $base.select
	destroy $window
	toplevel $window -class Toplevel
	wm transient $window $base
	wm geometry $window 300x300+$x+$y
	wm minsize $window 1 1
	wm overrideredirect $window 0
	wm resizable $window 1 1
	wm title $window "pfm - Select value for $attribute"

	# Label, entry and Next button
	label $window.lblSearch -text {Search for:}
	place $window.lblSearch \
	    -anchor nw -x 0 -y 0 \
	    -width 75 \
	    -height 30 

	entry $window.entSearch -textvariable ::form::searchString \
	    -background white
	place $window.entSearch \
	    -anchor nw -x 75 -y 0 \
	    -width {-125} -relwidth 1 \
	    -height 30 

	button $window.btnSearch -text {Next} -command \
	    "::form::cmdSearchInList $window" -takefocus 0 -underline 0
	place $window.btnSearch \
	    -anchor nw -x {-50} -y 0 -relx 1 \
	    -width 50 \
	    -height 30

	# Listbox and scrollbar
	listbox $window.lsb -yscrollcommand "$window.scroll set" \
	    -background white
	scrollbar $window.scroll -orient vertical \
	    -command "$window.lsb yview" -takefocus 0
	place $window.lsb \
	    -anchor nw -x 0 -y 30 \
	    -relwidth 1 -width -20 \
	    -height {-60} -relheight 1 
	place $window.scroll \
	    -anchor ne -x 0 -relx 1 -y 30 \
	    -width 20 \
	    -height -60 -relheight 1

	# Get data for listbox
	switch $attributeArray($attribute,typeofget) {
	    "tgList" {
		set queryDef "SELECT value,description FROM pfm_value "
		append queryDef \
		    "WHERE valuelist='$attributeArray($attribute,valuelist)'"
	    }
	    "tgLink" {
		set queryDef $attributeArray($attribute,sqlselect)
	    }
	}
	set queryRes [pg_exec $currentDB $queryDef]
	set lastItem [expr [pg_result $queryRes -numTuples] - 1]
	set valueList {}
	set displayList {}
	for {set item 0} {$item <= $lastItem} {incr item } {
	    set listItem [pg_result $queryRes -getTuple $item]
	    lappend valueList [lindex $listItem 0]
	    set displayItem {}
	    set itemNr 0
	    foreach subItem $listItem {
		set displayItem "$displayItem $subItem"
		if {$itemNr == 0} then {
		    set displayItem "$displayItem :"
		}
		incr itemNr
	    }
	    lappend displayList $displayItem
	    $window.lsb insert end $displayItem
	}
	pg_result $queryRes -clear


	# Select current item and scroll listbox
	if {$action eq {fillout}} then {
	    set attribValue $txtRecord($attribute)
	    set currentItem [lsearch $valueList $attribValue]
	    if {$currentItem < 0} then {
		set currentItem 0
	    }
	} else {
	    set currentItem 0
	}
	$window.lsb selection clear 0 end
	$window.lsb activate $currentItem
	$window.lsb see $currentItem
	$window.lsb selection set $currentItem $currentItem

	# Buttons OK and Cancel
	set command \
	    "::form::cmdValueSelected \{$valueList\} $attribute $window $action"
	button $window.btnOK -text {OK} \
	    -command $command -underline 0 -takefocus 0
	button $window.btnCancel -text {Cancel}  \
	    -command "destroy $window" -underline 0 -takefocus 0
	place $window.btnOK \
	    -anchor sw -x 0 -y 0 -rely 1 \
	    -relwidth 0.5 
	place $window.btnCancel \
	    -anchor sw -x 0 -y 0 -rely 1 -relx 0.5 \
	    -relwidth 0.5

	# Bindings
	bind $window.lsb <ButtonPress-1> "after idle \{$command\}"
	bind $window.lsb <KeyPress-Return> $command

	bind $window.entSearch <KeyPress-Return> "::form::cmdSearchInList $window"
	focus $window.lsb
	bind $window <KeyPress-Escape> "destroy $window"
	bind $window <Alt-KeyPress-o> "::pfm::PressButton $window.btnOK"
	bind $window <Alt-KeyPress-c> "::pfm::PressButton $window.btnCancel"
	bind $window <Alt-KeyPress-n> "::pfm::PressButton $window.btnSearch"
	return
    }

    proc cmdValueSelected {valueList attribute base action} {

	variable txtRecord
	variable WhereSelected
	set item [$base.lsb curselection]
	switch $action {
	    "fillout" {
		set txtRecord($attribute) [lindex $valueList $item]
	    }
	    "paste" {
		if {$WhereSelected} then {
		    .query.entWhere insert insert [lindex $valueList $item]
		} else {
		    .query.entOrderBy insert insert [lindex $valueList $item]
		}
	    }
	}
	destroy $base
	return
    }

    proc cmdSearchInList {window} {
	variable displayList
	variable searchString
	
	set lastIndex [expr [llength $displayList] - 1]
	set startPosition [expr [lindex [$window.lsb curselection] 0] + 1]
	if {$startPosition > $lastIndex} then {
	    $window.lsb selection clear 0 end
	    $window.lsb selection set 0 0
	    $window.lsb see 0
	    $window.lsb activate 0
	} else {
	    set newPosition -1
	    for {set index $startPosition} {$index <= $lastIndex} {incr index} {
		if {[string match -nocase "*$searchString*" \
			 [lindex $displayList $index]]} then {
		    set newPosition $index
		    break
		}
	    }
	    if { $newPosition >= 0 } then {
		$window.lsb selection clear 0 end
		$window.lsb selection set $newPosition $newPosition
		$window.lsb see $newPosition
		$window.lsb activate $newPosition
	    } else {
		$window.lsb selection clear 0 end
		$window.lsb selection set 0 0
		$window.lsb see 0
		$window.lsb activate 0
	    }
	}
	focus $window.lsb
	return
    }

    ###############################################################################
    # Procedures for opening the active form with a defined query                 #
    ###############################################################################

    proc OpenForm {queryDef intro displayKey} {

	variable ::pfm::currentDB
	variable ::pfm::formsArray
	variable activeForm
	variable lastRecord
	variable recordArray
	variable curRecord
	variable formAttribList
	variable formState
	variable txtRecord
	variable focusRow

	showFormWindow
	wm title .form "pfm - Form : $activeForm"
	displayOnForm .form.txtQuery "$intro\n" {blue}
	displayOnForm .form.txtQuery "$queryDef\n" {black}
	array unset recordArray
	array unset txtRecord
	set queryRes [pg_exec $currentDB $queryDef]
	set status [pg_result $queryRes -status]
	if { [string equal $status {PGRES_TUPLES_OK}] } then {
	    displayOnForm .form.txtQuery "$status\n" {green}
	    # lastRecord is a dummy empty record.
	    set lastRecord [pg_result $queryRes -numTuples]
	    pg_result $queryRes -assign recordArray
	} else {
	    set lastRecord 0
	    set status $status\n[pg_result $queryRes -error]
	    bell
	    displayOnForm .form.txtQuery "$status\n" {red}
	}
	pg_result $queryRes -clear
	for {set recordNr 0} {$recordNr < $lastRecord} {incr recordNr} {
	    set recordArray($recordNr,23status47) "Not modified"
	}
	foreach attribute $formAttribList {
	    set recordArray($lastRecord,$attribute) {}
	}
	set recordArray($lastRecord,23status47) "After last"
	if {$displayKey eq {}} then {
	    set curRecord 0
	} else {
	    set curRecord 0
	    for {set recordNr 0} {$recordNr < $lastRecord} {incr recordNr} {
		set match 1
		set i 0
		foreach pkey $formsArray($activeForm,pkey) {
		    if {[info exist recordArray($recordNr,$pkey)]} then {
			if {[lindex $displayKey $i] ne $recordArray($recordNr,$pkey)} then {
			    set match 0
			    break
			}
			incr i
		    } else {
			set errMsg "The pkey attribute '$pkey' of form '$activeForm' is not returned by the form's SQL statement. Check the form's definition!"
			tk_messageBox -type ok -icon error -message $errMsg \
			    -parent .form
			break
		    }
		}
		if {$match} then {
		    set curRecord $recordNr
		    break
		}
	    }
	}
	set focusRow 0
	set formState "browse"
	hideBrowseButtons
	hideEditButtons
	displayBrowseButtons
	displayAttribNames
	filltxtRecord $curRecord
	displayAttribLabels
	displayLinks
	return
    }

    proc displayAttribNames {} {
	variable formAttribList
	variable widget

	set rowidx 0
	foreach attribute $formAttribList {
	    label $widget(record).lb1$attribute \
		-text $attribute -anchor w \
	        -relief sunken -borderwidth 1
	    grid $widget(record).lb1$attribute \
		-column 0 -row $rowidx \
		-columnspan 1 -rowspan 1 -sticky we
	    incr rowidx
	}
	# Menu Search
	destroy .form.menu.mnSearch
	menu .form.menu.mnSearch -tearoff 0
	foreach attribute $formAttribList {
	    .form.menu.mnSearch add command \
		-command "::form::cmdOpenSearch $attribute" \
		-label $attribute
	}
	return
    }

    proc cmdOpenSearch {attribute} {
	variable formState
	variable windowSize

	if {($formState ne {update}) && ($formState ne {add})} then {
	    if {[winfo exists .form.search]} then {
		destroy .form.search
	    }
	    toplevel .form.search -class Toplevel
	    wm transient .form.search .form
	    set x [expr [winfo rootx .form] + 20]
	    set y [expr [winfo rooty .form] + 5]
	    wm geometry .form.search "+$x+$y"
	    wm title .form.search "pfm - Form: Search $attribute"
	    label .form.search.lb1 -text "Search for next record in buffer with"
	    label .form.search.lb2 -text "$attribute ="
	    entry .form.search.ent -background white -width 50
	    button .form.search.bnSearch -text Next -command \
		"::form::searchInBuffer $attribute \[.form.search.ent get\] \$bnCase" \
		-takefocus 0 -underline 0
	    button .form.search.bnCancel -text Cancel \
		-command {destroy .form.search} -underline 0 -takefocus 0
	    checkbutton .form.search.bnCase -text {Match case} \
		-underline 0 -takefocus 0
	    set help "*     matches any sequence of characters;"
	    set help "$help\n?     matches any single character."
	    message .form.search.msgHelp -width 500 -text $help -borderwidth 2 \
		-relief groove
	    focus .form.search.ent
	    bind .form.search.ent <KeyPress-Return> \
		"::form::searchInBuffer $attribute \[.form.search.ent get\] \$bnCase"
	    bind .form.search <KeyPress-Escape> \
		"::pfm::PressButton .form.search.bnCancel"
	    bind .form.search <Alt-KeyPress-c> \
		"::pfm::PressButton .form.search.bnCancel"
	    bind .form.search <Alt-KeyPress-m> \
		"::pfm::PressButton .form.search.bnCase"
	    bind .form.search <Alt-KeyPress-n> \
		"::pfm::PressButton .form.search.bnSearch"
	    grid .form.search.lb1 -column 0 -row 0 -columnspan 3
	    grid .form.search.lb2 -column 0 -row 1 -columnspan 1
	    grid .form.search.ent -column 1 -row 1 -columnspan 1
	    grid .form.search.bnSearch -column 2 -row 1 -columnspan 1 -sticky we
	    grid .form.search.bnCase -column 0 -row 2 -columnspan 1
	    grid .form.search.msgHelp -column 1 -row 2 -columnspan 1
	    grid .form.search.bnCancel -column 2 -row 2 -columnspan 1 -sticky we
	}
	return
    }

    proc searchInBuffer {attribute pattern matchcase} {
	variable curRecord
	variable lastRecord
	variable recordArray

	set searching 1
	set startSearch [expr $curRecord + 1]
	while {$searching} {
	    set found 0
	    for {set tuple $startSearch} {$tuple <= $lastRecord} {incr tuple} {
		if {$matchcase} then {
		    set found [string match $pattern $recordArray($tuple,$attribute)]
		} else {
		    set found \
			[string match -nocase $pattern $recordArray($tuple,$attribute)]
		}
		if {$found} then {
		    set curRecord $tuple
		    filltxtRecord $curRecord
		    .form.txtResult delete 1.0 end
		    break
		}
	    }
	    if {$found} then {
		set searching 0
	    } else {
		set searching \
		    [tk_messageBox -type yesno -icon question -parent .form.search \
			 -message "End of buffer reached. Wrap around?"]
		set startSearch 0
	    }
	}
	return
    }

    proc displayAttribLabels {} {

	variable formAttribList
	variable txtRecord
	variable attributeArray
	variable widget
	variable focusRow

	set rowidx 0
	foreach attribute $formAttribList {
	    entry $widget(record).ent$attribute -state readonly -width 45 \
		-textvar ::form::txtRecord($attribute) \
		-borderwidth 1 -highlightthickness 2
	    bind $widget(record).ent$attribute <Alt-KeyPress-x> \
		"::pfm::PressButton $widget(record).bn$attribute"
	    grid $widget(record).ent$attribute -column 1 -row $rowidx \
		-columnspan 1 -rowspan 1
	    button $widget(record).bn$attribute -text {Expand} -pady 0 \
		-command "::form::cmdExpand $attribute 1" \
	        -takefocus 0 -underline 1
	    grid $widget(record).bn$attribute -column 2 -row $rowidx \
		-columnspan 1 -rowspan 1
	    bind $widget(record).ent$attribute <FocusIn> \
		"::form::scrollForm $rowidx"
	    incr rowidx
	}
	# update is necessary to get a correct result from the canvas
	# bbox command
	update
	set bbox [.form.canvas bbox all]
	set rightEdge [expr [lindex $bbox 2] + 20]
	set bottomEdge [expr [lindex $bbox 3] + 20]
	.form.canvas configure -scrollregion \
	    [list 0 0 $rightEdge $bottomEdge]
	set attribute [lindex $formAttribList $focusRow]
	focus $widget(record).ent${attribute}
	return
    }

    proc hideAttribLabels {} {

	variable formAttribList
	variable widget
	foreach attribute $formAttribList {
	    destroy $widget(record).ent$attribute
	    destroy $widget(record).bn$attribute
	}
	return
    }

    proc displayAttribEntries {} {

	variable formAttribList
	variable txtRecord
	variable attributeArray
	variable widget
	variable focusRow

	set rowidx 0
	foreach attribute $formAttribList {
	    switch $attributeArray($attribute,typeofget) {
		"tgDirect" -
		"tgExpression" {
		    entry $widget(record).ent$attribute \
			-borderwidth 1 -highlightthickness 2 \
			-textvar ::form::txtRecord($attribute) \
			-relief sunken -width 45 \
			-background white
		    button $widget(record).bn$attribute -text {Expand} \
			-pady 0 \
			-command "::form::cmdExpand $attribute 0" \
		        -takefocus 0 -underline 0
		    grid $widget(record).bn$attribute -column 2 \
			-row $rowidx -columnspan 1 -rowspan 1
		}
		"tgReadOnly" {
		    entry $widget(record).ent$attribute \
			-borderwidth 1  -highlightthickness 2 \
			-textvar ::form::txtRecord($attribute) \
			-relief sunken \
			-width 45 \
			-state readonly
		    button $widget(record).bn$attribute -text {Expand} \
			-pady 0 \
			-command "::form::cmdExpand $attribute 1" \
		        -takefocus 0 -underline 0
		    grid $widget(record).bn$attribute -column 2 \
			-row $rowidx -columnspan 1 -rowspan 1
		}
		"tgList" -
		"tgLink" {
		    button $widget(record).ent$attribute -pady 0 -padx 0 \
			-anchor w -textvar ::form::txtRecord($attribute) \
			-relief raised -width 45 -command \
			"::form::cmdSelectFromList $attribute $widget(record).ent$attribute fillout" \
			-highlightthickness 2
		    bind $widget(record).ent$attribute <KeyPress-Return> \
			"::form::cmdSelectFromList $attribute $widget(record).ent$attribute fillout"
		    button $widget(record).bn$attribute -text {Expand} \
			-pady 0 \
			-command "::form::cmdExpand $attribute 0" \
			-takefocus 0 -underline 1
		    grid $widget(record).bn$attribute -column 2 \
			-row $rowidx -columnspan 1 -rowspan 1
		}
	    }
	    grid $widget(record).ent$attribute -column 1 -row $rowidx \
		-columnspan 1 -rowspan 1
	    bind $widget(record).ent$attribute <Alt-KeyPress-x> \
		"::pfm::PressButton $widget(record).bn$attribute"
	    bind $widget(record).ent$attribute <FocusIn> \
		"::form::scrollForm $rowidx"
	    incr rowidx
	}
	# update is necessary to get a correct result from the canvas
	# bbox command
	update
	set bbox [.form.canvas bbox all]
	set rightEdge [expr [lindex $bbox 2] + 20]
	set bottomEdge [expr [lindex $bbox 3] + 20]
	.form.canvas configure -scrollregion \
	    [list 0 0 $rightEdge $bottomEdge]
	set attribute [lindex $formAttribList $focusRow]
	focus $widget(record).ent${attribute}
	return
    }

    proc scrollForm {i} {
	variable formAttribList
	variable focusRow

	# Scroll form such that attribute i is visible in the middle
	# of the form.
	#
	# Let: n1 be index of first visible attribute before scrolling
	#      n2 be index of last visible attribute before scrolling
	#      nn1 be index of first visible attribute after scrolling
	#      nn2 be index of last visible attribute after scrolling
	#      f1 = 1st fraction returned by yview before scrolling (known)
	#      f2 = 2nd fraction returned by yview before scrolling (known)
	#      nf1 = 1st fraction to be given to yview for scrolling
	#      nf2 = 2nd fraction returned by yview after scrolling
	#      n = nr of attributes on form (known)
	#      i = index of attribute with input focus (arg of function)
	#      s = number of visible attributes on screen
	# Known data:
	#     n, f1, f2, i
	# To be calculated:
	#     nf1
	# Calculation:
	#     f1 = (y-coord of first visible horizontal line)
	#                                     /(height of canvas)
	#        = (n1 + 1)/(n + 2)  
	#     f2 = (y-coord of last visible horizontal line)
	#                                     /(height of canvas)
	#        = (n2 + 1)/(n + 2)
	#     -- to take into account the empty space at top and bottom
	#     -- we assume 2 dummy attributes, one at top (index "-1"),
	#     -- one at bottom (index "n")
	#     n1 = (n + 2) * f1 - 1
	#     n2 = (n + 2) * f2 - 1
	#     s = n2 - n1 + 1
	#       = (n + 2) * (f2 - f1) + 1   (1)
	#     Afters scrolling, same relationship:
	#     nn1 = (n + 2) * nf1 - 1
	#     nn2 = (n + 2) * nf2 - 1
	#     s = (n + 2) * (nf2 - nf1) + 1  (2)
	#     After scrolling we want i to be in the middle between nn1 and nn2
	#     i = (nn1 + nn2)/2
	#       = (nf1 + nf2) * (n + 2)/2 - 1 (3)
	#     s we can calculate with (1)
	#     From (2) and (3) it follows:
	#     nf1 = (2*i - s + 3)/(2 * (n + 2))

	# Remeber which row gets focus
	set focusRow $i
	set n [llength $formAttribList]
	set yviewList [.form.canvas yview]
	set f1 [lindex $yviewList 0]
	set f2 [lindex $yviewList 1]
	set s [expr ($n + 2) * ($f2 - $f1) + 1.0]
	set nf1 [expr (2.0*$i - $s + 3.0)/(2.0 * ($n + 2.0))]
	if {$nf1 < 0} then {
	    set nf1 0
	}
	if {$nf1 > 1} then {
	    set nf1 1
	}
	.form.canvas yview moveto $nf1
	return
    }

    proc cmdExpand {attribute readonly} {
	variable wrapOn
	variable txtRecord
	variable windowSize

	set wrapOn 1
	destroy .form.expand
	toplevel .form.expand -class Toplevel
	bind .form.expand <Destroy> {
	    set ::form::windowSize(.form.expand) \
		[string map {{+0+0} {}} [wm geometry .form.expand]]
	}
	wm transient .form.expand .form
	wm geometry .form.expand $windowSize(.form.expand)
	wm title .form.expand "$attribute"
	frame .form.expand.buttons
	place .form.expand.buttons \
	    -anchor sw -x 0 -y 0 -rely 1 \
	    -relwidth 1 \
	    -height 30
	button .form.expand.buttons.bnCancel -text {Cancel} \
	    -command {destroy .form.expand} -underline 0
	bind .form.expand <Alt-KeyPress-c> \
	    "::pfm::PressButton .form.expand.buttons.bnCancel"
	bind .form.expand <KeyPress-Escape> \
	    "::pfm::PressButton .form.expand.buttons.bnCancel"
	pack .form.expand.buttons.bnCancel -side right

	if { $readonly } then {
	    text .form.expand.text -wrap word -background white \
		-xscrollcommand {.form.expand.hscroll set} \
		-yscrollcommand {.form.expand.vscroll set} \
		-takefocus 0
	    focus .form.expand.buttons.bnCancel
	} else {
	    text .form.expand.text -background white -wrap word \
		-xscrollcommand {.form.expand.hscroll set} \
		-yscrollcommand {.form.expand.vscroll set}
	    button .form.expand.buttons.bnOK -text {OK} \
		-command "::form::cmdExpandOK $attribute" -underline 0
	    bind .form.expand <Alt-KeyPress-o> \
	    "::pfm::PressButton .form.expand.buttons.bnOK"
	    pack .form.expand.buttons.bnOK -side right
	    focus .form.expand.text
	}
	scrollbar .form.expand.hscroll -orient horizontal \
	    -command {.form.expand.text xview}
	scrollbar .form.expand.vscroll -orient vertical \
	    -command {.form.expand.text yview}
	place .form.expand.text \
	    -anchor nw -x 0 -y 0 \
	    -width -20 -relwidth 1 \
	    -relheight 1 -height -50
	place .form.expand.hscroll \
	    -anchor sw -x 0 -y -30 -rely 1 \
	    -width -20 -relwidth 1 \
	    -height 20
	place .form.expand.vscroll \
	    -anchor ne -x 0 -y 0 -relx 1 \
	    -width 20 \
	    -height -50 -relheight 1 
	bind .form.expand <KeyPress-Prior> {
	    .form.expand.text yview scroll -1 pages
	}
	bind .form.expand <KeyPress-Next> {
	    .form.expand.text yview scroll 1 pages
	}
	bind .form.expand <KeyPress-Up> {
	    .form.expand.text yview scroll -1 units
	}
	bind .form.expand <KeyPress-Down> {
	    .form.expand.text yview scroll 1 units
	}

	radiobutton .form.expand.buttons.rbWrap -text {Wrap} -value 1 \
	    -underline 0 \
	    -variable ::form::wrapOn -takefocus 0 \
	    -command {.form.expand.text configure -wrap word}
	radiobutton .form.expand.buttons.rbTruncate -text {Truncate} -value 0 \
	    -variable ::form::wrapOn -takefocus 0 \
	    -command {.form.expand.text configure -wrap none} -underline 0
	pack .form.expand.buttons.rbWrap -side left
	pack .form.expand.buttons.rbTruncate -side left
	bind .form.expand <Alt-KeyPress-w> \
	    "::pfm::PressButton .form.expand.buttons.rbWrap"
	bind .form.expand <Alt-KeyPress-t> \
	    "::pfm::PressButton .form.expand.buttons.rbTruncate"
	.form.expand.text insert end $txtRecord($attribute)
	if { $readonly } then {
	    .form.expand.text configure -state disabled
	}
	return
    }

    proc cmdExpandOK { attribute} {
	variable txtRecord
	variable windowSize

	set txtRecord($attribute) [.form.expand.text get 1.0 "end -1 chars"]
	destroy .form.expand
	return
    }

    proc hideAttribEntries {} {

	variable formAttribList
	variable widget
	foreach attribute $formAttribList {
	    destroy $widget(record).ent$attribute
	    destroy $widget(record).bn$attribute
	}
	return
    }

    proc displayBrowseButtons {} {

	variable activeForm
	variable ::pfm::formsArray
	variable lastRecord
	variable curRecord

	# enable Form menu
	.form.menu entryconfigure 0 -state normal
	# enable Go to menu
	.form.menu entryconfigure 2 -state normal
	# enable Help menu
	.form.menu entryconfigure 3 -state normal
	set view $formsArray($activeForm,view)
	pack .form.frmButtons.left -side left -expand 1
	if { [string equal $view {f}] } then {
	    # enable Record menu
	    .form.menu entryconfigure 1 -state normal
	    pack .form.frmButtons.btnUpdate -side left
	    pack .form.frmButtons.btnAdd -side left
	    pack .form.frmButtons.btnDelete -side left
	    bind .form <Alt-KeyPress-u> \
		{::pfm::PressButton .form.frmButtons.btnUpdate}
	    bind .form <Alt-KeyPress-a> \
		{::pfm::PressButton .form.frmButtons.btnAdd}
	    bind .form <Alt-KeyPress-d> \
		{::pfm::PressButton .form.frmButtons.btnDelete}
	}
	pack .form.frmButtons.right -side left -expand 1
	bind .form <KeyPress-Next> {::form::cmdNext}
	bind .form <KeyPress-Right> {::form::cmdNext}
	bind .form <KeyPress-Prior> {::form::cmdPrev}
	bind .form <KeyPress-Left> {::form::cmdPrev}
	bind .form <KeyPress-Escape> {
	    if {[winfo exists .form.frmLinkBtn.btnBack]} then {
		::pfm::PressButton .form.frmLinkBtn.btnBack
	    } else {
		::form::cmdQuitForm
	    }
	}
	bind .form <KeyPress-Home> {::form::cmdFirst}
	bind .form <KeyPress-End> {::form::cmdLast}
	return
    }

    proc hideBrowseButtons {} {

	foreach widget [pack slaves .form.frmButtons] {
	    pack forget $widget
	}
	# unbind keys
	bind .form <KeyPress-Next> {}
	bind .form <KeyPress-Right> {}
	bind .form <KeyPress-Prior> {}
	bind .form <KeyPress-Left> {}
	bind .form <Alt-KeyPress-u> {}
	bind .form <Alt-KeyPress-a> {}
	bind .form <Alt-KeyPress-d> {}
	bind .form <KeyPress-Escape> {}
	bind .form <KeyPress-Home> {}
	bind .form <KeyPress-End> {}
	# disable all menus
	.form.menu entryconfigure 0 -state disabled
	.form.menu entryconfigure 1 -state disabled
	.form.menu entryconfigure 2 -state disabled
	.form.menu entryconfigure 3 -state disabled
	return
    }

    proc displayEditButtons {} {

	pack .form.frmButtons.left -side left -expand 1
	pack .form.frmButtons.btnOK -side left
	pack .form.frmButtons.btnCancel -side left
	pack .form.frmButtons.right -side left -expand 1
	bind .form <Alt-KeyPress-o> "::pfm::PressButton .form.frmButtons.btnOK"
	bind .form <Alt-KeyPress-c> "::pfm::PressButton .form.frmButtons.btnCancel"
	bind .form <KeyPress-Escape> "::pfm::PressButton .form.frmButtons.btnCancel"
	return
    }

    proc hideEditButtons {} {

	foreach widget [pack slaves .form.frmButtons] {
	    pack forget $widget
	}
	bind .form <Alt-KeyPress-o> {}
	bind .form <Alt-KeyPress-c> {}
	bind .form <KeyPress-Escape> {}

	return
    }

    proc newFormState {newState} {

	variable formState
	variable txtRecord

	set formState $newState
	displayRecordFrame
	displayAttribNames
	switch $newState {
	    "update" {
		hideAttribLabels
		displayAttribEntries
		hideBrowseButtons
		displayEditButtons
		newLinkState disabled
		set txtRecord(23status47) "Updating"
	    }
	    "add" {
		hideAttribLabels
		displayAttribEntries
		hideBrowseButtons
		displayEditButtons
		newLinkState disabled
		set txtRecord(23status47) "Adding"
	    }
	    "browse" {
		hideAttribEntries
		displayAttribLabels
		hideEditButtons
		displayBrowseButtons
		newLinkState normal
	    }
	}
	return
    }

    proc filltxtRecord {recordNr} {

	variable recordArray
	variable txtRecord
	variable formAttribList
	variable lastRecord
	variable widget
	variable activeForm
	foreach attribute $formAttribList {
	    if {[info exists recordArray($recordNr,$attribute)]} then {
		set txtRecord($attribute) $recordArray($recordNr,$attribute)
	    } else {
		set txtRecord($attribute) "Undefined"
		set errMsg "The attribute '$attribute' of form '$activeForm' is not returned by the form's SQL statement. Check the form's definition!"
		tk_messageBox -type ok -icon error -message $errMsg -parent .form
	    }
	}
	set txtRecord(23nr47) "Record [expr $recordNr + 1]/$lastRecord"
	set txtRecord(23status47) $recordArray($recordNr,23status47)
	return
    }

    proc getAttributes {formName} {

	variable ::pfm::currentDB
	variable attributeArray
	variable formAttribList
	variable tableAttribList
	array unset attributeArray
	set fields "attribute,typeofattrib,typeofget,valuelist,sqlselect,\"default\""
	set queryDef \
	    "SELECT $fields FROM pfm_attribute WHERE form = '$formName' ORDER BY nr"
	set queryRes [pg_exec $currentDB $queryDef]
	set formAttribList [list]
	set tableAttribList [list]
	set lastAttribute [expr [pg_result $queryRes -numTuples] - 1]
	pg_result $queryRes -assign attribRecords
	for {set attribNr 0} {$attribNr <= $lastAttribute} {incr attribNr} {
	    set attribute [string trim $attribRecords($attribNr,attribute)]
	    lappend formAttribList $attribute
	    set typeofattrib [string trim $attribRecords($attribNr,typeofattrib)]
	    set attributeArray($attribute,typeofattrib) $typeofattrib
	    set typeofget [string trim $attribRecords($attribNr,typeofget)]
	    set attributeArray($attribute,typeofget) $typeofget
	    set valuelist [string trim $attribRecords($attribNr,valuelist)]
	    set attributeArray($attribute,valuelist) $valuelist
	    set sqlselect [string trim $attribRecords($attribNr,sqlselect)]
	    set attributeArray($attribute,sqlselect) $sqlselect
	    set defVal [string trim $attribRecords($attribNr,default)]
	    set attributeArray($attribute,default) $defVal
	    if { ![string equal $typeofget {tgReadOnly}] } then {
		lappend tableAttribList $attribute
	    }
	}
	pg_result $queryRes -clear
	array unset attribRecords
	return
    }

    ########################################################################
    # Procedures that modify the data base                                 #
    ########################################################################

    proc addRecord {} {
	# insert txtRecord to data base and to recordArray

	variable ::pfm::currentDB
	variable ::pfm::formsArray
	variable recordArray
	variable txtRecord
	variable curRecord
	variable activeForm
	variable lastRecord
	variable attributeArray
	variable tableAttribList
	variable formAttribList

	set success 1
	set colDef "("
	set valueDef "("
	foreach attribute $tableAttribList {
	    set colDef "$colDef \"$attribute\","
	    if { [string equal $attributeArray($attribute,typeofget) {tgExpression}] } then {
		set txtRecord($attribute) [expr $txtRecord($attribute)]
	    }
	    switch $attributeArray($attribute,typeofattrib) {
		"taQuoted" {
		    set convertedValue [string map {' ''} $txtRecord($attribute)]
		    append valueDef " '$convertedValue',"
		}
		"taNotQuoted" {
		    append valueDef " $txtRecord($attribute),"
		}
	    }
	}
	set colDef "[string trimright $colDef ","])"
	set valueDef "[string trimright $valueDef ","])"
	set queryDef "INSERT INTO \"$formsArray($activeForm,tablename)\" $colDef\nVALUES $valueDef"
	displayOnForm .form.txtResult "$queryDef\n" {black}
	set queryRes [pg_exec $currentDB $queryDef]
	set status [pg_result $queryRes -status]
	if { [string equal $status {PGRES_COMMAND_OK}] } then {
	    displayOnForm .form.txtResult "$status\n" {green}
	    set curRecord $lastRecord
	    incr lastRecord
	    set recordArray($curRecord,23status47) "Added"
	    set recordArray($lastRecord,23status47) "After last"
	    set txtRecord(23status47) "Added"
	    set txtRecord(23nr47) "Record [expr $curRecord + 1]/$lastRecord"
	    if {[lsearch $formsArray($activeForm,pkey) {oid}] >= 0} then {
		set oid [pg_result $queryRes -oid]
		set recordArray($curRecord,oid) $oid
	    }
	    foreach attribute $tableAttribList {
		set recordArray($curRecord,$attribute) $txtRecord($attribute)
	    }
	    foreach attribute $formAttribList {
		set recordArray($lastRecord,$attribute) {}
	    }
	} else {
	    set status $status\n[pg_result $queryRes -error]
	    set success 0
	    displayOnForm .form.txtResult "$status\n" {red}
	    set curRecord $lastRecord
	    set txtRecord(23nr47) "Record [expr $lastRecord + 1]/$lastRecord"
	    set txtRecord(23status47) "Not added"
	}
	pg_result $queryRes -clear
	if {!$success} then {
	    bell
	    tk_messageBox -type ok -icon warning -message "Add record has failed" \
		-parent .form
	}
	return $success
    }


    proc updateRecord {} {
	# copy txtRecord to data base and to recordArray

	variable ::pfm::currentDB
	variable ::pfm::formsArray
	variable recordArray
	variable txtRecord
	variable curRecord
	variable activeForm
	variable attributeArray
	variable tableAttribList

	# Reworked because of bugs 679 and 680.
	# selectForUpdate starts a transaction and checks that the record
	# still exists and has not been modified by another user. If that check
	# fails, selectForUpdate returns 'false' and the update operation is
	# canceled.
	set success 1
	if { [selectForUpdate] } then {
	    set queryDef {}
	    foreach attribute $tableAttribList {
		if {[info exists recordArray($curRecord,$attribute)]} then {
		    if { $txtRecord($attribute) ne \
			     $recordArray($curRecord,$attribute)} then {
			if { $attributeArray($attribute,typeofget) eq \
				 {tgExpression}} then {
			    set txtRecord($attribute) [expr $txtRecord($attribute)]
			}
			switch $attributeArray($attribute,typeofattrib) {
			    "taQuoted" {
				set quotesDoubled [string map {' ''} $txtRecord($attribute)]
				append queryDef " \"$attribute\"='$quotesDoubled',"
			    }
			    "taNotQuoted" {
				append queryDef " \"$attribute\"=$txtRecord($attribute),"
			    }
			}
		    }
		} else {
		    set success 0
		    set errMsg "The attribute '$attribute' of form '$activeForm' is not returned by the form's SQL statement. Check the form's definition!\n"
		    displayOnForm .form.txtResult $errMsg {red}
		}
	    }
	    if { ![string equal $queryDef {}] && $success } then {
		set queryDef [string trimright $queryDef ","]
		set queryDef "UPDATE \"$formsArray($activeForm,tablename)\"\nSET $queryDef"
		append queryDef "\nWHERE [identCurRecord 0]"
		displayOnForm .form.txtResult "$queryDef\n" {black}
		set queryRes [pg_exec $currentDB $queryDef]
		set status [pg_result $queryRes -status]
		if { [string equal $status {PGRES_COMMAND_OK}] } then {
		    set recordArray($curRecord,23status47) "Updated"
		    set txtRecord(23status47) "Updated"
		    foreach attribute $tableAttribList {
			set recordArray($curRecord,$attribute) $txtRecord($attribute)
		    }
		    displayOnForm .form.txtResult "$status\n" {green}
		    set endTransaction {COMMIT WORK}
		    displayOnForm .form.txtResult "$endTransaction\n" {black}
		} else {
		    set status $status\n[pg_result $queryRes -error]
		    set success 0
		    displayOnForm .form.txtResult "$status\n" {red}
		    set endTransaction {ROLLBACK WORK}
		    displayOnForm .form.txtResult "$endTransaction\n" {black}
		}
		pg_result $queryRes -clear
		filltxtRecord $curRecord
	    } else {
		set status {No updates.}
		set success 0
		displayOnForm .form.txtResult "$status\n" {blue}
		set endTransaction {ROLLBACK WORK}
		displayOnForm .form.txtResult "$endTransaction\n" {black}
		set txtRecord(23status47) $recordArray($curRecord,23status47)
	    }
	} else {
	    set status {Your update is cancelled.}
	    set success 0
	    displayOnForm .form.txtResult "$status\n" {blue}
	    set endTransaction {ROLLBACK WORK}
	    displayOnForm .form.txtResult "$endTransaction\n" {black}
	    set txtRecord(23status47) $recordArray($curRecord,23status47)
	}
	set commitResult [pg_exec $currentDB $endTransaction]
	set commitStatus [pg_result $commitResult -status]
	if { ![string equal $commitStatus {PGRES_COMMAND_OK}] } then {
	    set commitStatus "$commitStatus\n[pg_result $commitResult -error]"
	    displayOnForm .form.txtResult "$commitStatus\n" {red}
	    set success 0
	    set txtRecord(23status47) $recordArray($curRecord,23status47)
	} else {
	    displayOnForm .form.txtResult "$commitStatus\n" {green}
	}
	pg_result $commitResult -clear
	if {!$success} then {
	    bell
	    tk_messageBox -type ok -icon warning -message "No updates done" \
		-parent .form
	}
	return $success
    }

#########################################################################
# selectForUpdate and reloadRecord                                      #
#########################################################################

    proc selectForUpdate { } {

	# Introduced because of bugs 679 and 680.
	# selectForUpdate starts a transaction and checks that the record
	# still exists and has not been modified by another user. If that check
	# fails, selectForUpdate returns 'false'.

	variable ::pfm::currentDB
	variable ::pfm::formsArray
	variable txtRecord
	variable recordArray
	variable curRecord
	variable tableAttribList
	variable activeForm
	set success 1
	set status {}
	set queryDef {}
	set sqlAttrib \"[join $tableAttribList "\", \""]\"
	set beginWork [pg_exec $currentDB "BEGIN WORK"]
	displayOnForm .form.txtResult "BEGIN WORK\n" {black}
	set beginStatus [pg_result $beginWork -status]
	if { [string equal $beginStatus {PGRES_COMMAND_OK}] } then {
	    displayOnForm .form.txtResult "$beginStatus\n" {green}
	    set queryDef "SELECT $sqlAttrib FROM \"$formsArray($activeForm,tablename)\""
	    set queryDef "$queryDef\nWHERE [identCurRecord 0] FOR UPDATE"
	    displayOnForm .form.txtResult "$queryDef\n" {black}
	    set queryRes [pg_exec $currentDB $queryDef]
	    set status [pg_result $queryRes -status]
	    if { [string equal $status {PGRES_TUPLES_OK}] } then {
		displayOnForm .form.txtResult "$status\n" {green}
		switch -- [pg_result $queryRes -numTuples] {
		    1 {
			pg_result $queryRes -assign recordForUpdate
			foreach attribute $tableAttribList {
			    if {[info exists recordArray($curRecord,$attribute)]} then {
				if { ![string equal $recordArray($curRecord,$attribute) \
					   $recordForUpdate(0,$attribute)] } then {
				    set success 0
				    set status \
					"Record was modified by another user after you pressed the \[Update\] button.\nReconsider your input."
				    displayOnForm .form.txtResult "$status\n" {red}
				    break
				}
			    } else {
				set success 0
				set status \
				    "Attribute '$attribute' of form '$activeForm' was not returned by the form's SQL statement.\nCheck the form's definition!"
				displayOnForm .form.txtResult "$status\n" {red}
				break
			    }
			}
		    }
		    0 {
			set success 0
			# Bug 680 : selectForUpdate has to take into account that the the
			# current record may have been deleted by another user.
			set status \
			    "Record was deleted by another user after you pressed \[Update\]."
			displayOnForm .form.txtResult "$status\n" {red}
		    }
		    default {
			set success 0
			set status "Error in definition of form '$activeForm'."
			set status "$status\npkey '$formsArray($activeForm,pkey)' does not uniquely define a record."
			set status "$status\n[pg_result $queryRes -numTuples] records where returned by the query."
			displayOnForm .form.txtResult "$status\n" {red}
		    }
		}
	    } else {
		set success 0
		set status $status\n[pg_result $queryRes -error]
		displayOnForm .form.txtResult "$status\n" {red}
	    }
	    pg_result $queryRes -clear
	} else {
	    set success 0
	    set beginStatus "$beginStatus\n[pg_result $beginWork -error]"
	    displayOnForm .form.txtResult "$beginStatus\n" {red}
	}
	pg_result $beginWork -clear
	if {!$success} then {
	    bell
	    tk_messageBox -type ok -icon warning -parent .form \
		-message "Select for update failed"
	}
	return $success
    }

    proc reloadRecord { } {

	variable ::pfm::currentDB
	variable ::pfm::formsArray
	variable txtRecord
	variable recordArray
	variable curRecord
	variable formAttribList
	variable activeForm
	set success 1
	if { ![string equal $txtRecord(23status47) {After last}] && \
		![string equal $txtRecord(23status47) {Deleted}] && \
		![string equal $txtRecord(23status47) {Not added}] } then {
	    set sqlAttrib $formsArray($activeForm,sqlselect)
	    set sqlFrom $formsArray($activeForm,sqlfrom)
	    set groupby $formsArray($activeForm,groupby)
	    set tableName $formsArray($activeForm,tablename)
	    set queryDef "SELECT $sqlAttrib\nFROM $sqlFrom"
	    if { [string length $groupby] != 0 } then {
		set queryDef "$queryDef\nGROUP BY $groupby"
		# If there is a GROUP BY clause, the WHERE clause must become a
		# HAVING clause
		set queryDef "$queryDef\nHAVING [identCurRecord 1]"
	    } else {
		set queryDef "$queryDef\nWHERE [identCurRecord 1]"
	    }
	    displayOnForm .form.txtResult "Reload record:\n" {blue}
	    displayOnForm .form.txtResult "$queryDef\n" {black}
	    set queryRes [pg_exec $currentDB $queryDef]
	    set status [pg_result $queryRes -status]
	    if { [string equal $status {PGRES_TUPLES_OK}]} then {
		displayOnForm .form.txtResult "$status\n" {green}
		switch -- [pg_result $queryRes -numTuples] {
		    1 {
			pg_result $queryRes -assign reloadedRecord
			foreach attribute $formAttribList {
			    if {[info exists reloadedRecord(0,$attribute)]} then {
				set recordArray($curRecord,$attribute) $reloadedRecord(0,$attribute)
				set txtRecord($attribute) $recordArray($curRecord,$attribute)
			    } else {
				set errMsg "The attribute '$attribute' of form '$activeForm' is not returned by the form's SQL statement. Check the form's definition!"
				displayOnForm .form.txtResult "$errMsg\n" {red}
				set succes 0
			    }
			}
		    }
		    0 {
			set success 0
			set status "The query did not return any tuples."
			displayOnForm .form.txtResult "$status\n" {red}
			set status "The record is deleted from the internal buffer,"
			append status " but the database is not modified."
			displayOnForm .form.txtResult "$status\n" {blue}
			set recordArray($curRecord,23status47) "Deleted"
			set txtRecord(23status47) "Deleted"
			foreach attribute $formAttribList {
			    set recordArray($curRecord,$attribute) ""
			    set txtRecord($attribute) ""
			}
		    }
		    default {
			set success 0
			set status "Error in definition of form $activeForm."
			set status "$status\npkey '$formsArray($activeForm,pkey)' does not uniquely define a record."
			set status "$status\n[pg_result $queryRes -numTuples] records where returned by the query."
			displayOnForm .form.txtResult "$status\n" {red}
		    }
		}
	    } else {
		set success 0
		set status "$status\n[pg_result $queryRes -error]"
		displayOnForm .form.txtResult "$status\n" {red}
	    }
	    pg_result $queryRes -clear
	} else {
	    set success 0
	    set status {Record not reloaded}
	    displayOnForm .form.txtResult "$status\n" {blue}
	}
	if {!$success} then {
	    bell
	    tk_messageBox -type ok -icon warning -parent .form -message \
		"Reload record has failed"
	}
	return $success
    }

    ###########################################################################
    #                                                                         #
    # Procedure for writing text on the text windows txtQuery and txtResult   #
    #                                                                         #
    ###########################################################################

    proc displayOnForm {txtWidget text colour} {

	if {$colour ne {black}} then {
	    set begin [$txtWidget index "end - 1 chars"]
	    $txtWidget insert end $text
	    set end [$txtWidget index "end -1 chars"]
	    $txtWidget tag add ${colour}Tag $begin $end
	} else {
	    $txtWidget insert end $text
	}
	return
    }

    #############################################################################
    #                                                                           #
    #  Procedures treating the links                                            #
    #                                                                           #
    #############################################################################

    proc displayLinks {} {

	variable ::pfm::currentDB
	variable activeForm
	variable linksArray
	variable lastFormOnStack

	if {$lastFormOnStack != 0} then {
	    button .form.frmLinkBtn.btnBack -text "0: Back (Esc)" -pady 0 -anchor w \
		-command ::form::cmdBack -takefocus 0 \
		-underline 0
	    bind .form <Alt-KeyPress-0> \
		"::pfm::PressButton .form.frmLinkBtn.btnBack"
	    grid .form.frmLinkBtn.btnBack -row 0 \
		    -column 0 -rowspan 1 -columnspan 1 -sticky we
	}
	array unset linksArray
	set queryDef "SELECT * from pfm_link WHERE fromform = '$activeForm'"
	set queryRes [pg_exec $currentDB $queryDef]
	pg_result $queryRes -assign linksArray
	set lastLink [expr [pg_result $queryRes -numTuples] -1]
	pg_result $queryRes -clear
	set rowidx 1
	for {set link 0} {$link <= $lastLink} {incr link} {
	    set linkName $linksArray($link,linkname)
	    button .form.frmLinkBtn.btn$link -text "   $linkName" \
		-pady 0 -anchor w \
		-command "::form::cmdFollowLink $link" \
		-takefocus 0
	    if {$rowidx <= 9} then {
		.form.frmLinkBtn.btn$link configure -text "${rowidx}: $linkName" \
		    -underline 0
		set key "<Alt-KeyPress-${rowidx}>"
		bind .form $key "::pfm::PressButton .form.frmLinkBtn.btn${link}"
	    }
	    grid .form.frmLinkBtn.btn$link -row $rowidx \
		    -column 0 -rowspan 1 -columnspan 1 -sticky we
	    incr rowidx
	}
	return
    }

    proc newLinkState {newState} {

	if { [winfo exists .form.frmLinkBtn.btnBack] } then {
	    .form.frmLinkBtn.btnBack configure -state $newState
	}
	set link 0
	while { [winfo exists .form.frmLinkBtn.btn$link] } {
	    .form.frmLinkBtn.btn$link configure -state $newState
	    incr link
	}
	return
    }

    proc cmdFollowLink {link} {

	variable ::pfm::formsArray
	variable linksArray
	variable txtRecord
	variable attributeArray
	variable activeForm
	variable recordArray
	variable curRecord
	variable lastFormOnStack
	variable formStack
	variable windowSize

	if { ![string equal $txtRecord(23status47) {After last}] && \
		![string equal $txtRecord(23status47) {Deleted}] && \
		![string equal $txtRecord(23status47) {Not added}] } then {
	    if {[winfo exists .form.search]} then {
		destroy .form.search
	    }
	    set whereDef [expandSqlWhere $linksArray($link,sqlwhere) $link]
	    set orderDef $linksArray($link,orderby)
	    set dispAttribList $linksArray($link,displayattrib)
	    set displayDef " "
	    foreach attribute $dispAttribList {
		if {[info exists txtRecord($attribute)]} then {
		    set displayDef "$displayDef$txtRecord($attribute) "
		} else {
		    set errMsg "Wrong definition of 'displayattrib' for link '$linksArray($link,linkname)'."
		    set errMsg "$errMsg\n'$attribute' is not an attribute of form '$activeForm'"
		    tk_messageBox -type ok -icon error -parent .form -message $errMsg
		}
	    }
	    set From "$activeForm \($displayDef\)"
	    set To "$linksArray($link,toform)"
	    # remember view attribute of 'fromform'
	    set view $formsArray($activeForm,view)
	    # remember active form
	    set fromForm $activeForm
	    # prepare form pointed to by the link
	    set activeForm $linksArray($link,toform)
	    getAttributes $activeForm
	    set sqlAttrib $formsArray($activeForm,sqlselect)
	    if {!$formsArray($activeForm,view) && \
		     ([lsearch $formsArray($activeForm,pkey) {oid}] >= 0) && \
		     ([regexp {\moid\M} $sqlAttrib] == 0)} then {
		set sqlAttrib \
		    "\"$formsArray($activeForm,tablename)\".oid, $sqlAttrib"
	    }
	    set sqlFrom $formsArray($activeForm,sqlfrom)
	    set groupby $formsArray($activeForm,groupby)
	    set queryDef "SELECT $sqlAttrib\nFROM $sqlFrom"
	    set intro "Link \'$linksArray($link,linkname) : $From -> $To\'"
	    if { [string length $groupby] != 0 } then {
		append queryDef "\nGROUP BY $groupby"
		# If there is a GROUP BY clause, the whereDef must become a
		# HAVING clause.
		append queryDef "\nHAVING $whereDef"
	    } else {
		append queryDef "\nWHERE $whereDef"
	    }
	    if { [string length $orderDef] != 0 } then {
		append queryDef "\nORDER BY $orderDef"
	    }
	    set displayKey {}
	    foreach pkey $formsArray($fromForm,pkey) {
		if {[info exists recordArray($curRecord,$pkey)]} then {
		    lappend displayKey $recordArray($curRecord,$pkey)
		} else {
		    set errMsg "The pkey attribute '$pkey' of form '$activeForm' is not returned by the form's SQL statement. Check the form's definition!"
		    tk_messageBox -type ok -icon error -parent .form \
			-message $errMsg
		}
	    }
	    set formStack($lastFormOnStack,displayKey) $displayKey
	    incr lastFormOnStack
	    set formStack($lastFormOnStack,formId) $activeForm
	    set formStack($lastFormOnStack,queryDef) $queryDef
	    set formStack($lastFormOnStack,intro) $intro
	    OpenForm $queryDef $intro {}
	}
	return
    }

    proc cmdBack {} {

	variable activeForm
	variable lastFormOnStack
	variable formStack
	if {$lastFormOnStack >= 1} then {
	    incr lastFormOnStack -1
	    set activeForm $formStack($lastFormOnStack,formId)
	    getAttributes $activeForm
	    set queryDef $formStack($lastFormOnStack,queryDef)
	    set intro $formStack($lastFormOnStack,intro)
	    set displayKey $formStack($lastFormOnStack,displayKey)
	    OpenForm $queryDef $intro $displayKey
	}
	return
    }

    proc expandSqlWhere {sqlWhere link} {
	# This procedure finds the $(name) variables,
	# and replaces them with $txtRecord(name)
	# and also doubles all single quotes

	variable txtRecord
	variable linksArray

	set expandWhere $sqlWhere
	set first [string first "\$(" $expandWhere]
	while {$first >= 0} {
	    set last [string first ")" $expandWhere $first]
	    set parName [string range $expandWhere [expr $first + 2] [expr $last -1]]
	    if {[info exists txtRecord($parName)]} then {
		set parameter [string map {' ''} $txtRecord($parName)]
		set expandWhere [string replace $expandWhere $first $last $parameter]
	    } else {
		set errMsg \
		    "Error in 'sqlwhere' of link '$linksArray($link,linkname)' "
		append errMsg \
		    "from '$linksArray($link,fromform)' to '$linksArray($link,toform)': "
		append errMsg \
		    "No attribute '$parName' in form '$linksArray($link,fromform)'"
		tk_messageBox -type ok -icon error -parent .form -message $errMsg
	    }
	    set first [string first "\$(" $expandWhere $last]
	}
	return $expandWhere
    }

    ############################################################
    # init namespace form                                      #
    ############################################################

    set windowSize(.query) {600x200}
    set windowSize(.form) {700x500}
    set windowSize(.form.help) {600x400}
    set windowSize(.form.expand) {600x400}
    set windowSize(.form.search) {}

}


##################################################################
# End of namespace form                                          #
##################################################################

##################################################################
# Begin namespace report                                         #
##################################################################

# commandHistory is an associative array where
#
#     - commandHistory(top) is the index in the history where
#           the next command will be stored, i.e. it points
#           to the first free element in the command history array.
#
#     - commandHistory(cursor) is the index in the history of
#               the command that is currently displayed
#
#     - commandHistory($n) is the $n-th command in the history
#
# reportDef is an array which contains the data stored in table
#           pfm_report. It is filled out by cmdDisplayReportList.
#
# layout is an array where layout($level) = row, column or table
#         - "row" means that the labels and values are printed on 1 row
#         - "column" means that the labels are printed in the
#            first column and the values in the second column
#         - "table" means that the values are printed in a
#            table with the labels as table header
#               
#           It is filled out by collectReportLayout.
#
# fieldNamesList($level) contains, for each level, the list
# of field names to be displayed in that level.
#
# printInfo($name,label) contains the label to be used for the
# field "$name". It is filled out by collectReportLayout
#
# printInfo($name,alignment) contains the alignment info:
#    l: left
#    r: right
#
# printInfo($name,width) contains the maximum width for the
# field $name. It is filled out by cmdRunReport.
#
# printInfo($name,columnWidth) contains the maximum of
# the length of the field's label and printInfo($name,width).
# It is the optimum width for a table column displaying the field data.
#
# maxLabelWidth($level,section) contains, for each level, the maximum label width
# for the section $level. It is filled out by collectReportLayout.
#
# maxLabelWidth($level,summary) contains, for each level, the maximum label width
# for the summary $level. It is filled out by collectReportLayout.
#
# targetFields($level) contains the list of summary fields defined in section $level,
# i.e. summary fields meant to be printed at the end of each $level section
#
# sourceFields($level) contains the list of summary fields that are defined as
# normal field in section $level.
#
# summaryDef($level,$seqnr,field), summaryDef($level,$seqnr,aggregate) and
# summaryDef($level,$seqnr,format)
# contain the field, the aggregate operator and the format of the summary field
# $seqnr defined in section $level. The $seqnr is assigned automatically by pfm
# to provide a unique identification of a summary field within a section and to allow
# e.g. both SUM and AVG of the same field in a section.
#
# summaryList($level) the list of seqnr of summary fields for section $level.
#
# runBegin($tuple,$level) is a boolean which indicates whether or not $tuple
# is the begin of a run of $level.
#
# runEnd($tuple,$level) is a boolean which indicates whether or not $tuple
# is the end of a run of $level.
#
# lowestPL($tuple) is the lowest level section that needs to be printed
# for record $tuple.
#
# reportData contains the result of the query associated with the
# report. It is filled out by cmdRunReport.
#
# summary($field,$targetLevel) contains a list of all the values to be used for
# the aggregate functions operating on $field for $targetLevel.


namespace eval report {

    variable ::pfm::currentDB
    variable ::pfm::dbName
    variable ::pfm::psqlChannel
    variable commandHistory
    variable reportDef
    variable layout
    variable fieldNamesList
    variable printInfo
    variable maxLabelWidth
    variable targetFields
    variable sourceFields
    variable summaryList
    variable summaryDef
    variable runBegin
    variable runEnd
    variable lowestPL
    variable summary
    variable reportData
    variable reportMode {sql}
    variable printCommand
    variable parmlist
    variable windowSize

########################################################################
#  This is the entry point for this namespace                          #
########################################################################

    proc cmdReportSQL {requestedMode} {
	variable ::pfm::currentDB
	variable ::pfm::dbName
	variable reportMode
	variable wrapOn
	variable windowSize

	set wrapOn 0
	set reportMode $requestedMode
	if { [info exists currentDB] } then {
	    if {[winfo exists .report]} {
		destroy .report
	    }
	    toplevel .report -class Toplevel
	    bind .report <Destroy> {
		set ::report::windowSize(.report) \
		    [string map {{+0+0} {}} [wm geometry .report]]
	    }
	    initCommandHistory
	    # wm focusmodel .report passive
	    wm geometry .report $windowSize(.report)
	    wm minsize .report 1 1
	    # wm overrideredirect .report 0
	    wm resizable .report 1 1
	    wm deiconify .report
	    # wm title .report "pfm - Reports and Queries : $dbName"

	    # Define menubar

	    menu .report.menubar -tearoff 0
	    .report.menubar add cascade \
		-menu .report.menubar.mnWindow \
		-label {Window} -underline 0
	    .report.menubar add cascade \
		-menu .report.menubar.mnSQL \
		-label {SQL} -underline 0
	    .report.menubar add cascade \
		-menu .report.menubar.mnOutput \
		-label {Output} -underline 0
	    .report.menubar add cascade \
		-menu .report.menubar.mnHelp \
		-label {Help} -underline 0

	    # Make it the menubar
	    .report configure -menu .report.menubar

	    # Window menu
	    menu .report.menubar.mnWindow -tearoff 0
	    .report.menubar.mnWindow add radiobutton \
		-label {SQL} -underline 0 -accelerator {Cntrl-s} \
		-variable ::report::reportMode \
		-value {sql} \
		-command ::report::cmdEnterSQLmode
	    .report.menubar.mnWindow add radiobutton \
		-label {Report} -underline 0 -accelerator {Cntrl-r} \
		-variable ::report::reportMode \
		-value {report} \
		-command ::report::cmdEnterReportMode
	    .report.menubar.mnWindow add command \
		-label {Close} -underline 0 -accelerator {Esc} \
		-command ::report::cmdQuitReport
	    bind .report <KeyPress-Escape> ::report::cmdQuitReport
	    bind .report <Control-KeyPress-s> {
		set ::report::reportMode {sql}
		::report::cmdEnterSQLmode
	    }
	    bind .report <Control-KeyPress-r> {
		set ::report::reportMode {report}
		::report::cmdEnterReportMode
	    }
	    
	    # SQL menu
	    menu .report.menubar.mnSQL -tearoff 0
	    .report.menubar.mnSQL add command \
		-label {Import SQL from file} -underline 0 \
		-command "::report::cmdImportFile {Open SQL-file}"
	    .report.menubar.mnSQL add command \
		-label {Save SQL to file} -underline 0 \
		-command {::report::cmdSaveFile .report.fmSQL.text \
			      "Save SQL as" "sql"}
	    .report.menubar.mnSQL add command \
		-label {List of Databases} -underline 8 \
		-command {::report::cmdShortCut {\l}}
	    .report.menubar.mnSQL add command \
		-label {List of Relations in database} -underline 8 \
		-command {::report::cmdShortCut {\d}}

	    # Output menu
	    menu .report.menubar.mnOutput -tearoff 0
	    .report.menubar.mnOutput add command \
		-label {Save} -underline 0 \
		-command {::report::cmdSaveFile .report.txtOutput \
			      "Save Output as" "text"}
	    .report.menubar.mnOutput add command \
		-label {Print} -underline 0 \
		-command {::report::cmdPrint .report.txtOutput}
	    .report.menubar.mnOutput add command \
		-label {Clear} -underline 0 \
		-command {.report.txtOutput delete 1.0 end}
	    .report.menubar.mnOutput add separator
	    .report.menubar.mnOutput add radiobutton \
		-label {Wrap} -underline 0 \
		-variable ::report::wrapOn -value 1 \
		-command {.report.txtOutput configure -wrap char}
	    .report.menubar.mnOutput add radiobutton \
		-label {Truncate} -underline 0 \
		-variable ::report::wrapOn -value 0 \
		-command {.report.txtOutput configure -wrap none}
	    .report.menubar.mnOutput add separator
	    .report.menubar.mnOutput add command \
		-label {Scroll Up} -underline 7 -accelerator {Alt-Up} \
		-command {.report.txtOutput yview scroll -1 pages}
	    .report.menubar.mnOutput add command \
		-label {Scroll Down} -underline 7 -accelerator {Alt-Down} \
		-command {.report.txtOutput yview scroll 1 pages}
	    .report.menubar.mnOutput add command \
		-label {Scroll Left} -underline 7 -accelerator {Alt-Left} \
		-command {.report.txtOutput xview scroll -1 pages}
	    .report.menubar.mnOutput add command \
		-label {Scroll Right} -underline 7 -accelerator {Alt-Right} \
		-command {.report.txtOutput xview scroll 1 pages}
	    bind .report <Alt-KeyPress-Up> \
		{.report.txtOutput yview scroll -1 pages}
	    bind .report <Alt-KeyPress-Down> \
		{.report.txtOutput yview scroll 1 pages}
	    bind .report <Alt-KeyPress-Left> \
		{.report.txtOutput xview scroll -1 pages}
	    bind .report <Alt-KeyPress-Right> \
		{.report.txtOutput xview scroll 1 pages}
	    
	    # Help menu
	    menu .report.menubar.mnHelp -tearoff 0
	    .report.menubar.mnHelp add command \
		-label {List of SQL-commands} -underline 0 \
		-command {::report::cmdShortCut {\h}}
	    .report.menubar.mnHelp add command \
		-label {psql \-commands} -underline 0 \
		-command {::report::cmdShortCut {\?}}
	    set helpText \
		"-- Type name of SQL-command on the next line after the \\h."
	    append helpText "\n\\h "
	    set command "::report::cmdClear\n"
	    append command ".report.fmSQL.text insert end \{${helpText}\}"
	    .report.menubar.mnHelp add command \
		-label {Help for a particular SQL command} -underline 0 \
		-command $command

	    # define widgets

	    # define middle frame and buttons for SQL command

	    frame .report.fmMiddle -borderwidth 2 -relief groove
	    place .report.fmMiddle \
		-anchor nw -x 0 -y -30 -rely 0.4 \
		-relwidth 1 \
		-height 30
	    button .report.fmMiddle.bnRun -text {Run} \
		-command {::report::cmdRun} \
		-underline 0 -takefocus 0
	    bind .report <Alt-KeyPress-r> \
		"::pfm::PressButton .report.fmMiddle.bnRun"
# 	    bind .report <Alt-KeyPress-Return> \
# 		"::pfm::PressButton .report.fmMiddle.bnRun"
	    button .report.fmMiddle.bnClear -text {Clear} -underline 0 \
 		-command ::report::cmdClear -takefocus 0
	    bind .report <Alt-KeyPress-c> \
		"::pfm::PressButton .report.fmMiddle.bnClear"
	    button .report.fmMiddle.bnNext -text {Next} -underline 0 \
		-command ::report::cmdNext -takefocus 0
	    bind .report <Alt-KeyPress-n> \
		"::pfm::PressButton .report.fmMiddle.bnNext"
	    bind .report <Control-KeyPress-Down> \
		"::pfm::PressButton .report.fmMiddle.bnNext"
	    button .report.fmMiddle.bnPrev -text {Previous} -underline 0 \
		-command ::report::cmdPrev -takefocus 0
	    bind .report <Alt-KeyPress-p> \
		"::pfm::PressButton .report.fmMiddle.bnPrev"
	    bind .report <Control-KeyPress-Up> \
		"::pfm::PressButton .report.fmMiddle.bnPrev"
	    
	    pack .report.fmMiddle.bnRun -side right
	    pack .report.fmMiddle.bnNext -side right
	    pack .report.fmMiddle.bnPrev -side right
	    pack .report.fmMiddle.bnClear -side right

	    # define scrolled text widget for displaying Outputs

	    label .report.lblOutput -text {Output}
	    text .report.txtOutput -wrap none -background white \
		-xscrollcommand {.report.hscrollOutput set} \
		-yscrollcommand {.report.vscrollOutput set} -takefocus 0

	    # Tag errTag which will be used to highlight error
	    # messages from psql.

	    .report.txtOutput tag configure errTag -foreground red3

	    # Scrollbars
	    scrollbar .report.hscrollOutput -orient horizontal \
		-command {.report.txtOutput xview} -takefocus 0
	    scrollbar .report.vscrollOutput -orient vertical \
		-command {.report.txtOutput yview} -takefocus 0

	    # Place the output widgets
	    place .report.lblOutput \
		-anchor n -x 0 -y 0 -relx 0.5 -rely 0.4 \
		-height 15
	    place .report.txtOutput \
		-anchor nw -x 0 -y 15 -rely 0.4 \
		-width -20 -relwidth 1 \
		-height -35 -relheight 0.6 
	    place .report.hscrollOutput \
		-anchor sw -x 0 -y 0 -rely 1 \
		-width -20 -relwidth 1 \
		-height 20
	    place .report.vscrollOutput \
		-anchor ne -x 0 -y 15 -relx 1 -rely 0.4 \
		-width 20 \
		-height -35 -relheight 0.6 

	    # Initially enter mode according to $reportMode

	    switch $reportMode {
		sql {
		    cmdEnterSQLmode
		}
		report {
		    cmdEnterReportMode
		}
	    }
	} else {
	    tk_messageBox -message "There is no data base open!" -type ok \
		-icon error -parent .
	}
	return
    }

###########################################################################
# Common procedures, i.e. for both SQL and report mode                    #
###########################################################################

    proc cmdSaveFile {txtWidget title type} {
	
	set textToSave [$txtWidget get 1.0 end]
	switch $type {
	    "sql" {
		set fileTypes {
		    {{SQL statements} {.sql} }
		    {{All files} *}
		}
		set defaultExt ".sql"
	    }
	    "text" {
		set fileTypes {
		    {{Text files} {.txt} }
		    {{All files} *}
		}
		set defaultExt ".txt"
	    }
	}
	set fileName [tk_getSaveFile -title $title -filetypes $fileTypes \
			  -defaultextension $defaultExt -parent .report]
	if { $fileName !=  "" } then {
	    set file_ch [open $fileName w]
	    puts $file_ch $textToSave
	    close $file_ch
	}
	return
    }


    proc cmdImportFile {title} {
	variable ::pfm::pfmConfig
	
	set txtWidget ".report.fmSQL.text"
	set initialDir [file normalize ~]
	set fromEncoding [encoding system]
	set fileTypes {
	    {{SQL statements} {.sql} }
	    {{All files} *}
	}
	set defaultExt ".sql"
	set fileName [tk_getOpenFile -title $title -filetypes $fileTypes \
			  -defaultextension $defaultExt -parent .report \
			  -initialdir $initialDir]
	    
	if {![string equal $fileName {}]} then {
	    toplevel .report.import -class Toplevel
	    wm transient .report.import .report
	    wm title .report.import "Import SQL - Select encoding"
	    set shortFileName [file tail $fileName]
	    set message \
		"Select the character encoding with which ${shortFileName} was made.\n"
	    append message \
		"Your system's default encoding is '${fromEncoding}'\n"
	    append message \
		"That is usually OK, but if you know that ${shortFileName}\n"
	    append message \
		"was made with another encoding, you can select it now."
	    label .report.import.lbMessage -text $message
	    label .report.import.lbEncoding -text {Selected encoding:}
	    button .report.import.bnEncoding -text $fromEncoding \
		-command ::report::GetEncoding
	    set cmdContinue "::report::ProcessSQLscript \{${fileName}\} "
	    append cmdContinue "\[.report.import.bnEncoding cget -text\] $txtWidget"
	    button .report.import.bnContinue -text {Continue} -command $cmdContinue
	    grid .report.import.lbMessage -row 0 -column 0 -columnspan 3
	    grid .report.import.lbEncoding -row 1 -column 0 -columnspan 1 -sticky we
	    grid .report.import.bnEncoding -row 1 -column 1 -columnspan 1 -sticky we
	    grid .report.import.bnContinue -row 1 -column 2 -columnspan 1 -sticky we
	    focus .report.import.bnEncoding
	}
	return
    }

    proc GetEncoding {} {

	toplevel .report.import.encoding -class Toplevel
	wm transient .report.import.encoding .report.import
	set x [winfo rootx .report.import.bnEncoding]
	set y [winfo rooty .report.import.bnEncoding]
	wm geometry .report.import.encoding 300x400+$x+$y
	wm title .report.import.encoding "Select encoding"
	listbox .report.import.encoding.lsb -yscrollcommand \
	    {.report.import.encoding.scrolly set}
	scrollbar .report.import.encoding.scrolly -orient vertical \
	    -command {.report.import.encoding.lsb yview} -takefocus 0
	place .report.import.encoding.lsb \
	    -anchor nw -x 0 -y 0 \
	    -relwidth 1 -width -20 \
	    -relheight 1
	place .report.import.encoding.scrolly \
	    -anchor ne -x 0 -y 0 -relx 1 \
	    -width 20 \
	    -relheight 1

	set encodingList [lsort [encoding names]]
	foreach item $encodingList {
	    .report.import.encoding.lsb insert end $item
	}
	.report.import.encoding.lsb selection clear 0 end
	.report.import.encoding.lsb selection set 0 0
	set command {
	    .report.import.bnEncoding configure -text \
		[.report.import.encoding.lsb get \
		     [.report.import.encoding.lsb curselection]]
	    destroy .report.import.encoding
	    focus .report.import.bnContinue
	}
	bind .report.import.encoding.lsb <ButtonPress-1> \
	    "after idle \{$command\}"
	bind .report.import.encoding.lsb <KeyPress-Return> $command
	focus .report.import.encoding.lsb
	return
    }

    proc ProcessSQLscript {fileName fromEncoding txtWidget} {

	destroy .report.import
	set title "Import SQL"
	set message "Offer file to psql using \\i (recommended for large files),"
	append message "\nor import directly in SQL window?"
	set buttonList {}
	lappend buttonList {Offer file to psql}
	lappend buttonList {Import file in SQL window}
	lappend buttonList {Cancel}
	switch -- \
	    [::pfm::pfmDialog .report.dialog $title $message 400 0 $buttonList] {
		0 {
		    $txtWidget insert end "\\i '[ConvertToUTF-8 $fileName $fromEncoding]'"
		}
		1 {
		    set file_ch [open $fileName r]
		    fconfigure $file_ch -encoding $fromEncoding
		    $txtWidget insert end [read $file_ch]
		    close $file_ch
		}
		default {
		}
	    }
	return
    }

    proc ConvertToUTF-8 {fileName fromEncoding} {
	# This procedures converts $fileName from $fromEncoding to UTF-8.
	# It writes the converted file in pfmOptions(tmpdir) and returns
	# the name of the converted file.
	#
	# Even if the $fromEncoding = utf-8, we execute this conversion.
	# Tcl seems to be rather clever to recognise encodings such that
	# even if the user has specifed utf-8 when that is not correct,
	# Tcl converts it to utf-8.

	variable ::options::pfmOptions

	set outFileName {}
	if {![file exists $pfmOptions(tmpdir)]} then {
	    if {[catch {file mkdir $pfmOptions(tmpdir)} errMsg]} then {
		tk_messageBox -type ok -icon error -message $errMsg
	    }
	}
	if {[catch {open $fileName r} inFile]} then {
	    tk_messageBox -type ok -icon error -message $inFile
	} else {
	    fconfigure $inFile -encoding $fromEncoding
	    set tmpName "pfm_$pfmOptions(user)_[pid].sql"
	    set outFileName [file join $pfmOptions(tmpdir) $tmpName]
	    if {[catch {open $outFileName w} outFile]} then {
		tk_messageBox -type ok -icon error -message $outFile
		set outFileName {}
	    } else {
		fconfigure $outFile -encoding utf-8
		while {![eof $inFile]} {
		    puts $outFile [gets $inFile]
		}
		close $inFile
		close $outFile
	    }
	}
	return $outFileName
    }

    proc cmdPrint {txtWidget} {
	variable ::options::pfmOptions
	variable printCommand
	variable parmlist

	set x [expr [winfo rootx .report] +20]
	set y [expr [winfo rooty .report] +20]
	toplevel .report.tpPrint -class Toplevel
	wm transient .report.tpPrint .report
	wm geometry .report.tpPrint +$x+$y
	wm title .report.tpPrint "pfm - Print Output"
	set printCommand $pfmOptions(printcmd)
	message .report.tpPrint.msgPrintcmd -width 500 -justify center -text \
	    "$printCommand\n\nLongest line is: [longestLine .report.txtOutput]"
	grid .report.tpPrint.msgPrintcmd -column 0 -row 0 -columnspan 3
	set n 1
	# Get the parameters for the printcommand
	set parmlist {}
	set startOfParm [string first "\$(" $printCommand 0]
	if { $startOfParm >= 0 } then {
	    set n 1
	    while { $startOfParm >= 0 } {
		set endOfParm [string first ")" $printCommand $startOfParm]
		if { $endOfParm >= 0 } then {
		    set parm [string range $printCommand $startOfParm $endOfParm]
		    set equalSign [string first "=" $parm 0]
		    if { $equalSign >= 0 } then {
			set defVal [string range $parm [expr $equalSign + 1] "end-1"]
			set parmName [string range $parm 2 [expr $equalSign - 1]]
		    } else {
			set defVal {}
			set parmName [string range $parm 2 "end-1"]
		    }
		    label .report.tpPrint.lb$n -text $parmName
		    grid .report.tpPrint.lb$n -column 0 -row $n
		    entry .report.tpPrint.en$n -width 40 -background white
		    .report.tpPrint.en$n insert end $defVal
		    grid .report.tpPrint.en$n -column 1 -columnspan 2 -row $n
		    lappend parmlist $parm
		    set startOfParm \
			[string first "\$(" $printCommand [expr $endOfParm + 1]]
		    incr n
		} else {
		    set startOfParm -1
		}
	    }
	    focus .report.tpPrint.en1
	}
	button .report.tpPrint.bnOK -text OK \
	    -command "::report::cmdPrintOK $txtWidget" \
	    -underline 0
	button .report.tpPrint.bnCancel -text Cancel \
	    -command {destroy .report.tpPrint} \
	    -underline 0
	grid .report.tpPrint.bnOK -column 1 -row $n -sticky we
	grid .report.tpPrint.bnCancel -column 2 -row $n -sticky we
	bind .report.tpPrint <Alt-KeyPress-o> \
	    "::pfm::PressButton .report.tpPrint.bnOK"
	bind .report.tpPrint <Alt-KeyPress-c> \
	    "::pfm::PressButton .report.tpPrint.bnCancel"
	bind .report.tpPrint <KeyPress-Escape> \
	    "::pfm::PressButton .report.tpPrint.bnCancel"
	return
    }

    proc cmdPrintOK {txtWidget} {
	variable printCommand
	variable parmlist
	variable ::options::pfmOptions

	# This part was completely rewritten.
	# The lappend function ensures that the arguments of the openCommand
	# are properly delimited with { and } where necessary.
	# Also possibility to use temp file was added: %s in printCommand
	# represents the temporary file.

	if {[string first {%s} $printCommand] <= 0} then {
	    # input via stdin
	    set tempFile 0
	} else {
	    # input via temporary file
	    if {![file exists $pfmOptions(tmpdir)]} then {
		if {[catch {file mkdir $pfmOptions(tmpdir)} errMsg]} then {
		    tk_messageBox -type ok -icon error -message $errMsg
		}
	    }
	    set tempFile 1
	    set tmpName "pfm_$pfmOptions(user)_[pid].txt"
	    set fileName [file normalize [file join $pfmOptions(tmpdir) $tmpName]]
	    set mapTemp {%s}
	    lappend mapTemp $fileName
	}
	set execCommand {}
	foreach arg $printCommand {
	    set n 1
	    foreach parm $parmlist {
		set value [.report.tpPrint.en$n get]
		set map $parm
		lappend map $value
		set arg [string map $map $arg]
		incr n
	    }
	    if {$tempFile} then {
		set arg [string map $mapTemp $arg]
	    }
	    lappend execCommand $arg
	}
	destroy .report.tpPrint
	if {$tempFile} then {
	    set openCommand $fileName
	} else {
	    set openCommand [linsert $execCommand 0 {|}]
	}
	# puts $openCommand
	if { [catch {open $openCommand w} printch] } then {
	    tk_messageBox -type ok -icon error -message $printch
	} else {
	    # Next line added because of feature request 693.
	    fconfigure $printch -encoding $pfmOptions(printencoding)
	    puts $printch [$txtWidget get 1.0 end]
	    if { [catch {close $printch} errMsg] } then {
		tk_messageBox -message $errMsg -type ok -icon info
	    }
	    if {$tempFile} then {
		set execCommand [linsert $execCommand 0 {exec}]
		lappend execCommand {&}
		# puts $execCommand
		if {[catch $execCommand errMsg]} then {
		    tk_messageBox -type ok -icon error -message $errMsg
		}
	    }
	}
	return
    }

    proc longestLine {txtWidget} {

	set longest 0
	set lastIndex [$txtWidget index end]
	set index [$txtWidget index 1.0]
	while { $index < $lastIndex } {
	    set thisLineLength [string length [$txtWidget get $index "$index lineend"]]
	    if { $longest < $thisLineLength } then {
		set longest $thisLineLength
	    }
	    set index [$txtWidget index "$index +1 lines"]
	}
	return $longest
    }

    proc cmdRun {} {
	variable reportMode

	switch $reportMode {
	    sql {
		cmdRunSQL
	    }
	    report {
		cmdRunReport [.report.fmReport.lsb curselection]
	    }
	}
	return
    }


    proc {cmdQuitReport} {} {
	variable commandHistory
	variable windowSize

	destroy .report
	array unset commandHistory
    }

###########################################################################
#  Procedures for queries (SQL mode)                                      #
###########################################################################

    proc cmdEnterSQLmode { } {
	variable ::pfm::dbName
	variable reportMode

	wm title .report "pfm - Run SQL : $dbName"
	destroy .report.fmReport
	destroy .report.fmSQL
	frame .report.fmSQL
	label .report.fmSQL.title -text {SQL statement}
	text .report.fmSQL.text -wrap none \
	    -background white \
	    -xscrollcommand {.report.fmSQL.hscroll set} \
	    -yscrollcommand {.report.fmSQL.vscroll set}
	scrollbar .report.fmSQL.hscroll -orient horizontal \
	    -command {.report.fmSQL.text xview} -takefocus 0
	scrollbar .report.fmSQL.vscroll -orient vertical \
	    -command {.report.fmSQL.text yview} -takefocus 0
	place .report.fmSQL \
	    -anchor nw -x 0 -y 0 \
	    -relwidth 1 \
	    -height -30 -relheight 0.4
	place .report.fmSQL.title \
	    -anchor n -x 0 -y 0 -relx 0.5 \
	    -height 15
	place .report.fmSQL.text \
	    -anchor nw -x 0 -y 15 \
	    -width -20 -relwidth 1 \
	    -height -35 -relheight 1
	place .report.fmSQL.hscroll \
	    -anchor sw -x 0 -y 0 -rely 1 \
	    -width -20 -relwidth 1 \
	    -height 20
	place .report.fmSQL.vscroll \
	    -anchor ne -x 0 -y 15 -relx 1 \
	    -width 20 \
	    -height -35 -relheight 1
	    
	focus .report.fmSQL.text
	# enable clear button
	.report.fmMiddle.bnClear configure -state normal

	# the status of the Previous and Next buttons is controlled by
	# procedures handling the command history and by
	# cmdEnterReportMode.

	# enable SQL menu
	.report.menubar entryconfigure 1 -state normal
	# enable Help menu
	.report.menubar entryconfigure 3 -state normal
	cmdClear
	return
    }

    proc showResult { } {
	variable ::pfm::psqlChannel

	if { ![winfo exists .report.txtOutput] } then {
	    # If psql sends output when the .report window is not existing,
	    # we create the .report window.
	    ::report::cmdReportSQL {sql}
	}
	if { ![eof $psqlChannel] } then {
	    .report.txtOutput insert end "[gets $psqlChannel]\n"
	    .report.txtOutput see end
	} else {
	    # Bug 690: pfm hangs when psql exits.
	    # This branch is necessary to avoid endless loops when the user types '\q',
	    # or in the unlikely event that psql dies.
	    fileevent $psqlChannel readable {}
	    unset psqlChannel
	    puts [info exists psqlChannel]
	    .report.txtOutput insert end "Connection with psql is closed.\n"
	    .report.txtOutput see end
	}
	return
    }

    proc showError { } {
	variable ::pfm::errChannel

	if { ![winfo exists .report.txtOutput] } then {
	    # If psql sends output when the .report window is not existing,
	    # we create the .report window.
	    ::report::cmdReportSQL {sql}
	}
	if { ![eof $errChannel] } then {
	    set begin [.report.txtOutput index "end - 1 chars"]
	    .report.txtOutput insert end "[gets $errChannel]\n"
	    .report.txtOutput see end
	    set end [.report.txtOutput index "end - 1 chars"]
	    .report.txtOutput tag add errTag $begin $end
	} else {
	    fileevent $errChannel readable {}
	    unset errChannel
	    .report.txtOutput insert end "Error channel has died.\n"
	    .report.txtOutput see end
	}
	return
    }

    proc cmdClear {} {
	variable commandHistory

	.report.fmSQL.text delete 1.0 end
	set commandHistory($commandHistory(top)) {}
	# Added because of bug 691. If user presses [Clear], cursor must
	# be reset to top of command history.
	set commandHistory(cursor) $commandHistory(top)
	if { $commandHistory(top) > 0 } then {
	    .report.fmMiddle.bnPrev configure -state normal
	} else {
	    .report.fmMiddle.bnPrev configure -state disabled
	}
	.report.fmMiddle.bnNext configure -state disabled
	return
    }

    proc initCommandHistory {} {
	variable commandHistory

	set commandHistory(top) 0
	set commandHistory(cursor) 0
	set commandHistory(0) {}
	return
    }

    proc storeCommand {} {
	variable commandHistory

	# Store command in first free element of commandHistory array
	# Do not store \i commands
	set commandStart [.report.fmSQL.text get 1.0 1.2]
	if {$commandStart ne {\i}} then {
	    set commandHistory($commandHistory(top)) [.report.fmSQL.text get 1.0 "end -1 chars"]
	    # Increment top to first free element.
	    incr commandHistory(top)
	    set commandHistory($commandHistory(top)) {}
	    set commandHistory(cursor) $commandHistory(top)
	    .report.fmMiddle.bnPrev configure -state normal
	    .report.fmMiddle.bnNext configure -state disabled
	}
	return
    }

    proc cmdPrev {} {
	variable commandHistory

	if { $commandHistory(cursor) == $commandHistory(top) } then {
	    set commandHistory($commandHistory(top)) [.report.fmSQL.text get 1.0 "end -1 chars"]
	}
	# Reworked because of bug 691.
	
	if { $commandHistory(cursor) > 0 } then {
	    incr commandHistory(cursor) -1
	    .report.fmSQL.text delete 1.0 end
	    .report.fmSQL.text insert end $commandHistory($commandHistory(cursor))
	    .report.fmMiddle.bnNext configure -state normal
	    if { $commandHistory(cursor) == 0} then {
		.report.fmMiddle.bnPrev configure -state disabled
	    }
	} else {
	    bell
	}
	return
    }

    proc cmdNext {} {
	variable commandHistory

	# Reworked because of bug 691.
	if { $commandHistory(cursor) < $commandHistory(top) } then {
	    incr commandHistory(cursor)
	    .report.fmSQL.text delete 1.0 end
	    .report.fmSQL.text insert end $commandHistory($commandHistory(cursor))	    
	    if { $commandHistory(cursor) == $commandHistory(top) } then {
		.report.fmMiddle.bnNext configure -state disabled
	    }
	    .report.fmMiddle.bnPrev configure -state normal
	} else {
	    bell
	}
	return
    }

    proc cmdRunSQL {} {
	variable ::pfm::psqlChannel
	variable ::pfm::errChannel

	storeCommand
        set sqlCmd [.report.fmSQL.text get 1.0 end]
	if { [info exists psqlChannel] } then {
	    if { [catch {
		puts $psqlChannel $sqlCmd
		flush $psqlChannel
	    } errMsg] } then {
		tk_messageBox -message $errMsg -type ok -icon error
	    }
	} else {
	    puts $errChannel \
		"No connection with psql.\nTry to close and reopen the database.\n"
	    flush $errChannel
	}
	.report.fmSQL.text delete 1.0 end
	return
    }

    proc cmdShortCut {command} {

	.report.fmSQL.text insert end $command
	cmdRunSQL
	return
    }

###################################################################
#  Procedures for report mode                                     #
###################################################################

    proc cmdEnterReportMode { } {
	variable reportMode
	variable ::pfm::dbName

	wm title .report "pfm - Run report : $dbName"
	destroy .report.fmSQL
	destroy .report.fmReport
	.report.fmMiddle.bnClear configure -state disabled
	.report.fmMiddle.bnNext configure -state disabled
	.report.fmMiddle.bnPrev configure -state disabled

	# disable SQL menu
	.report.menubar entryconfigure 1 -state disabled

	# disable Help Menu
	.report.menubar entryconfigure 3 -state disabled
	cmdDisplayReportList
	return
    }


    proc cmdDisplayReportList {} {
	variable ::pfm::currentDB
	variable reportDef
	variable ::options::pfmOptions

	if { ![info exists currentDB] } then {
	    tk_messageBox -message "There is no data base open!" -type ok \
		-icon error
	} else {
	    set queryDef "SELECT * FROM pfm_report ORDER BY name"
	    set queryRes [pg_exec $currentDB $queryDef]
	    set lastTuple [expr [pg_result $queryRes -numTuples] - 1]
	    pg_result $queryRes -assign reportDef
	    pg_result $queryRes -clear
	    if { $lastTuple < 0} then {
		tk_messageBox -message "There are no reports defined!" \
		    -type ok -icon info -parent .report
	    } else {
		set maxNameWidth 0
		for {set tuple 0} {$tuple <= $lastTuple} {incr tuple } {
		    set nameLength [string length $reportDef($tuple,name)]
		    if { $nameLength > $maxNameWidth } then {
			set maxNameWidth $nameLength
		    }
		}
		set reportList {}
		for {set tuple 0} {$tuple <= $lastTuple} {incr tuple } {
		    set name [format "%-$maxNameWidth\s" $reportDef($tuple,name)]
		    lappend reportList \
			    "$name : $reportDef($tuple,description)"
		}
		destroy .report.fmReport
		frame .report.fmReport
		label .report.fmReport.title -text "List of reports"
		listbox .report.fmReport.lsb \
		    -background white \
		    -yscrollcommand {.report.fmReport.vscroll set}
		scrollbar .report.fmReport.vscroll -orient vertical \
		    -command {.report.fmReport.lsb yview} -takefocus 0
		foreach report $reportList {
		    .report.fmReport.lsb insert end $report
		}
		.report.fmReport.lsb selection clear 0 end
		.report.fmReport.lsb selection set 0 0
		.report.fmReport.lsb activate 0
		bind .report.fmReport.lsb <KeyPress-Return> {
		    ::pfm::PressButton .report.fmMiddle.bnRun
		}
		bind .report.fmReport.lsb <ButtonPress-1> {
		    focus .report.fmReport.lsb
		}
		place .report.fmReport \
		    -anchor nw -x 0 -y 0 \
		    -relwidth 1 \
		    -height -30 -relheight 0.4
       		place .report.fmReport.title \
		    -anchor n -x 0 -y 0 -relx 0.5 \
		    -height 15
		place .report.fmReport.lsb \
		    -anchor nw -x 0 -y 15 \
		    -width -20 -relwidth 1 \
		    -height -15 -relheight 1
		place .report.fmReport.vscroll \
		    -anchor ne -x 0 -y 15 -relx 1 \
		    -width 20 \
		    -height -15 -relheight 1
		focus .report.fmReport.lsb
	    }
	}
	return
    }

    proc cmdRunReport {selectedReport} {
	variable reportDef
	variable ::pfm::currentDB
	variable parmlist

	# Get the parameters necessary to execute the query
	set sqlselect $reportDef($selectedReport,sqlselect)
	set startOfParm [string first "\$(" $sqlselect 0]
	if { $startOfParm >= 0 } then {
	    set parmlist {}
	    set x [expr [winfo rootx .report] + 20]
	    set y [expr [winfo rooty .report] + 20]
	    toplevel .report.getparm -class Toplevel
	    wm transient .report.getparm .report
	    wm geometry .report.getparm 600x400+$x+$y
	    wm title .report.getparm "pfm - Get report parameters"
	    text .report.getparm.txtselect -wrap word \
		-yscrollcommand {.report.getparm.vscroll set} \
		-background white -takefocus 0
	    scrollbar .report.getparm.vscroll -orient vertical \
		-command {.report.getparm.txtselect yview} \
		-takefocus 0
	    place .report.getparm.txtselect \
		-anchor nw -x 0 -y 0 \
		-width -20 -relwidth 1 \
		-height 150
	    place .report.getparm.vscroll \
		-anchor ne -x 0 -y 0 -relx 1 \
		-width 20 \
		-height 150
	    .report.getparm.txtselect insert end \
		$reportDef($selectedReport,sqlselect)
	    frame .report.getparm.frmparm -relief sunken -borderwidth 2
	    set frmparm .report.getparm.frmparm
	    place .report.getparm.frmparm \
		-anchor nw -x 0 -y 150 \
		-relwidth 1 \
		-height -180 -relheight 1
	    set n 0
	    while { $startOfParm >= 0 } {
		set endOfParm [string first ")" $sqlselect $startOfParm]
		if { $endOfParm >= 0 } then {
		    set parm [string range $sqlselect $startOfParm $endOfParm]
		    if { [lsearch -exact $parmlist $parm] == -1} then {
			# It is a new parameter. If the parameter is already
			# in parmlist, there is no need to prompt the user again.
			set labelText [string range $sqlselect \
					   [expr $startOfParm +2] [expr $endOfParm - 1]]
			label .report.getparm.frmparm.lb$n -text $labelText
			grid .report.getparm.frmparm.lb$n -column 0 -row $n
			entry .report.getparm.frmparm.en$n -width 40 \
			    -background white
			grid .report.getparm.frmparm.en$n -column 1 -row $n
			lappend parmlist $parm
			incr n
		    }
		    set startOfParm \
			[string first "\$(" $sqlselect [expr $endOfParm + 1]]
		} else {
		    set startOfParm -1
		}
	    }
	    focus .report.getparm.frmparm.en0
	    button .report.getparm.bnOK -text OK \
		-command "::report::completeSqlselect $selectedReport" \
		-underline 0
	    place .report.getparm.bnOK \
		 -anchor s -x -30 -y 0 -relx 0.5 -rely 1
	    button .report.getparm.bnCancel -text Cancel \
		-command {destroy .report.getparm} \
		-underline 0
	    place .report.getparm.bnCancel \
		-anchor s -x 30 -y 0 -relx 0.5 -rely 1
	    bind .report.getparm <Alt-KeyPress-o> \
		"::pfm::PressButton .report.getparm.bnOK"
	    bind .report.getparm <Alt-KeyPress-c> \
		"::pfm::PressButton .report.getparm.bnCancel"
	    bind .report.getparm <KeyPress-Escape> \
		"::pfm::PressButton .report.getparm.bnCancel"
	} else {
	    executeQuery $selectedReport $sqlselect
	}
	return
    }

    proc completeSqlselect { selectedReport } {
	variable parmlist
	variable reportDef

	set sqlselect $reportDef($selectedReport,sqlselect)
	set n 0
	foreach parm $parmlist {
	    set value [.report.getparm.frmparm.en$n get]
	    set sqlselect [string map "$parm \"$value\"" $sqlselect]
	    incr n
	}
	destroy .report.getparm
	executeQuery $selectedReport $sqlselect
	return
    }

    proc executeQuery {selectedReport sqlselect } {
	variable reportDef
	variable ::pfm::currentDB
	variable ::pfm::errChannel
	variable reportData

	# Execute the query for the report and store the result in reportData

	set queryRes [pg_exec $currentDB $sqlselect]
	set queryStatus [pg_result $queryRes -status]
	if { [string equal $queryStatus "PGRES_TUPLES_OK"] } then {
	    pg_result $queryRes -assign reportData
	    set lastTuple [expr [pg_result $queryRes -numTuples] - 1]
	    printReport $selectedReport $lastTuple $sqlselect
	} else {
	    set errmsg "$sqlselect failed\n"
	    set errmsg "$errmsg [pg_result $queryRes -error]\n"
	    puts $errChannel $errmsg
	    flush $errChannel
	}
	pg_result $queryRes -clear
	return
    }

    proc printReport {selectedReport lastTuple sqlselect} {
	variable reportDef
	variable fieldNamesList
	variable printInfo
	variable reportData
	variable layout
	variable maxLabelWidth
	variable targetFields
	variable sourceFields
	variable summaryList
	variable summaryDef
	variable runBegin
	variable runEnd
	variable lowestPL
	variable summary	

	set lastLevel [collectReportLayout $selectedReport]

	if {[checkReportDef $lastLevel $lastTuple]} then {

	    calcRunsAndLowestPL $lastTuple $lastLevel

	    # Calculate maximum field width and the optimum columwidth
	    # for a table layout. Store result in printInfo.

	    for { set level 1 } { $level <= $lastLevel} { incr level} {
		foreach field $fieldNamesList($level) {
		    set printInfo($field,width) 0
		    for {set tuple 0} { $tuple <= $lastTuple } { incr tuple } {
			set width [string length $reportData($tuple,$field)]
			if { $printInfo($field,width) < $width } then {
			    set printInfo($field,width) $width
			}
		    }
		    set labelLength [string length $printInfo($field,label)]
		    if { $printInfo($field,width) < $labelLength } then {
			set printInfo($field,columnWidth) $labelLength
		    } else {
			set printInfo($field,columnWidth) $printInfo($field,width)
		    }
		}
	    }

	    # Print report title

	    .report.txtOutput insert end "$reportDef($selectedReport,name)\n"
	    .report.txtOutput insert end \
		"[string repeat - [string length $reportDef($selectedReport,name)]]\n\n"

	    .report.txtOutput insert end \
		"Description: $reportDef($selectedReport,description)\n"
	    set formattedSQL [string map {\n "\n             "} $sqlselect]
	    .report.txtOutput insert end "SQL        : $formattedSQL\n"
	    .report.txtOutput insert end \
		"Date       : [clock format [clock seconds] -format {%d-%b-%Y}]\n\n"

	    # print the reportData. See comments in procedure
	    # calcRunsAndLowestPL

	    for {set tuple 0} {$tuple <= $lastTuple} {incr tuple} {
		for {set level 1} {$level <= $lastLevel} {incr level} {
		    if {$runBegin($tuple,$level)} then {
			initSummary $level
		    }
		}
		for {set level 1} {$level <= $lastLevel} {incr level} {
		    if {$level != $lastLevel} then {
			if {$level >= $lowestPL($tuple)} then {
			    updateSummary $level $tuple
			    .report.txtOutput insert end \
				[printLevel 1 $level $tuple]
			}
		    } else {
			updateSummary $level $tuple
			.report.txtOutput insert end \
			    [printLevel $runBegin($tuple,$level) $level $tuple]
		    }
		}
		for {set level $lastLevel} {$level >= 1} {incr level -1} {
		    if {$runEnd($tuple,$level)} then {
			printSummary $level
		    }
		}
	    }
	    .report.txtOutput insert end "\n\n"
	    .report.txtOutput see end
	}
	unset fieldNamesList
	array unset printInfo
	array unset reportData
	array unset layout
	array unset maxLabelWidth
	unset targetFields
	unset sourceFields
	unset summaryList
	array unset summaryDef
	array unset runBegin
	array unset runEnd
	array unset lowestPL
	array unset summary
	return
    }

    proc collectReportLayout {selectedReport} {
	variable ::pfm::currentDB
	variable layout
	variable fieldNamesList
	variable printInfo
	variable reportDef
	variable maxLabelWidth
	variable targetFields
	variable sourceFields
	variable summaryDef
	variable summaryList
	variable ::pfm::errChannel

	# Get data from pfm_section and store them in sectionArray

	set queryDef "SELECT * FROM pfm_section WHERE report='$reportDef($selectedReport,name)'"
	set queryDef "$queryDef ORDER BY level"
	set queryRes [pg_exec $currentDB $queryDef]
	set lastTuple [expr [pg_result $queryRes -numTuples] - 1]
	pg_result $queryRes -assign sectionArray
	pg_result $queryRes -clear
	set lastLevel 0
	set shouldBeLevel 1

	# Store data in layout, fieldNamesList and printInfo

	for {set tuple 0} {$tuple <= $lastTuple} {incr tuple } {
	    if {$sectionArray($tuple,level) != $shouldBeLevel} then {
		set errMsg "\nWARNING: The levels of pfm_section should be numbered consecutively, starting from 1! "
		append errMsg "\nLevel '$sectionArray($tuple,level)' should become level '$shouldBeLevel'!\n"
		puts $errChannel $errMsg
		flush $errChannel
		set sectionArray($tuple,level) $shouldBeLevel
	    }
	    if { $sectionArray($tuple,level) > $lastLevel} then {
		set lastLevel $sectionArray($tuple,level)
	    }
	    set layout($shouldBeLevel) $sectionArray($tuple,layout)
	    set fieldDefList $sectionArray($tuple,fieldlist)
	    set fieldNamesList($shouldBeLevel) {}
	    foreach item $fieldDefList {
		set attribName [lindex $item 0]
		lappend fieldNamesList($shouldBeLevel) $attribName
		set printInfo($attribName,label) [lindex $item 1]
		set printInfo($attribName,alignment) [lindex $item 2]
	    }
	    incr shouldBeLevel
	}

	# Calculate maxLabelWidth per level

	for { set level 1 } { $level <= $lastLevel } { incr level } {
	    set maxLabelWidth($level,section) 0
	    foreach field $fieldNamesList($level) {
		set width [string length $printInfo($field,label)]
		if { $width > $maxLabelWidth($level,section) } then {
		    set maxLabelWidth($level,section) $width
		}
	    }
	}

	# Read summary definitions

	for {set tuple 0} {$tuple <= $lastTuple} {incr tuple} {
	    set level $sectionArray($tuple,level)
	    set targetFields($level) {}
	    set summaryList($level) {}
	    set seqnr 1
	    foreach summaryItem $sectionArray($tuple,summary) {
		set field [lindex $summaryItem 0]
		set summaryDef($level,$seqnr,field) $field
		if {[lsearch $targetFields($level) $field] == -1} then {
		    lappend targetFields($level) $field
		}
		set summaryDef($level,$seqnr,aggregate) [lindex $summaryItem 1]
		set summaryDef($level,$seqnr,format) [lindex $summaryItem 2]
		lappend summaryList($level) $seqnr
		incr seqnr
	    }
	}
	
	# Check summary definitions and calculate sourceFields($level) and
	# maxLabelWidth($level,summary)

	for {set level 1} {$level <= $lastLevel} {incr level} {
	    set sourceFields($level) {}
	}
	for {set level 1} {$level <= $lastLevel} {incr level} {
	    set maxLabelWidth($level,summary) 0
	    foreach seqnr $summaryList($level) {
		set sourceLevel 0
		set field $summaryDef($level,$seqnr,field)
		for {set searchLevel $level} {$searchLevel <= $lastLevel} \
		    {incr searchLevel} {
			if {[lsearch $fieldNamesList($searchLevel) $field] >= 0} then {
			    set sourceLevel $searchLevel
			    break
			}
		}
		if {$sourceLevel == 0} then {
		    set errMsg "\nERROR: Summary field '$field' in section '$level'"
		    append errMsg \
			"\neither is not defined as a normal field in any section,"
		    append errMsg \
			"\nor it is defined in a section with level lower than $level."
		    append errMsg \
			"\nSummary field '$field' has been discarded.\n"
		    puts $errChannel $errMsg
		    flush $errChannel
		    set fieldIndex [lsearch $targetFields($level) $field]
		    if {$fieldIndex >= 0} then {
			set targetFields($level) \
			    [lreplace $targetFields($level) $fieldIndex $fieldIndex]
		    }
		    unset summaryDef($level,$seqnr,aggregate)
		    unset summaryDef($level,$seqnr,format)
		    unset summaryDef($level,$seqnr,field)
		    set seqIndex [lsearch $summaryList($level) $seqnr]
		    if {$seqIndex >= 0} then {
			set summaryList($level) \
			    [lreplace $summaryList($level) $seqIndex $seqIndex]
		    }
		} else {
		    if {[lsearch $sourceFields($sourceLevel) $field] == -1} then {
			lappend sourceFields($sourceLevel) $field
		    }
		    set labelWidth \
			[expr [string length $printInfo($field,label)] + \
			     [string length $summaryDef($level,$seqnr,aggregate)] + 2]
		    if { $maxLabelWidth($level,summary) < $labelWidth} then {
			set maxLabelWidth($level,summary) $labelWidth
		    }
		}
	    }
	}
	array unset sectionArray
	return $lastLevel
    }

    proc calcRunsAndLowestPL {lastTuple lastLevel} {
	# If for a record, the fields of a section all have the same
	# values as the corresponding fields of the previous record, then
	# the fields of that section are not printed for that record.
	# The fields of the highest numbered section (lastLevel) are
	# an exception. They are always printed.
	#
	# A run of level n, consists of a number of consecutive records
	# of the reportdata in which all data of sections with level lower
	# than n remain unchanged.
	#
	# This procedure calculates for each $tuple which is the lowest
	# level section that needs to be printed and stores the result in
	# lowestPL($tuple).
	#
	# It also calculates for each $tuple whether it is the begin or end
	# of a run of $level and stores the results in runBegin($tuple,$level)
	# and runEnd($tuple,$level).

	variable runBegin
	variable runEnd
	variable lowestPL

	# for the 1st record, all sections need to be printed
	set lowestPL(0) 1
	# init the runBegin and runEnd arrays
	for {set tuple 0} {$tuple <= $lastTuple} {incr tuple} {
	    for {set level 1} {$level <= $lastLevel} {incr level} {
		set runBegin($tuple,$level) 0
		set runEnd($tuple,$level) 0
	    }
	}
	# The 1st record is the begin of a run for all sections (levels).
	# The last record is the end of a run for all sections.
	for {set level 1} {$level <= $lastLevel} {incr level} {
	    set runBegin(0,$level) 1
	    set runEnd($lastTuple,$level) 1
	}
	for {set tuple 1} {$tuple <= $lastTuple} {incr tuple} {
	    # The $lastLevel is printed always
	    set lowestPL($tuple) $lastLevel
	    for {set level 1} {$level <= [expr $lastLevel - 1]} {incr level} {
		if {[newValues $level $tuple]} then {
		    # The values of the fields of this level for this record
		    # are different from the values of the same fields for
		    # the previous record.
		    # The fields of this and all higher levels need
		    # to be printed.
		    set lowestPL($tuple) $level
		    # This record is the begin of run for all levels
		    # higher than this one.
		    # The previous record is the end of a run for all
		    # levels higher than this one.
		    for {set higherLevel [expr $level + 1]} \
			{$higherLevel <= $lastLevel} {incr higherLevel} {
			    set runEnd([expr $tuple - 1],$higherLevel) 1
			    set runBegin($tuple,$higherLevel) 1
			}
		    break
		}
	    }
	}
	return
    }

    proc newValues {level tuple} {
	variable fieldNamesList
	variable reportData

	if { $tuple == 0 } then {
	    set returnValue 1
	} else {
	    set returnValue 0
	    foreach field $fieldNamesList($level) {
		if { $reportData($tuple,$field) ne $reportData([expr $tuple - 1],$field) } then {
		    set returnValue 1
		}
	    }
	}
	return $returnValue
    }

    proc checkReportDef {lastLevel lastTuple} {
	variable fieldNamesList
	variable reportData
	variable ::pfm::errChannel

	set result 1
	if {$lastTuple == -1} then {
	    set errMsg "\nWARNING: The report's SQL SELECT statement did not return any data.\n"
	    puts $errChannel $errMsg
	    flush $errChannel
	} else {
	    for {set level 1} {$level <= $lastLevel} {incr level} {
		foreach field $fieldNamesList($level) {
		    if {![info exists reportData(0,$field)]} then {
			set result 0
			set errMsg \
			    "\nERROR: The report's SQL SELECT statement does not return data"
			append errMsg \
			    "\nfor field '$field' defined in section '$level' of the report."
			append errMsg \
			    "\nCheck the report definition!\n"
			puts $errChannel $errMsg
			flush $errChannel
		    }
		}
	    }
	}
	return $result
    }

    proc initSummary {targetLevel} {
	# procedure to be called at the begin of each $targetLevel run
	# it inits all lists
	# targetLevel is the level for which the summary is calculated.
	variable targetFields
	variable summary

	# .report.txtOutput insert end "initSummary $targetLevel\n"
	foreach field $targetFields($targetLevel) {
	    set summary($field,$targetLevel) {}
	}
	return
    }

    proc updateSummary {sourceLevel tuple} {
	# procedure to be called every time a new value for $sourecLevel section
	# is being printed.
	# sourceLevel is the level/section in which a field is defined as an
	# ordinary field.
	variable reportData
	variable summary
	variable sourceFields
	variable targetFields

	foreach field $sourceFields($sourceLevel) {
	    for {set targetLevel 1} {$targetLevel <= $sourceLevel} {incr targetLevel} {
		if {([lsearch $targetFields($targetLevel) $field] >= 0) && \
			($reportData($tuple,$field) ne {})} then {
		    # Null fields are discarded. Else most aggregate functions
		    # will not give the right result when a LEFT JOIN does not
		    # yield a record in the right table.
		    lappend summary($field,$targetLevel) $reportData($tuple,$field)
		}
	    }
	}
	return
    }
    
    proc printSummary {targetLevel} {
	variable summary
	variable summaryDef
	variable targetFields
	variable summaryList
	variable ::pfm::errChannel

	set firstLine 1
	if {$summaryList($targetLevel) ne {}} then {
	    .report.txtOutput insert end "\n"
	}
	foreach seqnr $summaryList($targetLevel) {
	    if {[info commands \
		     ::aggregate::$summaryDef($targetLevel,$seqnr,aggregate)] \
		    ne {}} then {
		set field $summaryDef($targetLevel,$seqnr,field)
		set value [::aggregate::$summaryDef($targetLevel,$seqnr,aggregate) \
			       $summary($field,$targetLevel)]
		printSummaryLine $targetLevel $seqnr $value $firstLine
		set firstLine 0
	    } else {
		set errMsg "ERROR: Aggregate function '$summaryDef($targetLevel,$seqnr,aggregate)' is not defined."
		puts $errChannel "$errMsg\n"
		flush $errChannel
	    }
	}
	return
    }

    proc printSummaryLine {level seqnr value firstLine} {
	variable summaryDef
	variable printInfo
	variable maxLabelWidth

	set offset [string repeat " " [expr 4 * ($level - 1)]]
	if {$firstLine} then {
	    set textToPrint "$offset\Summary: "
	} else {
	    set textToPrint "$offset         "
	}
	set field $summaryDef($level,$seqnr,field)
	set label $printInfo($field,label)
	set label "$summaryDef($level,$seqnr,aggregate)($label)"
	set formatLabel "%-$maxLabelWidth($level,summary)\s"
	set formatString $summaryDef($level,$seqnr,format)
	set textToPrint "$textToPrint[format $formatLabel $label] = "
	if {$formatString ne {}} then {
	    set textToPrint "$textToPrint[format $formatString $value]\n"
	} else {
	    set textToPrint "$textToPrint$value\n"
	}
	.report.txtOutput insert end $textToPrint
	.report.txtOutput see end
	return
    }

    proc printLevel {firstRecordOfTable level tuple} {
	variable layout

	switch $layout($level) {
	    row {
		set lineToPrint [printRow $level $tuple]
	    }
	    column {
		set lineToPrint [printColumn $firstRecordOfTable $level $tuple]
	    }
	    table {
		set lineToPrint [printTable $firstRecordOfTable $level $tuple]
	    }
	}
  	if { $firstRecordOfTable } then {
  	    set lineToPrint "\n$lineToPrint"
  	}
	return $lineToPrint
    }



    proc printRow {level tuple} {
	variable reportData
	variable fieldNamesList
	variable printInfo

	set offset [string repeat " " [expr 4 * ($level - 1)]]
	set lineToPrint $offset
	foreach field $fieldNamesList($level) {
	    append lineToPrint "$printInfo($field,label):"
	    append lineToPrint "$reportData($tuple,$field); "
	}
	append lineToPrint "\n"
	return $lineToPrint
    }


    proc printColumn {firstRecordOfTable level tuple} {
	variable reportData
	variable fieldNamesList
	variable printInfo
	variable maxLabelWidth

	set offset [string repeat " " [expr 4 * ($level - 1)]]
	if {$firstRecordOfTable} then {
	    # printColumn normally prints an empty line before each record,
	    # except for the first record because printLevel has already
	    # printed it.
	    set textToPrint {}
	} else {
	    set textToPrint "\n"
	}
	foreach field $fieldNamesList($level) {
	    set formatString "%-$maxLabelWidth($level,section)\s"
	    set labelToPrint \
		"$offset[format $formatString $printInfo($field,label)] : "
	    append textToPrint $labelToPrint
	    set nextLineOffset [string repeat { } [string length $labelToPrint]]
	    set startIdx 0
	    set nlIdx [string first "\n" $reportData($tuple,$field) $startIdx]
	    while { $nlIdx >= 0 } {
		append textToPrint \
		    [string range $reportData($tuple,$field) $startIdx $nlIdx]
		append textToPrint $nextLineOffset
		set startIdx [expr $nlIdx + 1]
		set nlIdx [string first "\n" $reportData($tuple,$field) $startIdx]
	    }
	    append textToPrint \
		"[string range $reportData($tuple,$field) $startIdx end]\n"
	}
	return $textToPrint
    }


    proc printHeader {level} {
	variable reportData
	variable fieldNamesList
	variable printInfo

	set offset [string repeat " " [expr 4 * ($level - 1)]]
	set header ""
	set underline ""
	foreach field $fieldNamesList($level) {
	    switch $printInfo($field,alignment) {
		r {
		    set formatString "%$printInfo($field,columnWidth)\s"
		}
		l -
		default {
		    set formatString "%-$printInfo($field,columnWidth)\s"
		}
	    }
	    set header "$header| [format $formatString $printInfo($field,label)] "
	    set underline "$underline\+-[string repeat - $printInfo($field,columnWidth)]-"
	}
	set header [string replace $header 0 0 ""]
	set underline [string replace $underline 0 0 ""]
	return "$offset$header\n$offset$underline\n"
    }

    proc printTable {firstRecordOfTable level tuple} {
	variable reportData
	variable fieldNamesList
	variable printInfo


	if { $firstRecordOfTable } then {
	    set header [printHeader $level]
	} else {
	    set header ""
	}
	set offset [string repeat " " [expr 4 * ($level - 1)]]
	set lineToPrint ""
	foreach field $fieldNamesList($level) {
	    switch $printInfo($field,alignment) {
		r {
		    set formatString "%$printInfo($field,columnWidth)\s"
		}
		l -
		default {
		    set formatString "%-$printInfo($field,columnWidth)\s"
		}
	    }
	    append lineToPrint \
		"| [format $formatString $reportData($tuple,$field)] "
	}
	set lineToPrint [string replace $lineToPrint 0 0 ""]
	return "$header$offset$lineToPrint\n"
    }

    ##################################################################
    # Init namespace report                                          #
    ##################################################################

    set windowSize(.report) {750x550}

}

########################################################################
# End of namespace report                                              #
########################################################################

########################################################################
# Begin of namespace aggregate                                         #
########################################################################

namespace eval aggregate {

    proc SUM {List} {

	set sum 0.0
	foreach item $List {
	    set sum [expr $sum + $item]
	}
	return $sum
    }

    proc COUNT {List} {

	# puts $List
	set count 0
	foreach item $List {
	    incr count
	}
	return $count
    }

    proc AVG {List} {

	set sum 0.0
	set count 0
	foreach item $List {
	    set sum [expr $sum + $item]
	    incr count
	}
	set avg [expr $sum / $count]
	return $avg
    }

    proc STDDEV {List} {

	set sum 0.0
	set count 0
	foreach item $List {
	    set sum [expr $sum + $item]
	    incr count
	}
	set avg [expr $sum / $count]
	set squareDev 0
	foreach item $List {
	    set squareDev [expr $squareDev + pow(($item - $avg),2)]
	}
	set stddev [expr sqrt($squareDev/$count)]
	return $stddev
    }

    proc MIN {List} {

	set min [lindex $List 0]
	foreach item $List {
	    if {$item < $min} then {
		set min $item
	    }
	}
	return $min
    }

    proc MAX {List} {

	set max [lindex $List 0]
	foreach item $List {
	    if {$item > $max} then {
		set max $item
	    }
	}
	return $max
    }

}

########################################################################
# End of namespace aggregate                                           #
########################################################################


###################################################################
# Intialisation of the application programmers  interface (API)   #
###################################################################

if { [catch {source [file join $::pfm::installDir pgin.tcl]} errMsg1 ] } then {
    if { [catch {package require Pgtcl} PgtclVersion] } then {
	set errMsg "$errMsg1\n$PgtclVersion."
	set errMsg "$errMsg\nNeither Pgtcl, nor pgintcl were found."
	set errMsg "$errMsg\npfm cannot connect to postgreSQL."
	tk_messageBox -message $errMsg -type ok -icon error
	set ::pfm::API "pfm cannot communicate with postgreSQL. "
	append ::pfm::API "Neither Pgtcl nor pgintcl are present."
    } else {
	set ::pfm::API "pfm is using Pgtcl $PgtclVersion "
	append ::pfm::API "to communicate with postgreSQL."
    }
} else {
    if { [catch { set pgintclVersion $::pgtcl::version } errMsg] } then {
	set pgintclVersion "???"
    }
    set ::pfm::API "pfm is using pgintcl $pgintclVersion "
    append ::pfm::API "to communicate with postgreSQL."
}


# Define arrow images

image create bitmap ::img::down \
    -file [file join $::pfm::installDir arrow_down.xbm] \
    -maskfile [file join $::pfm::installDir arrow_down_mask.xbm]

package provide pfm $::pfm::pfmVersion
